#include "TimeMeasure.hh"

using namespace std;
using namespace BIAS;

// this file only to instantiate this static member:
std::ofstream BIAS::TimeMeasure::TimerStream_;


// JW, woelk
TimeMeasure::TimeMeasure() 
{
  timerclear(&rtRealSum);
  timerclear(&rtKernelSum);
  timerclear(&rtLast);
  timerclear(&rtActual);

  user = 0;
  running = false;
  _dSum=0.0;

#ifdef WIN32
  kernel = 0;
  idle   = 0;
  _tscStart = _tscEnd = _tscTicks = 0;
#else // WIN32
#  ifdef __i386__
  _uiHiStart=_uiLoStart=_uiHiStop=_uiLoStop=0;
#  endif
#endif //WIN32
}



// JW
//static 
void TimeMeasure::
MSecondsToHumanReadable(const unsigned long & msIN, int & msecOUT,
                        int & seconds, int & minutes, int & hours,
                        int & days, int & weeks)
{     
  unsigned long rest = msIN; // remainder

  msecOUT = rest % 1000; // 1000 msec each second
  rest -= msecOUT;
  rest /= 1000; // now seconds

  seconds = rest % 60; // 60 seconds each minute
  rest -= seconds;
  rest /= 60; // now minutes

  minutes = rest % 60;// 60 minutes each hour
  rest -= minutes;
  rest /= 60; // now hours

  hours = rest % 24;// 24 hours each day
  rest -= hours;
  rest /= 24; // now days

  days = rest % 7;// 7 days a week
  rest -= days;
  rest /= 7; // now week

  weeks = rest;
  // do not split further 
  // because days pe rmonth vary (28-31 days, const 365 day year is too long for me to wait anyhow ;-)
}

//JW
//static 
std::ostream & TimeMeasure::
PrintMsecHumanReadable(const unsigned long msecIN, std::ostream &os)
{
  int ms, sec, minutes, hh, dd, ww;
  MSecondsToHumanReadable(msecIN, ms, sec, minutes, hh, dd, ww);

  bool display=false;
  if (ww>0) display=true;
  if (display) os<<setw(3)<<ww<<" weeks ";

  if (hh>0) display=true;
  if (display) os<<setw(3)<<dd<<" days ";

  if (dd>0) display=true;
  if (display) os<<setw(3)<<hh<<" hours ";

  if (minutes>0) display=true;
  if (display) os<<setw(3)<<minutes<<" minutes ";

  if (sec>0) display=true;
  if (display) os<<setw(3)<<sec<<" sec ";

  if (!display){
    // display zero at least:
    os<<setw(3)<<ms<<" msec";
  } else if (ms>0){
    // suppress displaying 0 msec.
    os<<setw(3)<<ms<<" msec";
  };

  return os;
}


double TimeMeasure::
MeasureOverhead()
{
  /* Do it twice to eliminate cache effects */
  int i;
  double result=0;

  for (i = 0; i < 2; i++) {
    Reset();
    Start();
    Stop();
    result = GetCycleCount();
  }
  return result;
}

double TimeMeasure::
MeasureProcessorFrequency()
{
  double rate;
  int sleeptime=2;

  Start();
#ifdef WIN32
  Sleep(sleeptime  *1000);
#else
  sleep(sleeptime);
#endif
  Stop();
  rate = GetCycleCount() / (1e6*sleeptime);

  return rate;
}

void TimeMeasure::
PrintRealTime(std::ostream& os, const bool &verbose) const
{
  os<<"real:   "
    <<std::setw(12)
    <<GetRealTime()<<" us ";
  if (verbose) {
    os<<"\t(";
    PrintMsecHumanReadable( long(GetRealTime() /1000.0 ), os);
    os<<")";
  }
  os<<std::endl;
}

void TimeMeasure::
PrintUserTime(std::ostream& os,  const bool & verbose) const
{
  os << "user:   "
    <<std::setw(12)
    <<GetUserTime()<<" ms ";
  if (verbose) {
    //os<<"         "; // align with printrealtime
    os<<"\t(";      
    PrintMsecHumanReadable( long(GetUserTime()), os);
    os<<")";
  }
  os<<std::endl;
}

void TimeMeasure::
PrintKernelTime(std::ostream& os,  const bool & verbose) const
{
  os << "kernel: "
    <<std::setw(12)
    <<GetKernelTime()<<" ms ";
  if (verbose) {
    os<<"\t(";      
    PrintMsecHumanReadable( long(GetKernelTime()), os);
    os<<")";
  }
  os<<std::endl;
}

void TimeMeasure::
PrintIdleTime(std::ostream& os,  const bool & verbose) const
{
  os << "idle:   "
    <<std::setw(12)
    <<GetIdleTime()<<" ms ";
  if (verbose) {
    os<<"\t(";      
    PrintMsecHumanReadable( long(GetIdleTime()), os);
    os<<")";
  }
  os<<std::endl;
}

void TimeMeasure::
PrintCycleCount(std::ostream& os) const
{
  os << "cycles: "
    <<std::setw(14)
    <<GetCycleCount()<<std::endl;
}

void TimeMeasure::
CloseTimerFile()
{
  TimerStream_.close();
}


void TimeMeasure::
SetTimerFile(const std::string &file)
{
  //   if (TimerStream_) TimerStream_.close();
  TimerStream_.open(file.c_str(), std::ios_base::app);
  if (!TimerStream_) {
    BIASERR("Can not open TimerFile "<<file
      <<" in TimeMeasure::SetTimerFile()")
      perror(file.c_str());
    BIASABORT;
  }
}

void TimeMeasure::
LogToFileAndReset(const std::string &comment)
{
  Stop();
  if (!TimerStream_) BIASERR("TimerStream_ not open");
  TimerStream_<<"real: "<<std::setw(10)<<GetRealTime()/1e3
    <<"\tuser: "<<std::setw(6)<<GetUserTime()
    <<"\t"<<comment<<std::endl;
  Reset();
}


double TimeMeasure::
GetRealTime() const
{
  return (double)(rtRealSum.tv_usec + (rtRealSum.tv_sec*1E6));
}

double TimeMeasure::
GetUserTime() const
{
  //cerr << "sysconf(_SC_CLK_TCK): " << sysconf(_SC_CLK_TCK) << endl;
  // see ` man 2 times`
#ifdef WIN32
  return (double)user/(double)CLOCKS_PER_SEC * 1e3;
#else // WIN32
  return (double)user/(double)sysconf(_SC_CLK_TCK) * 1e3;
#endif // WIN32
}

double TimeMeasure::GetKernelTime() const
{
#ifdef WIN32
  return (double)kernel/(double)CLOCKS_PER_SEC * 1e3;
#else // WIN32
  return 0.0;
#endif // WIN32
}

double TimeMeasure::GetIdleTime() const
{
#ifdef WIN32
  return (double)idle/(double)CLOCKS_PER_SEC * 1e3;
#else // WIN32
  return 0.0;
#endif // WIN32
}

double TimeMeasure::GetFreq() const
{
  return 1E6/GetRealTime();
}

double TimeMeasure::GetUserFreq() const
{
  return 1E3/GetUserTime();
}

bool TimeMeasure::IsRunning() const
{
  return running;
}


double TimeMeasure::GetCycleCount() const
{
#ifdef BIAS_DEBUG
  if(running) { 
    BIASERR("you called GetCycleCount() while timer is running"); 
  }
#endif
  return _dSum;
}


void TimeMeasure::Reset()
{
  if (running){
    BIASERR("cannot reset running TimeMeasure");
  } else {
#ifdef WIN32
    _tscStart = _tscEnd = _tscTicks = 0;
#else // WIN32
#  ifdef __i386__
    _uiHiStart=_uiLoStart=_uiHiStop=_uiLoStop=0;
#  endif // __i386__
#endif // WIN32

    _dSum=0.0;
    timerclear(&rtRealSum);
    timerclear(&rtKernelSum);
    user = 0;
  }
}


void TimeMeasure::Print(std::ostream& os) const
{
  PrintRealTime(os);
  PrintUserTime(os);
#ifdef WIN32
  PrintKernelTime(os);
  PrintIdleTime(os);
#endif
  PrintCycleCount(os);
}



