/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#ifndef __RParametrization_hh__
#define __RParametrization_hh__

#include <bias_config.h>

#include <Base/Math/Vector.hh>
#include <Base/Math/Matrix.hh>
#include <Base/Geometry/Quaternion.hh>
#include <Geometry/RMatrix.hh>

#include <iostream>

namespace BIAS {

  //template <class T> class BIASGeometryBase_EXPORT Quaternion;
  //class BIASGeometryBase_EXPORT RMatrixBase;

#define Q_TYPE double

  /** @class RParametrization
      @brief Slim class bundeling rotation matrix parametrization and 
      associated covariance matrix.

      The rotation is parametrized as a quaternion

      \author woelk 11/2006 */
  class BIASGeometryBase_EXPORT RParametrization 
  {
  public:
    RParametrization(); 

    RParametrization(const Vector<Q_TYPE>& Q,
                     const Matrix<Q_TYPE>& covariance); 

    /** Set this from R parametrization. Zeros associated
        covarinace matrix */
    void Set(const Vector<Q_TYPE>& Qvec);

    /** Set this from orientation and covariance */
    inline void Set(const Vector<Q_TYPE>& Qvec, const Matrix<Q_TYPE>& Cov)
    { Set(Qvec); SetCovarianceMatrix(Cov); }
      
    /** Sets covariance matrix from Cov. Cov must be a 4x4 symmetric positiv 
        definit matrix (i.e. a covariance matrix) */
    void SetCovarianceMatrix(const Matrix<Q_TYPE>& Cov);
      
    /** returns a vector of dimension 6 where first 2 entries are
        the epipole and the last 4 entries are the quaternion orientation */
    inline const Vector<Q_TYPE>& GetVec() const
    { return Q_; }
    
    /** returns the covariance matrix associated with the EQ vector */
    inline const Matrix<Q_TYPE>& GetCovarianceMatrix() const 
    { return Cov_; }

    /** returns the rotation matrix computed from parametrization */
    RMatrixBase GetRotationMatrix() const;

    /** returns the quaternion */
    Quaternion<Q_TYPE> GetQ() const;
      
    bool Load(const std::string& file);
      
    bool Save(const std::string& file) const;

  protected:
    Vector<Q_TYPE> Q_;
    Matrix<Q_TYPE> Cov_;

  };

  BIASGeometryBase_EXPORT std::ostream& operator<<(std::ostream &os, const RParametrization& p);

  BIASGeometryBase_EXPORT std::istream& operator>>(std::istream &is, RParametrization& p);
    
}


#endif
