#include <iostream>
#include <sstream>

#include <Base/Image/Image.hh>
#include <Base/Image/ImageIO.hh>
#include <Base/Image/ImageDraw.hh>

#include <Base/Geometry/HomgPoint2D.hh>
#include <Base/Geometry/HomgPoint3D.hh>
#include <Geometry/RMatrix.hh>

using namespace BIAS;
using namespace std;


/**
 * Simple example that draws
 * - a rotating line (like a clock hand)
 * - a moving line that moves one pixel
 * - a static line with yEnd - yStart = 1, i.e. with a very low slope
 *
 * A set of images is produced (according to var steps) that you can view in
 * biasviewwx as an animation.
 *
 * Not very efficient...
 */
int main(int argc, char** argv) {
  
  // leading zeroes in filenames will only be set correctly if this is below 999
  unsigned int steps = 90;
  int antialiased = 1; // 0 = no AA, 1 = AA
  
  Image<unsigned char> image(400, 400, 3); // min size: 200x200
  
  unsigned char bgCol[] = {127, 0, 127};
  
  std::vector<unsigned char> red, green, blue, white;
  red.push_back(255);
  red.push_back(0);
  red.push_back(0);
  green.push_back(0);
  green.push_back(255);
  green.push_back(0);
  blue.push_back(0);
  blue.push_back(0);
  blue.push_back(255);
  white.push_back(255);
  white.push_back(255);
  white.push_back(255);
  
  HomgPoint2D clockCenter(image.GetWidth() / 2, image.GetHeight() / 2);
  HomgPoint2D clockOuter(image.GetWidth() / 2, image.GetHeight() / 4);
  HomgPoint2D clockOuterRunner(clockOuter);
  HomgPoint2D moveStart(10, image.GetHeight() / 4);
  HomgPoint2D moveEnd(110, image.GetHeight() / 4);
  HomgPoint2D moveStartRunner(moveStart);
  HomgPoint2D moveEndRunner(moveEnd);
  HomgPoint2D staticStart(10, image.GetHeight() / 8);
  HomgPoint2D staticEnd(image.GetWidth() - 10, image.GetHeight() / 8 + 2);
  
  float angle = 2.0f * float(M_PI) / float(steps); // angle per iteration
  
  ostringstream filename;
  
  for (unsigned int i = 0; i < steps; i++) {
    cout << endl << endl << "step: " << (i + 1) << "/" << steps << endl;
    
    // prepare image and filename
    image.FillImageWithConstValue(bgCol);
    filename.str("");
    filename << "clock_";
    if (i < 100) {
      filename << 0;
    }
    if (i < 10) {
      filename << 0;
    }
    filename << i << ".mip";
    
    /////
    // ROTATING LINE (clock)
    /////
    
    // calc new postion of clock point:
    // - reset
    clockOuterRunner[0] = clockOuter[0];
    clockOuterRunner[1] = clockOuter[1];
    
    // - translate clockOuterRunner so that clockCenter is in coord center
    clockOuterRunner[0] -= clockCenter[0];
    clockOuterRunner[1] -= clockCenter[1];
    
    // - rotation
    RMatrix R;
    Vector3<ROTATION_MATRIX_TYPE> p3D1(clockOuterRunner[0], clockOuterRunner[1], 0);
    Vector3<ROTATION_MATRIX_TYPE> p3D2;
    R.SetXYZ(0.0, 0.0, i * angle);
    R.Mult(p3D1, p3D2);
    clockOuterRunner[0] = p3D2[0];
    clockOuterRunner[1] = p3D2[1];
    
    // - undo translation
    clockOuterRunner[0] += clockCenter[0];
    clockOuterRunner[1] += clockCenter[1];
    
    // draw line
    ImageDraw<unsigned char>::Line(image,
                                   clockCenter, clockOuterRunner,
                                   red,
                                   antialiased);
    
    /////
    // MOVING LINE
    /////
    
    moveStartRunner[0] = moveStart[0];
    moveStartRunner[0] += (HOMGPOINT2D_TYPE)i / (HOMGPOINT2D_TYPE)steps;
    moveStartRunner[1] = moveStart[1] + i / steps;
    moveStartRunner[1] += (HOMGPOINT2D_TYPE)i / (HOMGPOINT2D_TYPE)steps;
    moveEndRunner[0] = moveEnd[0] + i / steps;
    moveEndRunner[0] += (HOMGPOINT2D_TYPE)i / (HOMGPOINT2D_TYPE)steps;
    moveEndRunner[1] = moveEnd[1] + i / steps;
    moveEndRunner[1] += (HOMGPOINT2D_TYPE)i / (HOMGPOINT2D_TYPE)steps;
    
    ImageDraw<unsigned char>::Line(image,
                                   moveStartRunner, moveEndRunner,
                                   blue,
                                   antialiased);
    
    /////
    // STATIC LINE
    /////
    
    ImageDraw<unsigned char>::Line(image,
                                   staticStart, staticEnd,
                                   green,
                                   antialiased);
    
    // save image
    ImageIO::Save(filename.str(), image);
  }
  
  return 0;
}
