/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
Multimediale Systeme der Informationsverarbeitung
Institut fuer Informatik
Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include <Base/Common/W32Compat.hh>
#include "ImageDraw.hh"
#include <math.h>
#include "Bresenham.hh"
#include "BresenhamCircle.hh"
#include "ColourRGB.hh"

#ifdef BIAS_HAVE_OPENCV
#  include <Base/Image/WrapBias2Ipl.hh>
#endif

using namespace BIAS;
using namespace std;

namespace BIAS {

template <class StorageType>
StorageType ImageDraw<StorageType>::GetContrastValue(StorageType t)
{
    BIASERR("Dont know what good contrast for "<<t<<" would be for StorageType "
        <<PRINTTYPE(StorageType)); 
    return (StorageType)0;
}

template <>
unsigned char ImageDraw<unsigned char>::GetContrastValue(unsigned char t)
{
    return ((int)t>128)?(unsigned char)0:(unsigned char)255;
}

template <>
float ImageDraw<float>::GetContrastValue(float t)
{
    return float((t>0.0)?-1.0:1.0);
}


/////////
// rectangles

template <class StorageType>
int ImageDraw<StorageType>::RectangleCorners(Image<StorageType>& im, 
                                             unsigned int minx, 
                                             unsigned int miny,
                                             unsigned int maxx, 
                                             unsigned int maxy,
                                             const StorageType Value[])
{
    BIASASSERT(minx<im.GetWidth());
    BIASASSERT(miny<im.GetHeight());
    BIASASSERT(maxx<im.GetWidth());
    BIASASSERT(maxy<im.GetHeight());

    register StorageType **ida = im.GetImageDataArray();
    register unsigned int xp1, xp2;
    for (register unsigned int x = minx; x <= maxx; x++){
        xp1 = x * im.GetChannelCount();
        for (register unsigned int c=0; c< im.GetChannelCount(); c++)
            ida[miny][xp1+c]=ida[maxy][xp1+c]=Value[c];
    }
    // left and right lines
    xp1=minx*im.GetChannelCount();
    xp2=maxx*im.GetChannelCount();
    for (register unsigned int y = miny; y <= maxy; y++){
        for (register unsigned int c=0; c< im.GetChannelCount(); c++)
            ida[y][xp1+c]=ida[y][xp2+c]=Value[c];
    }
    return 0;
}


template <class StorageType>
int ImageDraw<StorageType>::RectangleCorners(Image<StorageType>& im, 
                                             unsigned int minx, 
                                             unsigned int miny,
                                             unsigned int maxx, 
                                             unsigned int maxy,
                                             const StorageType value)
{
    StorageType *val = new StorageType[im.GetChannelCount()];
    for (register unsigned int i=0; i<im.GetChannelCount(); i++)
        val[i]=value;
    return RectangleCorners(im, minx, miny, maxx, maxy, val);
}


template <class StorageType>
int ImageDraw<StorageType>::RectangleCorners(Image<StorageType>& im, 
                                             unsigned int minx, 
                                             unsigned int miny,
                                             unsigned int maxx, 
                                             unsigned int maxy)
{
    register StorageType **ida = im.GetImageDataArray();
    register unsigned int xp1, xp2;
    for (register unsigned int x = minx; x <= maxx; x++){
        xp1 = x * im.GetChannelCount();
        for (register unsigned int c=0; c< im.GetChannelCount(); c++){
            ida[miny][xp1+c]=GetContrastValue(ida[miny][xp1+c]);
            ida[maxy][xp1+c]=GetContrastValue(ida[maxy][xp1+c]);
        }
    }
    // left and right lines
    xp1=minx*im.GetChannelCount();
    xp2=maxx*im.GetChannelCount();
    for (register unsigned int y = miny+1; y < maxy; y++){
        for (register unsigned int c=0; c< im.GetChannelCount(); c++){
            ida[y][xp1+c]=GetContrastValue(ida[y][xp1+c]);
            ida[y][xp2+c]=GetContrastValue(ida[y][xp2+c]);
        }
    }
    return 0;
}

template <class StorageType>
int ImageDraw<StorageType>::RectangleCornersGrey(Image<StorageType>& im,
                                                 unsigned int minx, 
                                                 unsigned int miny,
                                                 unsigned int maxx, 
                                                 unsigned int maxy,
                                                 StorageType GreyValue)
{
    register StorageType *ImageDataPointer, *ImageDataPointer2; 
    register unsigned int Width = im.GetWidth();
    // horizontal lines lines
    ImageDataPointer = im.GetImageData() + 
        im.GetPixelPosition(minx,miny);
    ImageDataPointer2 = im.GetImageData() + 
        im.GetPixelPosition(minx,maxy);
    for (register unsigned int i = 0; i <= maxx - minx; i++){
        *ImageDataPointer = GreyValue;
        *ImageDataPointer2 = GreyValue;
        ImageDataPointer++;
        ImageDataPointer2++;
    }
    // left and right lines
    ImageDataPointer = im.GetImageData() + 
        im.GetPixelPosition(minx,miny+1);
    ImageDataPointer2 = im.GetImageData() + 
        im.GetPixelPosition(maxx,miny+1);
    // top and bottom lines
    const unsigned int rowsminusone = maxy - miny ;
    for (register unsigned int i = 1; i < rowsminusone; i++){
        *ImageDataPointer = GreyValue;
        *ImageDataPointer2 = GreyValue;
        ImageDataPointer += Width;
        ImageDataPointer2 += Width;
    }
    return 0;
}


template <class StorageType>
int ImageDraw<StorageType>::RectangleCornersGreyFill(Image<StorageType>& im,
                                                     unsigned int minx, 
                                                     unsigned int miny,
                                                     unsigned int maxx, 
                                                     unsigned int maxy,
                                                     StorageType GreyValue)
{
    register StorageType *ImageDataPointer;

    for (unsigned int row=miny;row<=maxy;row++) {
        ImageDataPointer = im.GetImageData() + 
            im.GetPixelPosition(minx,row);
        for (register unsigned int i = 0; i <= maxx - minx; i++){
            *ImageDataPointer = GreyValue;
            ImageDataPointer++;
        }
    }
    return 0;
}


template <class StorageType>
int ImageDraw<StorageType>::RectangleCenter(Image<StorageType>& im, 
                                            unsigned int X, unsigned int Y, 
                                            unsigned int Size, 
                                            const StorageType Value[])
{
#ifdef BIAS_DEBUG
    if (Size > (unsigned int)im.GetWidth() || 
        Size > (unsigned int)im.GetHeight())
        BIASERR("wrong Size "<<Size);
#endif
    unsigned int Halfsize = Size >> 1;
    if (im.GetChannelCount()==1){
        return RectangleCornersGrey(im, (X>Halfsize)?X-Halfsize:0, 
            (Y>Halfsize)?Y-Halfsize:0, 
            (X<im.GetWidth()-Halfsize)?
            (X+Halfsize):(im.GetWidth()-1), 
            (Y<im.GetHeight()-Halfsize)?
            (Y+Halfsize):(im.GetHeight()-1),
            Value[0]); 
    } else {
        return RectangleCorners(im, (X>Halfsize)?X-Halfsize:0, 
            (Y>Halfsize)?Y-Halfsize:0, 
            (X<im.GetWidth()-Halfsize)?
            (X+Halfsize):(im.GetWidth()-1), 
            (Y<im.GetHeight()-Halfsize)?
            (Y+Halfsize):(im.GetHeight()-1),
            Value);
    }
}


template <class StorageType>
int ImageDraw<StorageType>::RectangleCenterGrey(Image<StorageType>& im, 
                                                unsigned int X, unsigned int Y, 
                                                unsigned int Size, 
                                                const StorageType Value)
{
#ifdef BIAS_DEBUG
    if (Size > (unsigned int)im.GetWidth() || 
        Size > (unsigned int)im.GetHeight())
        BIASERR("wrong Size "<<Size);
#endif
    unsigned int Halfsize = Size >> 1;
    return RectangleCornersGrey(im, (X>Halfsize)?X-Halfsize:0, 
        (Y>Halfsize)?Y-Halfsize:0, 
        (X<im.GetWidth()-Halfsize)?
        (X+Halfsize):(im.GetWidth()-1), 
        (Y<im.GetHeight()-Halfsize)?
        (Y+Halfsize):(im.GetHeight()-1),
        Value); 
}

template <class StorageType>
int ImageDraw<StorageType>::RectangleCenterGreyFill(Image<StorageType>& im, 
                                                    unsigned int X, 
                                                    unsigned int Y, 
                                                    unsigned int Size, 
                                                    StorageType Value)
{
#ifdef BIAS_DEBUG
    if (Size > (unsigned int)im.GetWidth() || 
        Size > (unsigned int)im.GetHeight())
        BIASERR("wrong Size "<<Size);
#endif
    unsigned int Halfsize = Size >> 1;
    return RectangleCornersGreyFill(im, (X>Halfsize)?X-Halfsize:0, 
        (Y>Halfsize)?Y-Halfsize:0, 
        (X<im.GetWidth()-Halfsize)?
        (X+Halfsize):(im.GetWidth()-1), 
        (Y<im.GetHeight()-Halfsize)?
        (Y+Halfsize):(im.GetHeight()-1),
        Value); 
}



template <class StorageType>
int ImageDraw<StorageType>::RectangleCenter(Image<StorageType>& im, unsigned int x, 
                                            unsigned int y, unsigned int size)
{
    unsigned int Halfsize = size >> 1;
    return RectangleCorners(im, (x>Halfsize)?x-Halfsize:0, 
        (y>Halfsize)?y-Halfsize:0, 
        (x<im.GetWidth()-Halfsize)?(x+Halfsize):(im.GetWidth()-1),
        (y<im.GetHeight()-Halfsize)?
        (y+Halfsize):(im.GetHeight()-1));
}

template <class StorageType>
int ImageDraw<StorageType>::RectangleCenter(Image<StorageType>& im, 
                                            unsigned int coo[2], 
                                            unsigned int size)
{ return RectangleCenter(im, coo[0], coo[1], size); }

/////////////////////////////////////////////////////////////////////////
// lines 
/////////////////////////////////////////////////////////////////////////

template <class StorageType>
int ImageDraw<StorageType>::Line(Image<StorageType>& im,
                                 const int start[2],
                                 const int end[2], 
                                 const StorageType value[])
{
    BIASASSERT(start[0]>=0);
    BIASASSERT(start[1]>=0);
    BIASASSERT(end[0]>=0);
    BIASASSERT(end[1]>=0);

    unsigned int us[2], ue[2];
    us[0] = (unsigned)start[0];
    us[1] = (unsigned)start[1];
    ue[0] = (unsigned)end[0];
    ue[1] = (unsigned)end[1];

    return Line(im, us, ue, value);
}

template <class StorageType>
int ImageDraw<StorageType>::Line(Image<StorageType>& im,
                                 const unsigned int start[2],
                                 const unsigned int end[2], 
                                 const StorageType value[])
{
  if (start[0]>=im.GetWidth() || start[0]<0 ||
      start[1]>=im.GetHeight() || start[1]<0 ||
      end[0]>=im.GetWidth() || end[0]<0 ||
      end[1]>=im.GetHeight() || end[1]<0) {
#ifdef BIAS_DEBUG   
        BIASERR("start- or end-point of line outside of "<<im.GetWidth()<<"x"
            <<im.GetHeight()<<" image: ("<<start[0]<<", "
            <<start[1]<<") -> ("<<end[0]<<", "<<end[1]<<") "
            <<" skipping drawing Line.");
#endif
    return -1;
  }

  unsigned int next[2];
  Bresenham bres(start, end);

  next[0]=start[0];
  next[1]=start[1];

  StorageType **ida=im.GetImageDataArray();
  do {
    if (next[0]<0 || next[0]>=im.GetWidth() ||
        next[1]<0 || next[1]>=im.GetHeight())
      break;
    for (unsigned int c=0; c<im.GetChannelCount(); c++)
      ida[next[1]][next[0]*im.GetChannelCount()+c]=value[c];
  } while (bres.GetNext(next));
  return 0;
}

template <class StorageType>
int ImageDraw<StorageType>::Line(Image<StorageType>& im,
                                 const unsigned int start[2],
                                 const unsigned int end[2], StorageType value)
{
#ifdef BIAS_DEBUG
  if (start[0]<0 || start[0]>=im.GetWidth() ||
      start[1]<0 || start[1]>=im.GetHeight() ||
      end[0]<0 || end[0]>=im.GetWidth() ||
      end[1]<0 || end[1]>=im.GetHeight()){
    BIASERR("coordinates out of bounds "<<start[0]<<" "<<start[1]<<" "
            <<end[0]<<" "<<end[1]);
    BIASABORT;
  }
#endif
  unsigned int next[2];
  Bresenham bres(start, end);

  next[0]=start[0];
  next[1]=start[1];

  StorageType **ida=im.GetImageDataArray();
  do {
    for (unsigned int c=0; c<im.GetChannelCount(); c++)
      ida[next[1]][next[0]*im.GetChannelCount()+c]=value;
  } while (bres.GetNext(next));
  return 0;
}

template <class StorageType>
int ImageDraw<StorageType>::LineGrey(Image<StorageType>& im,
                                     const unsigned int start[2],
                                     const unsigned int end[2], 
                                     StorageType value)
{
#ifdef BIAS_DEBUG
  if (start[0]<0 || start[0]>=im.GetWidth() ||
      start[1]<0 || start[1]>=im.GetHeight() ||
      end[0]<0 || end[0]>=im.GetWidth() ||
      end[1]<0 || end[1]>=im.GetHeight()){
    BIASERR("coordinates out of bounds");
    BIASABORT;
  }
#endif
  unsigned int next[2];
  Bresenham bres(start, end);

  next[0]=start[0];
  next[1]=start[1];

  StorageType **ida=im.GetImageDataArray();
  do {
    ida[next[1]][next[0]]=value;
  } while (bres.GetNext(next));
  return 0;
}


template <class StorageType>
int ImageDraw<StorageType>::Line(Image<StorageType>& im,
                                 const unsigned int start[2],
                                 const unsigned int end[2])
{
#ifdef BIAS_DEBUG
  if (start[0]<0 || start[0]>=im.GetWidth() ||
      start[1]<0 || start[1]>=im.GetHeight() ||
      end[0]<0 || end[0]>=im.GetWidth() ||
      end[1]<0 || end[1]>=im.GetHeight()){
    BIASERR("coordinates out of bounds: "<<start[0]<<";"<<start[1]
            <<" to "<<end[0]<<";"<<end[1]);
    BIASABORT;
  }
#endif
  unsigned int next[2];
  Bresenham bres(start, end);

  next[0]=start[0];
  next[1]=start[1];

  StorageType **ida=im.GetImageDataArray();
  do {
    for (unsigned int c=0; c<im.GetChannelCount(); c++)
      ida[next[1]][next[0]*im.GetChannelCount()+c]=
        GetContrastValue(ida[next[1]][next[0]*im.GetChannelCount()+c]);
  } while (bres.GetNext(next));
  return 0;
}


template <class StorageType>
int ImageDraw<StorageType>::LineGrey(Image<StorageType>& im,
                                     const unsigned int start[2],
                                     const unsigned int end[2])
{
#ifdef BIAS_DEBUG
  if (start[0]<0 || start[0]>=im.GetWidth() ||
      start[1]<0 || start[1]>=im.GetHeight() ||
      end[0]<0 || end[0]>=im.GetWidth() ||
      end[1]<0 || end[1]>=im.GetHeight()){
    BIASERR("coordinates out of bounds");
    BIASABORT;
  }
  if (im.GetChannelCount() !=1){
    BIASERR("only for one channel images");
    BIASABORT;
  }
#endif
  unsigned int next[2];
  Bresenham bres(start, end);

  next[0]=start[0];
  next[1]=start[1];

  StorageType **ida=im.GetImageDataArray();
  do {
    ida[next[1]][next[0]]=GetContrastValue(ida[next[1]][next[0]]);
  } while (bres.GetNext(next));
  return 0;
}



template <class StorageType>
int ImageDraw<StorageType>::Line(Image<StorageType>& im, 
                                 const unsigned int StartX, 
                                 const unsigned int StartY, 
                                 const unsigned int EndX, 
                                 const unsigned int EndY, 
                                 const StorageType Value[])
{
    unsigned int start[]={StartX, StartY}, end[]={EndX, EndY};
    return Line(im, start, end, Value); 
}


template <class StorageType>
int ImageDraw<StorageType>::Line(Image<StorageType>& im, 
                                 const unsigned int StartX, 
                                 const unsigned int StartY, 
                                 const unsigned int EndX, 
                                 const unsigned int EndY, 
                                 const StorageType Value)
{
    unsigned int start[]={StartX, StartY}, end[]={EndX, EndY};
    return Line(im, start, end, Value); 
}

template <class StorageType>
int ImageDraw<StorageType>::Line(Image<StorageType>& im, 
                                 const unsigned int StartX, 
                                 const unsigned int StartY, 
                                 const unsigned int EndX, 
                                 const unsigned int EndY)
{
    unsigned int start[]={StartX, StartY}, end[]={EndX, EndY};
    return Line(im, start, end); 
}



template <class StorageType>
int ImageDraw<StorageType>::LineGrey(Image<StorageType>& im,
                                     const unsigned int StartX, 
                                     const unsigned int StartY,
                                     const unsigned int EndX, 
                                     const unsigned int EndY, 
                                     const StorageType Value)
{
    unsigned int start[]={StartX, StartY}, end[]={EndX, EndY};
    return LineGrey(im, start, end, Value); 
}

// template <class StorageType>
// int ImageDraw<StorageType>::LineGrey(Image<StorageType>& im,
//                                      const unsigned int StartX, 
//                                      const unsigned int StartY,
//                                      const unsigned int EndX, 
//                                      const unsigned int EndY, 
//                                      StorageType Value)
// {
// #ifdef BIAS_DEBUG
//   if (StartX < 0 || StartX > im.GetWidth()-1 ||
//       StartY < 0 || StartY > im.GetHeight()-1 ||
//       EndX < 0 || EndX > im.GetWidth()-1 ||
//       EndY < 0 || EndY >im.GetHeight()-1)
//     BIASERR("wrong coordinates");
// #endif

//   StorageType **ida=im.GetImageDataArray();
//   int DrawX, DrawY, StX=StartX, StY=StartY, EnX=EndX, EnY=EndY;
//   int OffsetX, OffsetY;
//   bool _bxinv, _byinv, _over_angle;
//   // seeding point, rise, ...
//   int X = 0, Y = 0;
//   int dY,dX, dYmX;
//   int E;
//   int tmp;

//   // shift seeding point to origin and transform
//   // line to first quadrant
//   //
//   OffsetX=StX;
//   OffsetY=StY;
//   StX=0; StY=0;
//   EnX=EnX-OffsetX;
//   EnY=EnY-OffsetY;

//   // transform line to first quadrant
//   if (EnX < 0) { EnX = -EnX; _bxinv = true; }
//   else _bxinv = false;

//   if (EnY < 0) { EnY = -EnY; _byinv = true; }
//   else _byinv = false;

//   // rise less equal 1
//   if (EnX < EnY) {
//     tmp = EnX; EnX = EnY; EnY = tmp;
//     _over_angle = true;
//   }
//   else _over_angle = false;

//   // calculate rise
//   dX = EnX-StX;
//   dY = EnY-StY;
//   dYmX = dY-dX;

//   E = dYmX+dY;

//   // compute Bresenham points
//   for (int i=0; i<dX; i++) {
//     if (_over_angle) { DrawX = Y; DrawY = X; } 
//     else { DrawX = X; DrawY = Y; }
//     if (_bxinv) DrawX = -DrawX;
//     if (_byinv) DrawY = -DrawY;
//     DrawY+=OffsetY;
//     DrawX+=OffsetX;
//     // draw point
//     ida[DrawY][DrawX]=Value;
//     // image->PutPixel(&GreyValue,XCoord(X,Y),YCoord(X,Y));
//     X++;

//     // calc next y-coordinate
//     if (E > 0) { Y++; E += 2*(dYmX); }
//     else  E += 2*dY;
//   }
//   if (_over_angle) { DrawX = Y; DrawY = X; }
//   else { DrawX = X; DrawY = Y; }
//   if (_bxinv) DrawX = -DrawX;
//   if (_byinv) DrawY = -DrawY;
//   DrawY+=OffsetY;
//   DrawX+=OffsetX;
//   ida[DrawY][DrawX]=Value;
//   // image->PutPixel(&GreyValue,XCoord(X,Y),YCoord(X,Y));
//   return 0;
// }

#define BIAS_DRAW_CROSS_PRODUCT(a, b, res)\
    res[0] = a[1] * b[2] - a[2] * b[1];\
    res[1] = a[2] * b[0] - a[0] * b[2];\
    res[2] = a[0] * b[1] - a[1] * b[0];

#define BIAS_DRAW_PRECISION 1e-8

template <class StorageType>
int ImageDraw<StorageType>::
Line(Image<StorageType>& im, const double start[2], 
     const double end[2], 
     const StorageType value[])
{
    int w=im.GetWidth();
    int h=im.GetHeight();
    double wm=(double)(w-1);
    double hm=(double)(h-1);
    unsigned uistart[2], uiend[2];
    double s[]={start[0], start[1], 1.0}, e[]={end[0], end[1], 1.0};

    // the line is build by the cross product of the homogenous 
    // start and end points
    double line[3];
    BIAS_DRAW_CROSS_PRODUCT(s, e, line);

    // now intersect the line with the image border lines
    // again by using the cross product
    double top[]={0, -1.0, 0.0}, bottom[]={0.0, -1.0, hm};
    double left[]={-1.0, 0.0, 0.0}, right[]={-1.0, 0.0, wm};
    double intersections[4][3]; // 0=left, 1=right, 2=top, 3=bottom
    BIAS_DRAW_CROSS_PRODUCT(line, left,   intersections[0]);
    BIAS_DRAW_CROSS_PRODUCT(line, right,  intersections[1]);
    BIAS_DRAW_CROSS_PRODUCT(line, top,    intersections[2]);
    BIAS_DRAW_CROSS_PRODUCT(line, bottom, intersections[3]);

    unsigned coo[4]={0,0,0,0};
    bool intersect[4];
    int intersectioncount=0;
    // deal with compiler precision
    for (unsigned int i=0; i<4; i++){
        if (fabs(intersections[i][0]) < BIAS_DRAW_PRECISION)
            intersections[i][0] = 0.0;
        if (fabs(intersections[i][1]) < BIAS_DRAW_PRECISION)
            intersections[i][1] = 0.0;
        if (fabs(intersections[i][0]-wm) < BIAS_DRAW_PRECISION)
            intersections[i][0] = wm;
        if (fabs(intersections[i][1]-hm) < BIAS_DRAW_PRECISION)
            intersections[i][1] = hm;
        if (fabs(intersections[i][2]) < BIAS_DRAW_PRECISION){
            intersections[i][2] = 0.0;
            intersect[i]=false;
        } else { // homogenize
            intersections[i][0]/=intersections[i][2];
            intersections[i][1]/=intersections[i][2];
            intersect[i]=true;
        }
    }

    // extract the intersection points into coo[]
    // left
    if (intersect[0] && intersections[0][1]>=0 && intersections[0][1]<=hm){
        intersectioncount++;
        coo[0] = 0;
        coo[1] = (unsigned int)rint(intersections[0][1]);
    } 
    // right
    if (intersect[1] && intersections[1][1]>=0 && intersections[1][1]<=hm){
        if (intersectioncount==0){
            coo[0] = w- 1;
            coo[1] = (unsigned int)rint(intersections[1][1]);
        } else {
            coo[2] = w - 1;
            coo[3] = (unsigned int)rint(intersections[1][1]);
        }
        intersectioncount++;
    } 
    // top
    if (intersect[2] && intersections[2][0]>=0 && intersections[2][0]<=wm){
        switch (intersectioncount){
    case 0:
        coo[0] = (unsigned int)rint(intersections[2][0]);
        coo[1] = 0; 
        break;
    case 1:
        coo[2] = (unsigned int)rint(intersections[2][0]);
        coo[3] = 0; 
        break;
    default:
        // do nothing, all intersections correctly calulated
        // because left and right intersection can not have a common point
        // cout << "found intersections "<<coo[0]<<", "<<coo[1]<<"  and  "<<coo[2]<<", "<<coo[3]<<"\nnew intersection "<<(unsigned int)rint(intersections[2][0])<<", 0\n";
        break;
        }
        intersectioncount++;
    } 
    // bottom
    if (intersect[3] && intersections[3][0]>=0 && intersections[3][0]<=wm){
        switch (intersectioncount){
    case 0:
        coo[0] = (unsigned int)rint(intersections[3][0]);
        coo[1] = h-1; 
        break;
    case 1:
        coo[2] = (unsigned int)rint(intersections[3][0]);
        coo[3] = h-1; 
        break;
    case 2:
        // this can only happen if upper left or upper right corner of image
        // is intersection
        // It follows that the top corner is always stored in coo[2/3], because
        // interscetioncount was 1 when checking for interscetions with top edge
        coo[0] = (unsigned int)rint(intersections[3][0]);
        coo[1] = h-1;
        break;
    default:
        // do nothing, all intersections correctly calulated
        // cout << "found intersections "<<coo[0]<<", "<<coo[1]<<"  and  "<<coo[2]<<", "<<coo[3]<<"\nnew intersection "<<(unsigned int)rint(intersections[2][0])<<", 0\n";
        break;
        }
        intersectioncount++;
    } 

    if (intersectioncount>1){

        // now draw the line
        if (start[0]>=0.0 && start[0]<=wm && start[1]>=0.0 && start[1]<=hm){
            // easy, start[0] in image and start[1] in image
            uistart[0]=(unsigned)rint(start[0]);
            uistart[1]=(unsigned)rint(start[1]);
        } else {
            // replace start point

            // take coo[0/1] if (end-coo[0/1]).ScalarProduct(end-start)==1.0
            double dex = end[0]-coo[0], dey = end[1]-coo[1];
            double dx = end[0]-start[0], dy = end[1]-start[1];
            if (dx*dex+dy*dey>0.0){ // angle < 90 degree take coo[0/1]
                uistart[0]=coo[0];
                uistart[1]=coo[1];
            } else {
                uistart[0]=coo[2];
                uistart[1]=coo[3];
            }

        }
        if (start[0]>=0.0 && start[0]<=wm && start[1]>=0.0 && start[1]<=hm){
            // easy, end[0] in image and end[1] in image
            uiend[0]=(unsigned)rint(end[0]);
            uiend[1]=(unsigned)rint(end[1]);
        } else {
            // replace end point

            // take coo[0/1] if (start-coo[0/1]).ScalarProduct(start-end)==1.0
            double dex = start[0]-coo[0], dey = start[1]-coo[1];
            double dx = start[0]-end[0], dy = start[1]-end[1];
            if (dx*dex+dy*dey>0.0){ // angle < 90 degree take coo[0/1]
                uistart[0]=coo[0];
                uistart[1]=coo[1];
            } else {
                uistart[0]=coo[2];
                uistart[1]=coo[3];
            }
        }
        //cerr << "drawing between "<<uistart[0]<<", "<<uistart[1]
        //     << "  and  "<<uiend[0]<<", "<<uiend[1]<<"\n";
        return Line(im, uistart, uiend, value);
    }

    //cerr << "nothing to draw\n";
    return 0;
}

#undef BIAS_DRAW_CROSS_PRODUCT
#undef BIAS_DRAW_PRECISION

////////////////////////////////////////////////////
// interpolated lines
///////////////////////////////////////////////////////

template <class StorageType>
int ImageDraw<StorageType>::InterpolatedLine(Image<StorageType>& im, 
                                             const int StartX, 
                                             const int StartY,
                                             const int EndX, 
                                             const int EndY, 
                                             const StorageType Value[])
{
    if (im.GetChannelCount()==1){
        return InterpolatedLineGrey(im, StartX, StartY, EndX, EndY, Value[0]);
    } else {
        BIASERR("DrawInterpolatedLineColor not jet implemented");
        return -1;
    }
}

template <class StorageType>
int ImageDraw<StorageType>::InterpolatedLineGrey(Image<StorageType>& im, 
                                                 const int StartX, 
                                                 const int StartY,
                                                 const int EndX, 
                                                 const int EndY, 
                                                 const StorageType GreyValue)
{
    StorageType **ida = im.GetImageDataArray();
    EightWaySymmetry<int> sym;
    enum TLineType type;
    int TStartX, TStartY, TEndX, TEndY, CurrX=0, CurrY=0;
    double TSlope;

    type = sym.DetermineLineType(StartX, StartY, EndX, EndY);

    sym.Transform(type, StartX, StartY, TStartX, TStartY);

    sym.Transform(type, EndX, EndY, TEndX, TEndY);

    TSlope = (double)(TEndY - TStartY)/(double)(TEndX - TStartX);

    //    cerr << "TSlope: " << TSlope << endl;

    register double g, g1, g2, gm, gtmp;
    g = 0.0;
    g1 = TSlope - 1.0;
    g2 = TSlope;

    if (TSlope > 1e-15){
        while (TStartX < TEndX) {

            if (g>1.0){
                g+=g1;
                gtmp = (g<1.0)?(g):(1.0);
                TStartY++;
                gm = (g<1.0)?(1.0 - g):(0.0);
                sym.InverseTransform(type, TStartX, TStartY, CurrX, CurrY);
                //  	cerr << " ++a type: " <<type << "  x " << CurrX << "  y " << CurrY << "   g " << g << "  gtmp " << "  g1: " << g1 << "  g2: " << g2 << endl;
                ida[CurrY][CurrX] = (StorageType)rint((double)GreyValue * gm) +
                    (StorageType)rint((double)ida[CurrY][CurrX] * gtmp);
                sym.InverseTransform(type, TStartX, TStartY+1, CurrX, CurrY);
                //  	cerr << " ++b type: " <<type << "  x " << CurrX << "  y " << CurrY << "   gm " << gm << "  gtmp " << gtmp << endl;
                ida[CurrY][CurrX] = (StorageType)rint((double)GreyValue * gtmp) + 
                    (StorageType)rint((double)ida[CurrY][CurrX] * gm);
                TStartX++;
            } else {
                gm = 1.0 - g;
                sym.InverseTransform(type, TStartX, TStartY, CurrX, CurrY);
                //  	cerr << " a type: " <<type << " x " << CurrX << "  y " << CurrY << "   g " << g << endl;
                ida[CurrY][CurrX] = (StorageType)rint((double)GreyValue * gm) + 
                    (StorageType)rint((double)ida[CurrY][CurrX] * g);
                sym.InverseTransform(type, TStartX, TStartY+1, CurrX, CurrY);
                //  	cerr << " b type: " <<type <<" x " << CurrX << "  y " << CurrY << "   gm " << gm << endl;
                ida[CurrY][CurrX] = (StorageType)rint((double)GreyValue * g) + 
                    (StorageType)rint((double)ida[CurrY][CurrX] * gm);
                TStartX++;
                g+=g2;
            }
        }
    } else {
        while (TStartX < TEndX){
            //cerr << " x " << TStartX << "  y " << StartY << "   g " << g << endl;
            sym.InverseTransform(type, TStartX, TStartY, CurrX, CurrY);
            ida[CurrY][CurrX] = (StorageType)rint((double)GreyValue);
            TStartX++;
        }
    }
    return 0;
}

// Helper method
// WARNING: no checking for correct channel count!!!
//          must ensure this in calling method!
template <class StorageType>
inline void DrawPixel(Image<StorageType>& im,
                      HOMGPOINT2D_TYPE x, HOMGPOINT2D_TYPE y,
                      const std::vector<StorageType> &color,
                      HOMGPOINT2D_TYPE weight,
                      unsigned int octet,
                      HOMGPOINT2D_TYPE xStart, HOMGPOINT2D_TYPE yStart,
                      HOMGPOINT2D_TYPE xEnd, HOMGPOINT2D_TYPE yEnd) {
  
  
  // TODO: pass data array as argument
  StorageType *imData = im.GetImageData();
  StorageType colNew;
  StorageType colOld;
  
  weight = abs(weight);
  
  HOMGPOINT2D_TYPE xTransf, yTransf;
  
  // undo transformation to first octet
  switch (octet) {
    case 1:
      xTransf = x;
      yTransf = y;
      break;
    case 2:
      xTransf = y;
      yTransf = x;
      break;
    case 3:
      xTransf = -(y - yStart) + yEnd;
      yTransf = x;
      break;
    case 4:
      xTransf = -(x - xStart) + xEnd;
      yTransf = y;
      break;
    case 5:
      xTransf = x;
      yTransf = y;
      break;
    case 6:
      xTransf = y;
      yTransf = x;
      break;
    case 7:
      xTransf = y;
      yTransf = -(x - xStart) + xEnd;
      break;
    case 8:
      xTransf = x;
      yTransf = -(y - yStart) + yEnd;
      break;
    default:
      return;
  }
  
  if (!im.IsPositionInImage (int(xTransf + 0.5), int(yTransf + 0.5))) {
    return;
  }
  
  // draw pixel
  unsigned int offset = int(yTransf + 0.5) * im.GetWidth() * im.GetChannelCount()
                        + int(xTransf + 0.5) * im.GetChannelCount();
  for (unsigned int i = 0; i < color.size(); i++) {
    // mix given color with background color according to weight
    colOld = StorageType((1 - weight) * imData[offset + i]);
    colNew = StorageType(weight * color[i]);
    imData[offset + i] = colOld + colNew;
  }
}


// Helper method
template <class StorageType>
inline void DrawLineBresenham(Image<StorageType>& im,
                              HOMGPOINT2D_TYPE xStart,
                              HOMGPOINT2D_TYPE yStart,
                              HOMGPOINT2D_TYPE xEnd,
                              HOMGPOINT2D_TYPE yEnd,
                              const std::vector<StorageType> &color,
                              unsigned int octet) {
  HOMGPOINT2D_TYPE deltaX, deltaY, d, deltaEast, deltaNortheast;
  
  DrawPixel(im, xStart, yStart, color, 1, octet, xStart, yStart, xEnd, yEnd);
  
  deltaX = xEnd - xStart;
  deltaY = yEnd - yStart;
  
  d = 2 * deltaY - deltaX;
  deltaEast = 2 * deltaY;
  deltaNortheast = 2 * (deltaY - deltaX);
  
  HOMGPOINT2D_TYPE y = yStart;
  for (HOMGPOINT2D_TYPE x = xStart + 1; x <= xEnd; x++) {
    if (d <= 0) {
      d = d + deltaEast;
    }
    else {
      d = d + deltaNortheast;
      if (deltaY >= 0) {
        y++;
      }
      else {
        y--;
      }
    }
    
    // draw
    DrawPixel(im, x, y, color, 1, octet, xStart, yStart, xEnd, yEnd);
  }
}


// Helper method
// TODO:
// - allow non-integer start and end points in anti-aliased mode
template <class StorageType>
inline void DrawLineGuptaSproull(Image<StorageType>& im,
                                 HOMGPOINT2D_TYPE xStart,
                                 HOMGPOINT2D_TYPE yStart,
                                 HOMGPOINT2D_TYPE xEnd,
                                 HOMGPOINT2D_TYPE yEnd,
                                 const std::vector<StorageType> &color,
                                 unsigned int octet) {
  int x, y;
  HOMGPOINT2D_TYPE p, m, c, s, firstY;
  
  // TODO: table as static class var
  // linear interpolation is performed between theses discrete values,
  // since for low slopes the steps are visible
  float lookupTable[] = {0.78f, 0.775f, 0.76f, 0.736f, 0.703f, 0.662f, 0.613f, 0.558f,
    0.5f, 0.441f, 0.383f, 0.328f, 0.276f, 0.228f, 0.184f, 0.145f, 0.11f, 0.08f, 0.056f,
    0.036f, 0.021f, 0.01f, 0.004f, 0.001f};
  int lookupTableLength = 24; // TODO: dynamically??
  
  float lookupTableEndPoints[][6] = {
    {0.0f,   0.0f,   0.0f, 0.056f, 0.393f, 0.056f},  // slope =  0/16
    {0.0f,   0.0f,   0.0f, 0.063f, 0.39f,  0.048f},  // slope =  1/16
    {0.0f,   0.0f,   0.0f, 0.072f, 0.39f,  0.042f},  // slope =  2/16
    {0.0f,   0.0f,   0.0f, 0.082f, 0.39f,  0.037f},  // slope =  3/16
    {0.0f,   0.0f,   0.0f, 0.094f, 0.39f,  0.032f},  // slope =  4/16
    {0.0f,   0.0f,   0.0f, 0.107f, 0.39f,  0.028f},  // slope =  5/16
    {0.0f,   0.001f, 0.0f, 0.121f, 0.39f,  0.025f},  // slope =  6/16
    {0.0f,   0.001f, 0.0f, 0.136f, 0.39f,  0.022f},  // slope =  7/16
    {0.0f,   0.002f, 0.0f, 0.152f, 0.391f, 0.019f},  // slope =  8/16
    {0.0f,   0.002f, 0.0f, 0.169f, 0.39f,  0.017f},  // slope =  9/16
    {0.0f,   0.003f, 0.0f, 0.187f, 0.39f,  0.015f},  // slope = 10/16
    {0.0f,   0.003f, 0.0f, 0.206f, 0.39f,  0.013f},  // slope = 11/16
    {0.0f,   0.004f, 0.0f, 0.225f, 0.39f,  0.012f},  // slope = 12/16
    {0.0f,   0.005f, 0.0f, 0.245f, 0.39f,  0.01f},   // slope = 13/16
    {0.0f,   0.006f, 0.0f, 0.264f, 0.39f,  0.009f},  // slope = 14/16
    {0.001f, 0.007f, 0.0f, 0.284f, 0.39f,  0.008f},  // slope = 15/16
    {0.001f, 0.007f, 0.0f, 0.304f, 0.391f, 0.007f}}; // slope = 16/16
  int index;
  
  m = (yEnd - yStart) / (xEnd - xStart);
  c = 1 / sqrt(m * m + 1);
  firstY = yStart + m * (int(xStart + 0.5) - xStart);
  p = 0;
  s = (0.5 - m) * c;
  y = (int)(firstY + 0.5);
  
  // draw end points
  index = int((m * 16) + 0.5);
  
  // left left side of line (start)
  int px = (int)(xStart + 0.5);
  int py = (int)(yStart + 0.5);
  
  DrawPixel(im, px - 1, py + 1, color, lookupTableEndPoints[index][0], octet,
            xStart, yStart, xEnd, yEnd);
  DrawPixel(im, px - 1, py, color, lookupTableEndPoints[index][1], octet,
            xStart, yStart, xEnd, yEnd);
  DrawPixel(im, px - 1, py - 1, color, lookupTableEndPoints[index][2], octet,
            xStart, yStart, xEnd, yEnd);
  DrawPixel(im, px, py + 1, color, lookupTableEndPoints[index][3], octet,
            xStart, yStart, xEnd, yEnd);
  DrawPixel(im, px, py, color, lookupTableEndPoints[index][4], octet,
            xStart, yStart, xEnd, yEnd);
  DrawPixel(im, px, py - 1, color, lookupTableEndPoints[index][5], octet,
            xStart, yStart, xEnd, yEnd);
  
  // right side of line (end)
  px = (int)(xEnd + 0.5);
  py = (int)(yEnd + 0.5);
  
  DrawPixel(im, px, py + 1, color, lookupTableEndPoints[index][5], octet,
            xStart, yStart, xEnd, yEnd);
  DrawPixel(im, px, py, color, lookupTableEndPoints[index][4], octet,
            xStart, yStart, xEnd, yEnd);
  DrawPixel(im, px, py - 1, color, lookupTableEndPoints[index][3], octet,
            xStart, yStart, xEnd, yEnd);
  DrawPixel(im, px + 1, py + 1, color, lookupTableEndPoints[index][2], octet,
            xStart, yStart, xEnd, yEnd);
  DrawPixel(im, px + 1, py, color, lookupTableEndPoints[index][1], octet,
            xStart, yStart, xEnd, yEnd);
  DrawPixel(im, px + 1, py - 1, color, lookupTableEndPoints[index][0], octet,
            xStart, yStart, xEnd, yEnd);
  
  HOMGPOINT2D_TYPE weight;
  HOMGPOINT2D_TYPE pAbs;
  HOMGPOINT2D_TYPE error;
  
  // iterate over line
  for (x = (int)(xStart + 0.5); x < (int)(xEnd + 0.5); x++) {
    
    
    
    // TODO: need some refactoring here!!! avoid duplicate code
    
    
    
    
    // skip firstpixel of line (handled seperately - see above)
    // TODO: better approach: determine y at xStart+1 and xEnd-1
    if (x != (int)(xStart + 0.5)) {
      // shade lower pixel
      pAbs = abs(p - c) * 16.0;
      index = int(pAbs + 0.5);
      error = pAbs - index;
      
      if (index < lookupTableLength) {
        if (error > 0 && index + 1 < lookupTableLength) {
          weight = (1.0 - error) * lookupTable[index];
          weight += error * lookupTable[index + 1];
        }
        else if (error < 0 && index > 0) {
          error = -error;
          weight = (1.0 - error) * lookupTable[index];
          weight += error * lookupTable[index - 1];
        }
        else {
          weight = lookupTable[index];
        }
        
        DrawPixel(im, x, y + 1, color, weight, octet,
                  xStart, yStart, xEnd, yEnd);
      }
      
      
      
      
      // determine shade of middle pixel
      pAbs = abs(p) * 16.0;
      index = int(pAbs + 0.5);
      error = pAbs - index;
      
      if (index < lookupTableLength) {
        if (error > 0 && index + 1 < lookupTableLength) {
          weight = (1.0 - error) * lookupTable[index];
          weight += error * lookupTable[index + 1];
        }
        else if (error < 0 && index > 0) {
          error = -error;
          weight = (1.0 - error) * lookupTable[index];
          weight += error * lookupTable[index - 1];
        }
        else {
          weight = lookupTable[index];
        }
        
        DrawPixel(im, x, y, color, weight, octet,
                  xStart, yStart, xEnd, yEnd);
      }
      
      
      
      
      
      
      
      
      // determine shade of upper pixel
      pAbs = abs(p + c) * 16.0;
      index = int(pAbs + 0.5);
      error = pAbs - index;
      
      if (index < lookupTableLength) {
        if (error > 0 && index + 1 < lookupTableLength) {
          weight = (1.0 - error) * lookupTable[index];
          weight += error * lookupTable[index + 1];
        }
        else if (error < 0 && index > 0) {
          error = -error;
          weight = (1.0 - error) * lookupTable[index];
          weight += error * lookupTable[index - 1];
        }
        else {
          weight = lookupTable[index];
        }
        
        DrawPixel(im, x, y - 1, color, weight, octet,
                  xStart, yStart, xEnd, yEnd);
      }
      
      
      
      
      
      
      
    }
    
    // prepare next iteration
    if (p >= s) {
      y++;
      p = p + (m - 1) * c; // TODO: effi
    }
    else {
      p = p + m * c; // TODO: effi
    }
  }
}


// TODO:
// - clip line to ROI
template <class StorageType>
int ImageDraw<StorageType>::Line(Image<StorageType>& im,
                                 const HomgPoint2D& pStart,
                                 const HomgPoint2D& pEnd,
                                 const std::vector<StorageType> &color,
                                 int antialiasing,
                                 double thickness) {
  // Bresenham (aliased) and Gupta & Sproull (anti-aliased)
  // for more info see
  // Bresenham: Algorithm for computer control of a digital plotter
  // Gupta and Sproull: Filtering Edges for Gray-Scale Images
  
  // sanity checks
  if (!color.empty() && color.size() != im.GetChannelCount()) {
    BIASERR("number of channels for line color must be equal to number of channels of image!");
    return -1;
  }
  
  // get start and end coords and homogenize them
  HOMGPOINT2D_TYPE xStart, yStart, xEnd, yEnd;
  xStart = pStart[0] / pStart[2];
  yStart = pStart[1] / pStart[2];
  xEnd = pEnd[0] / pEnd[2];
  yEnd = pEnd[1] / pEnd[2];
  
  // ensure that both points are inside image
  // TODO: clip automatically to ROI
  if (xStart < 0 || xStart > im.GetWidth()) {
    BIASERR("xStart is outside of image!");
    return -1;
  }
  if (yStart < 0 || yStart > im.GetHeight()) {
    BIASERR("yStart is outside of image!");
    return -1;
  }
  if (xEnd < 0 || xEnd > im.GetWidth()) {
    BIASERR("xEnd is outside of image!");
    return -1;
  }
  if (yEnd < 0 || yEnd > im.GetHeight()) {
    BIASERR("yEnd is outside of image!");
    return -1;
  }
  
  // determine octet and transform line so that it is in first octet
  // the transformation will be undone in the helper method DrawPixel()
  HOMGPOINT2D_TYPE xStartTransf, yStartTransf, xEndTransf, yEndTransf;
  unsigned int octet;
  
  HOMGPOINT2D_TYPE deltaX, deltaY;
  deltaX = xEnd - xStart;
  deltaY = yEnd - yStart;
  
  if (yStart < yEnd) {
    if (xStart < xEnd) {
      if (abs(deltaX) >= abs(deltaY)) {
        octet = 1;
        xStartTransf = xStart;
        yStartTransf = yStart;
        xEndTransf = xEnd;
        yEndTransf = yEnd;
      }
      else { // abs(deltaX) < abs(deltaY)
        octet = 2;
        xStartTransf = yStart;
        yStartTransf = xStart;
        xEndTransf = yEnd;
        yEndTransf = xEnd;
      }
    }
    else { // xStart >= xEnd
      if (abs(deltaX) >= abs(deltaY)) {
        octet = 4;
        xStartTransf = xEnd;
        yStartTransf = yStart;
        xEndTransf = xStart;
        yEndTransf = yEnd;
      }
      else { // abs(deltaX) < abs(deltaY)
        octet = 3;
        xStartTransf = yStart;
        yStartTransf = xEnd;
        xEndTransf = yEnd;
        yEndTransf = xStart;
      }
    }
  }
  else { // yStart >= yEnd
    if (xStart < xEnd) {
      if (abs(deltaX) >= abs(deltaY)) {
        octet = 8;
        xStartTransf = xStart;
        yStartTransf = yEnd;
        xEndTransf = xEnd;
        yEndTransf = yStart;
      }
      else { // abs(deltaX) < abs(deltaY)
        octet = 7;
        xStartTransf = yEnd;
        yStartTransf = xStart;
        xEndTransf = yStart;
        yEndTransf = xEnd;
      }
    }
    else { // xStart >= xEnd
      if (abs(deltaX) >= abs(deltaY)) {
        octet = 5;
        xStartTransf = xEnd;
        yStartTransf = yEnd;
        xEndTransf = xStart;
        yEndTransf = yStart;
      }
      else { // abs(deltaX) < abs(deltaY)
        octet = 6;
        xStartTransf = yEnd;
        yStartTransf = xEnd;
        xEndTransf = yStart;
        yEndTransf = xStart;
      }
    }
  }
  
  // draw the line
  if (antialiasing == 0) { // good old Bresenham (aliased)
    DrawLineBresenham(im, xStartTransf, yStartTransf, xEndTransf, yEndTransf,
                      color, octet);
  }
  else if (antialiasing == 1) { // Gupta and Sproull (anti-aliased)
    DrawLineGuptaSproull(im, xStartTransf, yStartTransf, xEndTransf, yEndTransf,
                         color, octet);
    return 0;
  }
  
  return 0;
}


///////////////////////////////////////////////////////////
// circles
//////////////////////////////////////////////////////////

template <class StorageType>
int ImageDraw<StorageType>::CircleCenter(Image<StorageType>& im,
                                         unsigned int CenterX, 
                                         unsigned int CenterY, 
                                         unsigned int Radius, 
                                         const StorageType Value[])
{
    int res=0;
    BresenhamCircle bc;
    int center[2], next[]={0,0}, radius=(int)Radius;
    StorageType **ida=im.GetImageDataArray();
    unsigned int cc=im.GetChannelCount();  
    const int height = (int)im.GetHeight();
    const int width = (int)im.GetWidth();

    register unsigned int c;

    center[0]=(int)CenterX;
    center[1]=(int)CenterY;

    bc.Init(center, radius);

    if (Value==NULL) {
        // no color supplied, use black and white dep. on background
        while (bc.GetNext(next)){
            if (next[1]>=0 && next[1]<height && 
                next[0]>=0 && next[0]<width) {
                    for (c=0; c<cc; c++) {
                        ida[next[1]][next[0]*cc+c]=
                            GetContrastValue(ida[next[1]][next[0]*cc+c]);
                    }
                }
        }
    } else {
        while (bc.GetNext(next)){
            if (next[1]>=0 && next[1]<height && 
                next[0]>=0 && next[0]<width) {
                    for (c=0; c<cc; c++) {
                        ida[next[1]][next[0]*cc+c]=Value[c];
                    }
                }
        }
    }
    return res;
}

template <class StorageType>
int ImageDraw<StorageType>::CircleCenterFilled(Image<StorageType>& im,
                                               unsigned int CenterX, 
                                               unsigned int CenterY, 
                                               unsigned int Radius, 
                                               const StorageType Value[])
{
    int res=0;
    BresenhamCircleEighth bc;
    int zerocenter[]={0,0};
    int center[2], next[2], radius=(int)Radius;
    unsigned int start[2], end[2];
    const int width = (int)im.GetWidth();
    const int height = (int)im.GetHeight();
    center[0]=(int)CenterX;
    center[1]=(int)CenterY;

    if (center[0]<=radius || center[1]<=radius || 
        center[0]+radius+1>=width || center[1]+radius+1>=height){
      return -1;
    }

    bc.Init(zerocenter, radius);

    while (bc.GetNext(next)){
        start[0]=center[0]+next[0]; 
        start[1]=center[1]+next[1];
        end[0]=center[0]-next[0]; 
        end[1]=center[1]+next[1];
        Line(im, start, end, Value);
        start[0]=center[0]+next[1]; 
        start[1]=center[1]+next[0];
        end[0]=center[0]-next[1]; 
        end[1]=center[1]+next[0];
        Line(im, start, end, Value);
        start[0]=center[0]+next[0]; 
        start[1]=center[1]-next[1];
        end[0]=center[0]-next[0]; 
        end[1]=center[1]-next[1];
        Line(im, start, end, Value);
        start[0]=center[0]+next[1]; 
        start[1]=center[1]-next[0];
        end[0]=center[0]-next[1]; 
        end[1]=center[1]-next[0];
        Line(im, start, end, Value);
    }

    return res;
}


template <class StorageType>
int ImageDraw<StorageType>::Arrow(Image<StorageType>& im,
                                  const unsigned start[2],
                                  const unsigned mend[2],
                                  const unsigned length, 
                                  const unsigned width, 
                                  const StorageType value[])
{
  int dx=(int)mend[0]-(int)start[0], dy=(int)mend[1]-(int)start[1];
  unsigned end[]={mend[0], mend[1]};
  double l = sqrt((double)dx*(double)dx+(double)dy*(double)dy);
  if (l<(double)length){
    unsigned num = (unsigned)ceil( (double)length / l );
    end[0] += dx * num;
    end[1] += dy * num;  
  }

  float thick=2.0; 
   
  //if (Line(im, start, end, value)!=0){
  //cout << "line: "<<start[0]<<"  "<<start[1]<<"  "<<end[0]<<"  "<<end[1]<<endl;
  if (Line(im, start[0], start[1], end[0], end[1], value, thick)!=0){
#ifdef BIAS_DEBUG   
    BIASERR("error drawing line");
#endif
    return -1;
  }
  if (dx==0 && dy==0){ 
    // do not draw arrow, zero length and hence direction indetermined
    return -2;
  } 

#ifdef BIAS_DEBUG
  if (sqrt(double(dx*dx+dy*dy))<length || sqrt(double(dx*dx+dy*dy))<width){
    BIASERR("funny arrow, head bigger than tail");
    // BIASABORT;
  }
#endif  
  int next[2]={0,0};
  Bresenham bres(end, start), bres2;

  for (unsigned i=0; i<length; i++)
    bres.GetNext(next);

  int left[]={0,0}, right[]={0,0};
  int lend[2], rend[2];
  rend[0]=(int)next[0]-dy;
  rend[1]=(int)next[1]+dx;
  lend[0]=(int)next[0]+dy;
  lend[1]=(int)next[1]-dx;
  //cerr << "next: ("<<next[0]<<", "<<next[1]<<")  lend: ("<<lend[0]<<", "
  //     <<lend[1]<<")  rend: ("<<rend[0]<<", "<<rend[1]<<")\n";
  bres.Init(next, rend);
  bres2.Init(next, lend);
  for (unsigned i=0; i<width; i++){
    bres.GetNext(right);
    bres2.GetNext(left);
  }

  int res = 0;
  if (right[0]>=0 && right[1]<(int)im.GetWidth()){
    //cout << "right: "<<end[0]<<"  "<<end[1]<<"  "<<right[0]<<"  "<<right[1]<<endl;
    if (Line(im, end[0], end[1], (unsigned)(right[0]), (unsigned)(right[1]), 
             value, thick)!=0){
      //if (Line(im, end, (unsigned *)right, value)!=0){
#ifdef BIAS_DEBUG  
      BIASERR("error drawing line");
#endif
      return -1;
    }
  } else { res = -1; }
  
  if (left[0]>=0 && left[1]<(int)im.GetWidth()){
    //cout << "left: "<<end[0]<<"  "<<end[1]<<"  "<<left[0]<<"  "<<left[1]<<endl;
    if (Line(im, end[0], end[1], (unsigned)(left[0]), (unsigned)(left[1]), 
             value, thick)!=0){
    //if (Line(im, end, (unsigned *)left, value)!=0){
#ifdef BIAS_DEBUG  
      BIASERR("error drawing line");
#endif
      return -1;
    } 
  } else { res = -1; }
  return res;
}


  /// JW: 
template <class StorageType>
int ImageDraw<StorageType>::Line(Image<StorageType>& im, 
                                 const unsigned int StartX, 
                                 const unsigned int StartY, 
                                 const unsigned int EndX, 
                                 const unsigned int EndY, 
                                 const StorageType color[],
                                 const float thickness )
{
#ifdef BIAS_HAVE_OPENCV
  int i_thickness=int(thickness);
  Line(im,
       StartX, StartY,
       EndX, EndY,
       ColourRGB<StorageType>(color[0],color[1], color[2]),
       i_thickness
       );
#else
  BIASERR("Please enable OpenCV to use Lines iwth thickness. (jw)");
  // workareound : draw multiple lines to get a thicker one:

  // direction of (original) line
  double dir[2]; //< direction od the line
  double ortho[2]; //< perpendicalur vector of line displacement for thickness
  dir[0] = EndX - StartX; //< dx
  dir[1] = EndY - StartY; //< dy

  // orthogonal offset is perpendicular to line direction: 
  ortho[0] =  -dir[1];
  ortho[1] = dir[0]; //< y, x as perpendicular of dir
  double norm = sqrt(ortho[0]*ortho[0] + ortho[1]*ortho[1]);
  if (norm ==0) {
    ortho[0] = 1.0; ortho[1] = 0.0;
  } else {
    ortho[0] /= norm; // length one 
    ortho[1] /= norm; // length one 
  };

  // draw multiple parallel lines to emulate thickness 
  // QUCKHACK. \todo implement Bresenham for thicked lines or even better with antialised lines. JW
  // \todo the line may actually be shifted for fractional part of thickness JW
  // it is not enogh to use 1.0 step...
  for (float l= float(-thickness/2.0); l <= float(thickness/2.0); l+=1.0) {
    //ImageDraw<StorageType::
    Line(im, 
         (unsigned int)(StartX +l * ortho[0]), 
         (unsigned int)(StartY +l * ortho[1]), 

         (unsigned int)(EndX   +l * ortho[0] ), 
         (unsigned int)(EndY   +l * ortho[1] ), 
         color 
         );
  };
#endif // OpenCV
  return 0;
}

  template <class StorageType>
    int ImageDraw<StorageType>::Ellipse(Image<StorageType>& im, 
                                        double center[2], double a[2], 
                                        double b[2], 
                                        const StorageType Value[])
    {
      // points on ellipse can be parametrized by angle t
      // (a*cos(t), b*sin(t)) if the ellipse-axes are parallel to coordinate axes

      const int w = im.GetWidth();
      const int h = im.GetHeight();
      StorageType **ida=im.GetImageDataArray();
      const int cc=im.GetChannelCount();

      double length_a = sqrt(a[0]*a[0]+a[1]*a[1]);
      double length_b = sqrt(b[0]*b[0]+b[1]*b[1]);

      // test for degenerate cases, where the ellipse 
      // consists only of a single line
      if (length_a<1e-8){
        double start[2], end[2];
        start[0]=center[0]-b[0];
        start[1]=center[1]-b[1];
        end[0]=center[0]+b[0];
        end[1]=center[1]+b[1];
        return ImageDraw<StorageType>::Line(im, start, end, Value);
      }
      if (length_b<1e-8){
        double start[2], end[2];
        start[0]=center[0]-a[0];
        start[1]=center[1]-a[1];
        end[0]=center[0]+a[0];
        end[1]=center[1]+a[1];
        return ImageDraw<StorageType>::Line(im, start, end, Value);
      }

#ifdef BIAS_DEBUG
      // assume axes are perpendicular
      if ((a[0]*b[0]+a[1]*b[1])/(length_a*length_b)>1e-8){
        BIASERR("half axes of ellipse are not perpendicular: ("<<a[0]<<", "<<a[1]
                <<")   ("<<b[0]<<", "<<b[1]<<")\n");
        BIASASSERT((a[0]*b[0]+a[1]*b[1])/(length_a*length_b)<1e-8);
      }
#endif

      double dt;
      if (length_a>length_b){
        dt = atan2(1.0, length_a);
      } else {
        dt = atan2(1.0, length_b);
      }
      //cout << "dt: "<<dt<<endl;

      // angel of axis a with x axis
      double ang=atan2(a[1], a[0]);
      //cout << "ang: "<<ang*180.0/M_PI<<" deg"<<endl;
      // rotation with respect to coordinate system
      double ca=cos(ang), sa=sin(ang); 

      double p[2], rp[2], ap[2];
      double t=0.0;
      int ix, iy;

      while (t<=M_PI_2){
        //cout << "t: "<<t<<endl;
        p[0] = length_a * cos(t);
        p[1] = length_b * sin(t);
        //cout << "p : "<<p[0]<<", "<<p[1]<<endl;

        ap[0] = p[0];
        ap[1] = p[1];
        // now rotate and translate the point
        rp[0] = ap[0]*ca - ap[1]*sa + center[0];
        rp[1] = ap[1]*ca + ap[0]*sa + center[1];
        //cout << "1 - rp : "<<rp[0]<<", "<<rp[1]<<endl;
        ix = (int)rint(rp[0]);
        iy = (int)rint(rp[1]);
        if (ix>=0 && ix<w  && iy>=0 && iy<h){
          for (int i=0; i<cc; i++)
            ida[iy][ix*cc+i]=Value[i];
        }

        // use 4-ways symmetry of ellipse
        ap[0] = -p[0];
        ap[1] = -p[1];
        // now rotate and translate the point
        rp[0] = ap[0]*ca - ap[1]*sa + center[0];
        rp[1] = ap[1]*ca + ap[0]*sa + center[1];
        //cout << "2 - rp : "<<rp[0]<<", "<<rp[1]<<endl;
        ix = (int)rint(rp[0]);
        iy = (int)rint(rp[1]);
        if (ix>=0 && ix<w  && iy>=0 && iy<h){
          for (int i=0; i<cc; i++)
            ida[iy][ix*cc+i]=Value[i];
        }

        // use 4-ways symmetry of ellipse
        ap[0] = p[0];
        ap[1] = -p[1];
        // now rotate and translate the point
        rp[0] = ap[0]*ca - ap[1]*sa + center[0];
        rp[1] = ap[1]*ca + ap[0]*sa + center[1];
        //cout << "3 - rp : "<<rp[0]<<", "<<rp[1]<<endl;
        ix = (int)rint(rp[0]);
        iy = (int)rint(rp[1]);
        if (ix>=0 && ix<w  && iy>=0 && iy<h){
          for (int i=0; i<cc; i++)
            ida[iy][ix*cc+i]=Value[i];
        }

        // use 4-ways symmetry of ellipse
        ap[0] = -p[0];
        ap[1] = p[1];
        // now rotate and translate the point
        rp[0] = ap[0]*ca - ap[1]*sa + center[0];
        rp[1] = ap[1]*ca + ap[0]*sa + center[1];
        //cout << "4 - rp : "<<rp[0]<<", "<<rp[1]<<endl;
        ix = (int)rint(rp[0]);
        iy = (int)rint(rp[1]);
        if (ix>=0 && ix<w  && iy>=0 && iy<h){
          for (int i=0; i<cc; i++)
            ida[iy][ix*cc+i]=Value[i];
        }

        t+=dt;
      }
      return 0;
    }


#ifdef BIAS_HAVE_OPENCV
  // JW
  template <class StorageType>
    void ImageDraw<StorageType>::Text(Image<StorageType>& dstImg,
                                      const std::string & message,
                                      const int & posX,
                                      const int & posY,
                                      const BIAS::ColourRGB<StorageType> &colorRGB,
                                      const int fontface,
                                      const double hscale,
                                      const double vscale,
                                      const double shear,
                                      const int thickness,
                                      const int linetype
                                      )
    {
      WrapBias2Ipl wrap( &dstImg );
      CvFont font;
      cvInitFont( &font, 
                  fontface,
                  hscale,
                  vscale, 
                  shear,
                  thickness,
                  linetype
                  );
      cvPutText( wrap.p_imgIpl, 
                 message.c_str(), 
                 cvPoint(posX, posY), 
                 &font, 
                 cvScalar(colorRGB[0], colorRGB[1], colorRGB[2]) // no swap!
                 );
    }

  // JW
  template <class StorageType>
    void ImageDraw<StorageType>::Circle(BIAS::Image<StorageType>& dstImg,
                                        const int & centerX, 
                                        const int & centerY,
                                        const int & radius,
                                        const BIAS::ColourRGB<StorageType> &colorRGB,
                                        const int & thickness,
                                        const int & linetype,
                                        const int & shift
                                        )
    {
      WrapBias2Ipl wrap( &dstImg );
      cvCircle( wrap.p_imgIpl, 
                cvPoint (centerX, centerY),
                radius, 
                cvScalar(colorRGB[0], colorRGB[1], colorRGB[2] ),
                thickness,
                linetype,
                shift
                );
    }

  // JW
  template <class StorageType>
    void ImageDraw<StorageType>::Line(BIAS::Image<StorageType>& dstImg,
                                      const int & pt1X,
                                      const int & pt1Y,
                                      const int & pt2X,
                                      const int & pt2Y,
                                      const BIAS::ColourRGB<StorageType> &colorRGB,
                                      const int thickness,
                                      const int linetype,
                                      const int shift
                                      )
    {
      WrapBias2Ipl wrap( &dstImg );
      cvLine( wrap.p_imgIpl, 
              cvPoint (pt1X, pt1Y),
              cvPoint (pt2X, pt2Y),
              cvScalar(colorRGB[0], colorRGB[1], colorRGB[2] ),
              thickness,
              linetype,
              shift
              );
    }

template <class StorageType>
void ImageDraw<StorageType>::
Arrow(Image<StorageType>& im, const int startX,
      const int startY, const int endX, const int endY,
      const BIAS::ColourRGB<StorageType> &colorRGB, 
      const int thickness)
{
  WrapBias2Ipl wrap( &im );
  int dx=endX-startX, dy=endY-startY;
  if (dx==0 && dy==0){ 
    // do not draw arrow, zero length and hence direction indetermined
    return;
  }
  CvPoint start, end;
  start.x = startX;
  start.y = startY;
  end.x = endX;
  end.y = endY;
  
  CvScalar color = cvScalar(colorRGB[0], colorRGB[1], colorRGB[2]);
  double angle = atan2( (double)(start.y - end.y), (double)(start.x - end.x) );
  int shift = 0;

  // main line
  cvLine(wrap.p_imgIpl, start, end, color, thickness, CV_AA, shift);

  /* Now draw the tips of the arrow. I do some scaling so that the
   * tips look proportional to the main line of the arrow.   */
  start.x = (int) (end.x + 6 * cos(angle + M_PI / 4));
  start.y = (int) (end.y + 6 * sin(angle + M_PI / 4));
  cvLine(wrap.p_imgIpl, start, end, color, thickness, CV_AA, shift);

  start.x = (int) (end.x + 6 * cos(angle - M_PI / 4));
  start.y = (int) (end.y + 6 * sin(angle - M_PI / 4));
  cvLine(wrap.p_imgIpl, start, end, color, thickness, CV_AA, shift);
  
}


  // JW
  template <class StorageType>
    void ImageDraw<StorageType>::Rectangle(BIAS::Image<StorageType>& dstImg,
                                           const int & pt1X,
                                           const int & pt1Y,                        
                                           const int & pt2X, 
                                           const int & pt2Y,
                                           const BIAS::ColourRGB<StorageType> &colorRGB,
                                           const int thickness,
                                           const int linetype,
                                           const int shift
                                           )
    {
      WrapBias2Ipl wrap( &dstImg );
      cvRectangle(wrap.p_imgIpl, 
                  cvPoint (pt1X, pt1Y),
                  cvPoint (pt2X, pt2Y),
                  cvScalar(colorRGB[0], colorRGB[1], colorRGB[2] ),
                  thickness,
                  linetype,
                  shift
                  );
    }  

  // JW
  template <class StorageType>
    void ImageDraw<StorageType>::Ellipse(BIAS::Image<StorageType>& dstImg,
                                         const int & centerX,
                                         const int & centerY,
                                         const int & axesX,
                                         const int & axesY,
                                         const double & angle,
                                         const double & start_angle,
                                         const double & end_angle,
                                         const BIAS::ColourRGB<StorageType> &colorRGB,
                                         const int & thickness,
                                         const int & linetype,
                                         const int & shift
                                         )
    {
      WrapBias2Ipl wrap( &dstImg );
      cvEllipse(wrap.p_imgIpl, 
                cvPoint(centerX, centerY), 
                cvSize (axesX, axesY),
                angle,
                start_angle,
                end_angle,
                cvScalar(colorRGB[0], colorRGB[1], colorRGB[2] ),
                thickness,
                linetype,
                shift );
    }


#endif //BIAS_HAVE_OPENCV

} // namespace BIAS


//
// solve explicit instantiation
//
#define INSTANCE_ImageDraw(type)\
template class BIASImageBase_EXPORT ImageDraw<type>;


// create instances
INSTANCE_ImageDraw(unsigned char)
INSTANCE_ImageDraw(float)
#ifdef BUILD_IMAGE_INT
INSTANCE_ImageDraw(int)
#endif
#ifdef BUILD_IMAGE_CHAR
INSTANCE_ImageDraw(char)
#endif
#ifdef BUILD_IMAGE_SHORT
INSTANCE_ImageDraw(short)
#endif
#if defined(BUILD_IMAGE_USHORT) 
INSTANCE_ImageDraw(unsigned short)
#endif
#ifdef BUILD_IMAGE_DOUBLE
INSTANCE_ImageDraw(double)
#endif
#ifdef BUILD_IMAGE_UINT
INSTANCE_ImageDraw(unsigned int)
#endif
