// author Jan Woetzel (c) 2006
#include "CheckerboardDetector.hh"
#include <Base/Common/W32Compat.hh>
#include <math.h>

#include <string>
#include <iostream>
#include <fstream>
#include <iomanip>
#include <sstream>

//BIAS
#include <Base/Debug/Error.hh>
#include <Base/Math/Vector3.hh>
#include <Base/Image/ImageIO.hh>
#include <Base/Image/ImageConvert.hh>

#include <cv.h>
#include <highgui.h>

using namespace std;
using namespace BIAS;


int CheckerboardDetector::Compute(BIAS::Image<unsigned char> & img 
                                  ,const unsigned int & xCorners
                                  ,const unsigned int & yCorners
                                  ,t_CoordVecf & coords
                                  ,const bool & saveDebugImage
                                  )
{
 
  IplImage *image = NULL;
  IplImage *threshold = NULL;
  IplImage *gray = NULL;

  // allocate
  const int numCorners=xCorners*yCorners;
  coords.resize(numCorners, std::pair<float,float>(0.0f,0.0f));
  int cornersFound(0);
  // conversion required fromBIAS to cv image
  BIAS::Image<unsigned char> imgGrey;
  BIAS::Image<unsigned char> imgRGB;
  ImageConvert::ToRGB(img, imgRGB);
  ImageConvert::BIAS2ipl(imgRGB, image);
  BIASASSERT(image!=NULL); // loaded?
  ImageConvert::ToGrey(img, imgGrey);
  ImageConvert::BIAS2ipl(imgGrey, gray);
  BIASASSERT(gray!=NULL); // loaded?

  threshold   = cvCreateImage(cvSize(image->width, image->height), IPL_DEPTH_8U, 1);
  BIASASSERT(threshold !=NULL);
  cornersFound = xCorners * yCorners;

  // chessboard detection 
  int bResult = cvFindChessBoardCornerGuesses(
                                              gray 
                                              ,threshold
                                              ,NULL 
                                              ,cvSize(xCorners, yCorners)
                                              ,(CvPoint2D32f*)&(coords[0].first)
                                              ,&cornersFound);
  if (!bResult)
    {
      //BIASERR("Could not find all "<<xCorners<<"x"<<yCorners<<" inner corners");
      return -3;
    }

  // subpixel refinement
  cvFindCornerSubPix(
                     gray
                     ,(CvPoint2D32f*)&coords[0].first
                     ,cornersFound 
                     ,cvSize(5,5)
                     ,cvSize(-1,-1)
                     ,cvTermCriteria(CV_TERMCRIT_ITER, 100, 0.1)
                     );

  if (saveDebugImage){
    // confirm correctness of pos. 
    DrawPoints(image
               ,(CvPoint2D32f*)&coords[0].first
               ,cornersFound, xCorners, yCorners);
    string outFilename("out__findChessBoardCornerGuesses.jpg");
    cvSaveImage(outFilename.c_str(), image);
  }

  cvReleaseImage(&image);
  cvReleaseImage(&gray);
  cvReleaseImage(&threshold);		

  cvDestroyAllWindows();
  return 0;
}





/* DrawPoints takes number of corners on x axis and y axis, 
and number of images to builds an array of corners.  
Assumes all corners on z = 0 plane. */
void CheckerboardDetector::DrawPoints(CvArr *img
                                      ,CvPoint2D32f *ptArray
                                      //,t_CoordVecf & coords
                                      ,int numPts
                                      ,const int xCorners
                                      ,const int yCorners
                                      ,const bool DRAW_IDEAL_OVERLAY                                      
                                      ,const bool DRAW_AXES
                                      ,const bool DRAW_POINT_LABEL
                                      )
{
  CvPoint pt;
  CvPoint ptOld;
  CvFont font;
  if (DRAW_POINT_LABEL) {
    cvInitFont( 
      &font
      ,CV_FONT_HERSHEY_PLAIN //font_face
      ,1.0f //hscale
      ,1.0f //vscale
      );
  }

  // init old as first 
  if (numPts>0){
    ptOld.x=(int)ptArray[0].x;
    ptOld.y=(int)ptArray[0].y;
  }
  for (int i = 0; i < numPts; i++)
  {
    int clr = (int)((float)i / numPts * 255.0);
    pt.x = (int)ptArray[i].x;
    pt.y = (int)ptArray[i].y;
    cvCircle(img, pt, 3, CV_RGB(255-clr,0,clr), CV_FILLED);

    if (DRAW_POINT_LABEL){
      // label the point
      stringstream ssID;
      ssID<<i;
      cvPutText(img
        ,ssID.str().c_str()
        ,pt
        ,&font
        ,CV_RGB(0,0,255) 
        );
    }

    // next
    ptOld=pt;
  };

  if (DRAW_IDEAL_OVERLAY){
    // mark ideal pattern
    // horizontal lines
    for (int xx=0; xx<xCorners*yCorners; xx+=xCorners)
    {
      cvLine(img
        ,cvPointFrom32f(ptArray[xx])
        ,cvPointFrom32f(ptArray[xx +xCorners -1])
        ,CV_RGB(0,0,255),1,CV_AA );
    };
    // vertical lines
    for (int yy=0; yy<xCorners; yy+=1)
    {
      cvLine(img
        ,cvPointFrom32f(ptArray[yy])
        ,cvPointFrom32f(ptArray[yy +xCorners*(yCorners-1) ])
        ,CV_RGB(0,0,255),1,CV_AA );
    };
  }

  if (DRAW_AXES){
    CvFont fontBig;
    cvInitFont( 
      &fontBig
      ,CV_FONT_HERSHEY_PLAIN //font_face
      ,5.0f //hscale
      ,5.0f //vscale
      ,0 //shear
      ,2 //thickness
      ,CV_AA //line_type
      );

    cvPutText(img
      ,"0"
      ,cvPointFrom32f(ptArray[0])
      ,&fontBig
      ,CV_RGB(0,0,255) 
      );

    cvLine(img
      ,cvPointFrom32f(ptArray[0])
      ,cvPointFrom32f(ptArray[xCorners -1])
      ,CV_RGB(255,0,0)
      ,1 // thickness=1
      ,CV_AA //line_type
      );
    cvPutText(img
      ,"X"
      ,cvPointFrom32f(ptArray[xCorners -1])
      ,&fontBig
      ,CV_RGB(255,0,0) 
      );
    cvLine(img
      ,cvPointFrom32f(ptArray[0])
      ,cvPointFrom32f(ptArray[xCorners*(yCorners-1)])
      ,CV_RGB(0,255,0)
      ,1 // thickness=1
      ,CV_AA //line_type
      );
    cvPutText(img
      ,"Y"
      ,cvPointFrom32f(ptArray[xCorners*(yCorners-1)])
      ,&fontBig
      ,CV_RGB(0,255,0) 
      );
  }

  cvNamedWindow("img",1);
  cvMoveWindow("img",0,0);
  cvShowImage("img", img);
  cvWaitKey(1000);
}


