/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#ifndef __BIAS_CoordSystsemTransform3D_hh__
#define __BIAS_CoordSystsemTransform3D_hh__

// never include this unless it is absolutely needed:
// #include "bias_config.h" 

#include <Base/Common/BIASpragmaStart.hh>
#include <Base/Debug/Debug.hh>
#include <Base/Geometry/Quaternion.hh>
#include <Base/Math/Matrix4x4.hh>
#include <Base/Geometry/HomgPoint3D.hh>
#include <Base/Geometry/HomgPoint3DCov.hh>
#include <Base/Math/Vector3.hh>

namespace BIAS {


  /** @class CoordinateTransform3D
   *  @ingroup g_geometry
   *  @brief transforms 3d points between two different coodinate systems, e.g.
   *         from camera to world coordinates using the camera pose
   *
   *  We define a local coordinate system LCS(imagine camera coordinate system)
   *  and a global coordinate system GCS (imagine world coordinate system). 
   *  Imagine a 4x4 matrix 
   *
   *               [                C_[0] ]<br>
   *               [   scale_*R     C_[1] ]<br>
   *         T  =  [                C_[2] ] = invTransform_<br>
   *               [  0    0    0     1   ]<br>
   *
   *  This matrix converts a point X_L in LCS to its representation in GCS like
   *
   *
   *                  X_G = T * X_L
   *
   *  R is the matrix equivalent to Q_.
   * 
   *  Obviously,      X_L = T.Invert() * X_G
   *
   *         T.invert() =  [   R^T/scale_      -R^T*C_/scale_ ] = (*this)<br>
   *                       [  0      0      0        1        ]<br>
   *
   *  C is the origin of the local coordinate system in global coordinates
   *  Q (or R) tranforms a direction in local coordinates into a direction
   *  in global coordinates.
   *  The terms global and local are only used here to distinguish between
   *  two arbitrary coordinate systems, neither of them has to be local/global.
   *
   * \attention do not confuse T with the matrix stored in (*this) (see above)!
   *
   *  Apply for: hand-eye transformation, global-local coordinates,
   *  camera to world coordinates, ...
   *  @author koeser/bartczak 03/2006
   */
  class BIASGeometry_EXPORT CoordinateTransform3D :
    protected Matrix4x4<double>
  {
  public:
    /** @brief sets identity transform with zero covariance */
    CoordinateTransform3D()
      : Matrix4x4<double>(MatrixIdentity), Q_(0.0, 0.0, 0.0, 1.0), 
        C_(0.0,0.0,0.0), scale_(1.0),  
        invTransform_(MatrixIdentity) {}

    /** @brief calls operator= */
    CoordinateTransform3D(const CoordinateTransform3D& c)
    {
      *this = c;
    }

    /** @brief copy all elements */
    const CoordinateTransform3D& 
    operator=(const CoordinateTransform3D& c) {
      Q_= c.Q_;
      C_ = c.C_;
      scale_ = c.scale_;
      invTransform_ = c.invTransform_;
      Matrix4x4<double>::operator=(c);
      return (*this);
    }

    /** @brief use matrix, which transforms points in global coordinate system
        to points in local coordinate system (like pmatrix ) 
        \attention use only for matrices that cannot be decomposed properly,
        otherwise use Set method!
    */
    void SetFromMatrix3x4(const Matrix3x4<double> &M);

//     /** @brief set from pose object
//     */
//     void SetFromPose(const Pose &pose);

    /** @brief get matrix, which transforms points in global coordinate system
        to points in local coordinate system */
    Matrix3x4<double> GetMatrix3x4() const;

    void SetQ(const Quaternion<double>& Q) { Q_ = Q; UpdateMatrix_(); }

    /** Sets the upper 3x3 matrix accordingly (copies values)
     */
    void SetR(const BIAS::RMatrixBase &R);

    /** @brief sets origin of local coordinate system in global coordinates */
    void SetC(const Vector3<double>& C) { C_ = C; UpdateMatrix_(); }

    /** \brief Define local coordinate frame from within the world frame.
     * \param[in] Q Tells the orientation of the local frame in 
     *    the world frame base: 
     *    Q*(global base vecor)*Q_conj = local base vector.
     * \param[in] C Tells the position of th origin of the local frame in the
     *    world coordinate frame:
     *     global base origin + C_ = local base origin.
     * \param[in] scale Is the isotropic scale of base vector length to get
     *    local base vector length: 
     *    ||global base vector||*scale = ||local base vector||.
     */
    virtual void Set(const Quaternion<double>& Q, const Vector3<double>& C, 
                     const double& scale = 1.0) {
      Q_ = Q; C_ = C; scale_ = scale; UpdateMatrix_();
    }

    /** @brief sets the length of the local coordinate axes in the world 
     *         coordinate frame.
     */
    inline void SetScale(const double& scale) { 
      scale_ = scale; 
      UpdateMatrix_();}

    HomgPoint3D GlobalToLocal(const HomgPoint3D& X_G) const {
      Vector4<double> X_L;
      this->Mult(X_G, X_L);
      return X_L;
    }
    
    HomgPoint3D LocalToGlobal(const HomgPoint3D& X_L) const {
      Vector4<double> X_G;
      invTransform_.Mult(X_L, X_G);
      return X_G;
    }

    void GlobalToLocal(HomgPoint3D& X_G, 
                       HomgPoint3DCov& Cov_G) const {
      Vector4<double> X(X_G);
      this->Mult(X, X_G);
      Cov_G = (*this) * Cov_G * (this->Transpose());
    }
    
    void LocalToGlobal(HomgPoint3D& X_L, 
                       HomgPoint3DCov& Cov_L) const {
      HomgPoint3D X(X_L);
      invTransform_.Mult(X, X_L);
      Cov_L = invTransform_ * Cov_L * invTransform_.Transpose();
    }

    const double& GetScale() const { return scale_; }
    
    const Quaternion<double>& GetQ() const { return Q_; }

    const Vector3<double>& GetC() const { return C_; }

    inline const Vector<double> GetCQ() const {
      Vector<double> res(7);
      for (int i=0; i<3; i++){
        res[i] = C_[i]; res[i+3] = Q_[i];
      }
      res[6] = Q_[3];
      return res;
    }

    /** Returns rotation that maps from local coordiantes to 
    * global coordinates.
    */
    inline BIAS::RMatrixBase GetR() const {
      RMatrixBase R; 
      R.SetFromQuaternion(GetQ());
      return R;
    }
 
    inline  Matrix4x4<double> GetGlobalToLocalTransform() const
    { return *this; }

    inline  Matrix4x4<double> GetLocalToGlobalTransform() const
    { return invTransform_; }

    friend std::ostream& operator<<(std::ostream &os, 
                                    const CoordinateTransform3D& p);
    friend std::istream& operator>>(std::istream &is, 
                                    CoordinateTransform3D& p);
    
    bool Load(const std::string& inputFile);
    bool Save(const std::string& outputFile);

    /** \brief Calculates transformation from global system of this to local 
     *  system of argument transformation.
     *  if this defines [global <-> local] and
     *  argument (localT) defines [local <-> local(2)].
     *  the res then defines [global <-> local(2)].
     */
    void ConcatenateLocalTransform(
                               const CoordinateTransform3D& localT,
                               CoordinateTransform3D& res ) const;

    /** \brief Changes this so that newLocal
     * becomes the local coordinate frame and newGlobal becomes the global
     * system. 
     * 
     * It is assumed that newLocal and newGlobal are related through the same
     * global frame meaning: a world point p(world) is described by the the
     * coordinates p(newGlobal) and p(newLocal) so that 
     * p(world) = T(newLocal)*p(newLocal) = T(newGlobal)*p(newGLobal)
     * it follows that p(newGlobal) (which is will become the new global 
     * coord frame point) is found through 
     * p(newGlobal) = T(newGlobal)^{-1}*T(newLocal)*p(newLocal)
     *
     * newLocal : [global -> local]
     * newGlobal : [global -> local(B)]
     * Methods then does:
     *  (*this) : [local(B) -> local] (local(B) becomes global in this context)
     */
    void BecomeRelativeTransform(const CoordinateTransform3D& newLocal,
                                 CoordinateTransform3D newGlobal);

    /** Swapps the meaning of global and local.
     */
    void InvertIP() { 
      Q_.Invert(); scale_ = 1.0/scale_; 
      C_ = -scale_ * Q_.MultVec(C_); 
      UpdateMatrix_(); }

    /** Applies the passed rotation to the orientation of the local frame.
    Convenience wrapper, which simluates
    \verbatim
       LocalCoordTrsf.Set(Q); 
       LocalCoordTrsf.Set(Vector3<double>(0,0,0));
       this->ConcatenateLocalTransform(Q, *this);
       \enverbatim
    **/
    void RotateLocalFrame(const Quaternion<double>& Q);

    /**
     * Prints Transformation together with more intutive rotation 
     * representations.
     */
    void Talk() const;

  protected:
    /** @brief computes the 4x4 matrix */
    virtual void UpdateMatrix_();

    /// this is the rotation between the systems
    Quaternion<double> Q_;

    /// the translation between the systems
    Vector3<double> C_;

    /**\brief the scale between the systems
     *||vec1Global|| * scale_ = ||vec1Local||
     */
    double scale_;

    /// the uncertainty of the orientation / translation
    // yet no complete model for statisitcis on a transformation
    //    Matrix<POSE_TYPE> Cov_;

    /// matrix representation of inverse transform (LocalToGlobalTranform)
    Matrix4x4<double> invTransform_;
  };

  inline std::ostream& operator<<(std::ostream &os, 
                                  const CoordinateTransform3D& p) {
    os<<"Q_= "<<p.Q_<<" C_= "<<p.C_<<" scale_= "<<p.scale_;
    return os;  
  }
  
  inline std::istream& operator>>(std::istream &is, 
                                  CoordinateTransform3D& p){
    std::string temp;
    is>>temp>>p.Q_>>temp>>p.C_>>temp>>p.scale_;
    return is;
  }
  
  
} // end namespace 


#include <Base/Common/BIASpragmaEnd.hh>

#endif 

