/* This file is part of the BIAS library (Basic ImageAlgorithmS).
   
   Copyright (C) 2003, 2004    (see file CONTACTS for details)
   Multimediale Systeme der Informationsverarbeitung
   Institut fuer Informatik
   Christian-Albrechts-Universitaet Kiel

   BIAS is free software; you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as published by
   the Free Software Foundation; either version 2.1 of the License, or
   (at your option) any later version.
   
   BIAS is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public License
   along with BIAS; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */
#ifndef __BIAS_ProjectionParametersSphericalSimple_hh__
#define __BIAS_ProjectionParametersSphericalSimple_hh__

#include <bias_config.h>

//including BIASpragmaEnd at end of file implies including BIASpragmaStart at first
#include <Base/Common/BIASpragmaStart.hh>

#include <Geometry/ProjectionParametersBase.hh>


namespace BIAS {

  /** @class ProjectionParametersSphericalSimple
      @brief projection parameters 
      camera parameters which define the mapping between rays in the 
      camera coordinate system and pixels in the image as well as external pose.
      the parameters phiOffset and thetaOffset are considered to be intrinsic 
      parameters describing the rotation of the principal point
      the parameters dPhi and dTheta describe the resolution of a pixel in
      radian per pixel.
      Distortion is ignored.
      This class is intended to describe spherical rectification results. */
  class BIASGeometry_EXPORT ProjectionParametersSphericalSimple
    : public ProjectionParametersBase 
  {
  public:

    ProjectionParametersSphericalSimple();
    
    ProjectionParametersSphericalSimple(const double phiOffs,
                                        const double thetaOffs,
                                        const double dPhi,
                                        const double dTheta,
                                        const unsigned int width,
                                        const unsigned int height);

    ProjectionParametersSphericalSimple(const 
                                        ProjectionParametersSphericalSimple& P)
      { *this=P; }
    
    virtual ~ProjectionParametersSphericalSimple() {};
    
    /** @brief position of image on sphere in phi direction
        can also considered to be rotation of the principal point. */
    virtual double GetPhiOffset() const { return phiOffset_;};
    
    /** @brief position of image on sphere in theta direction  */
    virtual double GetThetaOffset() const { return thetaOffset_;};

    /** @brief set resolution in radian/pixel for phi direction*/
    virtual double GetDPhi() const { return dPhi_;};

    /** @brief set resolution in radian/pixel for theta direction*/
    virtual double GetDTheta() const { return dTheta_;};

    virtual void Print(std::ostream& os = std::cout) const;

  
    /** @brief calculates the projection of a point 
        in the local camera coordinate system  
        to a pixel in the image plane of the camera
        
        In the simplest case perspective pinhole projection x = K * point  
        where point is transformed of X using  point = (R^T | -R^T C) X */
    virtual HomgPoint2D ProjectLocal(const Vector3<double>& point,
                                     bool IgnoreDistortion = false) const;
    
    /** @brief calculates the unit size viewing ray from the camera center 
        (in the camera coordinate system) which belongs to the image position
        pos e.g. pos=principal point results in (0,0,1)=optical axis */
    virtual Vector3<double> UnProjectLocal(const HomgPoint2D& pos,
                                           bool IgnoreDistortion = false)
      const;

    /** @brief map points from image onto unit diameter image plane in 3D.*/
    virtual HomgPoint3D UnProjectToImagePlane(const HomgPoint2D& pos,
                                              const double& depth = 1.0,
                                              bool IgnoreDistortion = false) 
      const;

    virtual void Distort(BIAS::HomgPoint2D& point2d) const {}; 

    virtual void Undistort(BIAS::HomgPoint2D& point2d) const {};

    virtual ProjectionParametersBase* Clone() const
    { return new ProjectionParametersSphericalSimple(*this); };

    void Rescale(float ratio);

    /** Checks if 3D point projects into specified image and returns
     *  belonging 2D image point.
     *  @param localX assumes homogenized point! */
    virtual bool 
      DoesPointProjectIntoImageLocal(const Vector3<double>& localX,
                                     HomgPoint2D& x,bool IgnoreDistort) const;
  
    /** Transform from local camera centric reference system to pixel coo */
    inline void 
      Local2Pixel(const Vector3<double>& p2d, Vector2<double>& pixel) const; 

    /** Transform from a semilocal reference frame to pixel coo.
        The semilocal reference  frame is defined to have its origin at 
        the camera center and to have the same orientation as the full
        global reference frame */
    inline void 
    GlobalOrientation2Pixel(const Vector3<double>& p2d, 
                            Vector2<double>& pixel) const;

    /** Transform from pixel coo to local reference frame */
    inline void 
      Pixel2Local(const Vector2<double>& pixel, Vector3<double>& p2d) const; 

    /** Transform from pixel coo to semilocal reference frame. 
        The semilocal reference  frame is defined to have its origin at 
        the camera center and to have the same orientation as the full
        global reference frame */
    inline void
      Pixel2GlobalOrientation(const Vector2<double>& pixel, 
                              Vector3<double>& p2d) const;

    /** compute the angles belonginng to pixel coos */
    inline void 
      Pixel2Angles(const Vector2<double>& pixel, double &phi, 
                   double &theta) const;

#ifdef BIAS_HAVE_XML2
    /** @brief specialization of XML block name function */
    virtual int 
      XMLGetClassName(std::string& TopLevelTag, double& Version) const;

    /** @brief specialization of XML write function */
    virtual int XMLOut(const xmlNodePtr Node, XMLIO& XMLObject) const;

    /** @brief specialization of XML read function */
    virtual int XMLIn(const xmlNodePtr Node, XMLIO& XMLObject); 
#endif

    ///////////////////
    // ensure that R_ stays consistent with Pose_
    /////////////////

    virtual void SetPoseParametrization(const PoseParametrization& pp);

    virtual void SetQ(const BIAS::Quaternion<double>& Q);

    virtual void SetQC(const BIAS::Quaternion<double>& Q,
                       const BIAS::Vector3<double>& C);

    virtual void SetPose(const BIAS::Pose pose);
    
    virtual void SetR(const BIAS::RMatrix& R);

    virtual void ValidatePose();

    RMatrixBase GetRCache() const
    { return R_; }

    /// friend functions
    friend BIASGeometry_EXPORT std::ostream& 
      operator<<(std::ostream &os, 
                 const ProjectionParametersSphericalSimple& p);

  protected:
    double phiOffset_;
    double thetaOffset_;
    double dPhi_;
    double dTheta_; 
    /// a cache for the orientation of the camera
    BIAS::RMatrixBase R_;
  };
 
  BIASGeometry_EXPORT std::ostream& 
  operator<<(std::ostream &os, const ProjectionParametersBase& p);

  /////////////////////////////////////////
  // inline code
  /////////////////////////////////////////
  

  void ProjectionParametersSphericalSimple::
  Local2Pixel(const Vector3<double>& p2d, Vector2<double>& pixel) const
  {
    // phi 
    if (p2d[2]<0.0){
      pixel[0] = (atan2(-p2d[0], -p2d[2]) - phiOffset_) / dPhi_;
      pixel[1] = (atan2(-p2d[1], -p2d[2]) - thetaOffset_) / dTheta_;
    } else {
      pixel[0] = (atan2(p2d[0], p2d[2]) - phiOffset_) / dPhi_;
      pixel[1] = (atan2(p2d[1], p2d[2]) - thetaOffset_) / dTheta_;
    }
  }
  
  void ProjectionParametersSphericalSimple::
  GlobalOrientation2Pixel(const Vector3<double>& p2d, 
                          Vector2<double>& pixel) const
  {
    BIASASSERT(QValid_);
    BIAS::Vector3<double> tmp;
    R_.TransposedMult(p2d, tmp);
    Local2Pixel(tmp, pixel);
  }

  void ProjectionParametersSphericalSimple::
  Pixel2Local(const Vector2<double>& pixel, Vector3<double>& p2d) const
  {
    //std::cout << "new: phi "<<pixel[0]*dPhi_ + phiOffset_<<", theta "<<pixel[1]*dTheta_ + thetaOffset_<<std::endl;
    p2d.Set(tan(pixel[0]*dPhi_ + phiOffset_), 
            tan(pixel[1]*dTheta_ + thetaOffset_), 1.0);
    //std::cout << "new: local: "<<p2d<<std::endl;
  }

  void ProjectionParametersSphericalSimple::
  Pixel2GlobalOrientation(const Vector2<double>& pixel, 
                          Vector3<double>& p2d) const
  {
    BIASASSERT(QValid_);
    Vector3<double> tmp;
    Pixel2Local(pixel, tmp);
    R_.Mult(tmp, p2d);
    //std::cout << "new: R: "<<R_<<std::endl;
    //std::cout << "new: global: "<<p2d<<std::endl;
  }

  void ProjectionParametersSphericalSimple::
  Pixel2Angles(const Vector2<double>& pixel, double &phi, double &theta) const
  {
    phi =   pixel[0] * dPhi_ + phiOffset_;
    theta = pixel[1]*dTheta_ + thetaOffset_;
  }

} // end namespace 


#include <Base/Common/BIASpragmaEnd.hh>

#endif 

