/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#ifndef __BIAS_SimilarityTransformation3D_hh__
#define __BIAS_SimilarityTransformation3D_hh__

// never include this unless it is absolutely needed:
// #include "bias_config.h" 

#include <Base/Common/BIASpragmaStart.hh>
#include <Base/Debug/Debug.hh>
#include <Base/Geometry/Quaternion.hh>
#include <Base/Math/Matrix4x4.hh>
#include <Base/Geometry/HomgPoint3D.hh>
#include <Base/Math/Vector3.hh>

namespace BIAS {


  /** @class SimilarityTransformation3D
   *  @brief calss rpresenting a similiraty transform, 
   *   that can either be a coordinate transform or a movement of points
   *
   *  This transform is identical to the following matrix
   *
   *                [                C_[0] ]
   *                [   scale_*R     C_[1] ]
   *         T  =   [                C_[2] ]
   *                [  0    0    0     1   ]
   *
   *  The internal representation is as quaternion, scale and 
   *  translation(center).
   *  May be used to represent a pose, a movement, or a coordinateTransform.
   *  @author grest 05/2006
   */
  class BIASGeometry_EXPORT SimilarityTransform : public BIAS::Debug
  {
  public:
    /** @brief sets identity transform with zero covariance */
    SimilarityTransform()
      : Q_(0.0, 0.0, 0.0, 1.0), 
        C_(0.0,0.0,0.0), scale_(1.0) {}

    /** @brief calls operator= */
    SimilarityTransform(const SimilarityTransform& c)
    {
      *this = c;
    }

    /** @brief copy all elements */
    const SimilarityTransform& 
    operator=(const SimilarityTransform& c) {
      Q_= c.Q_;
      C_ = c.C_;
      scale_ = c.scale_;
      return (*this);
    }
      
    inline void SetQ(const Quaternion<double>& Q) { Q_ = Q; }

    /** Sets the quaternion 
     */
    inline void SetR(const BIAS::RMatrixBase &R) 
    { R.GetQuaternion(Q_); }
   
    inline void SetC(const Vector3<double>& C) { C_ = C; }

 
    void Set(const Quaternion<double>& Q, const Vector3<double>& C, 
             const double& scale = 1.0) {
      Q_ = Q; C_ = C; scale_ = scale; ;
    }
    
    const Quaternion<double>& GetQ() const { return Q_; }

    const Vector3<double>& GetC() const { return C_; }

    /** returns the corresponding 4x4 matrix
     */
    Matrix4x4<double> GetMatrix() const;

    inline BIAS::RMatrixBase GetR() const {
      RMatrixBase R; 
      R.SetFromQuaternion(GetQ());
      return R;
    }

    friend std::ostream& operator<<(std::ostream &os, 
                                    const SimilarityTransform& p);
    friend std::istream& operator>>(std::istream &is, 
                                    SimilarityTransform& p);
    
    bool Load(const std::string& inputFile);
    bool Save(const std::string& outputFile);

    /** Inverts the transform
     */
    inline void InvertIP() { 
      Q_.Invert(); scale_ = 1.0/scale_; 
      C_ = -scale_ * Q_.MultVec(C_); }

    /** result = this * arg
     */
    inline void Mult(const SimilarityTransform &arg,
              SimilarityTransform &result)
    {
      Q_.Mult(arg.Q_,result.Q_);
      Q_.MultVec(arg.C_, result.C_);
      result.C_ *= scale_;
      result.C_+= C_;
      result.scale_= scale_ * arg.scale_;
    }
 
    inline void MultVec(const Vector3<double> &vec,
                        Vector3<double> &result) const
    { Q_.MultVec(vec,result); result*= scale_; result+= C_; } 

    /**
     * Prints Transformation together with more intuitive rotation 
     * representations.
     */
    void Dump();

  protected:

    /// this is the rotation 
    Quaternion<double> Q_;

    /// the translation 
    Vector3<double> C_;
 
    double scale_;
  };

  inline std::ostream& operator<<(std::ostream &os, 
                                  const SimilarityTransform& p) {
    os<<"Q_= "<<p.Q_<<" C_= "<<p.C_<<" scale_= "<<p.scale_;
    return os;  
  }
  
  inline std::istream& operator>>(std::istream &is, 
                                  SimilarityTransform& p){
    std::string temp;
    is>>temp>>p.Q_>>temp>>p.C_>>temp>>p.scale_;
    return is;
  }
  
  inline SimilarityTransform operator*(SimilarityTransform &arg1,
                                SimilarityTransform &arg2)
  {
    SimilarityTransform result;
    arg1.Mult(arg2,result);
    return result;
  }

  inline BIAS::Vector3<double> operator*(const SimilarityTransform &arg1,
                                  const BIAS::Vector3<double> &vec)
  {
    BIAS::Vector3<double> result;
    arg1.MultVec(vec,result);
    return result;
  }

} // end namespace 


#include <Base/Common/BIASpragmaEnd.hh>

#endif 

