#ifndef __ImageCanvas_hh__
#define __ImageCanvas_hh__

#include <bias_config.h>

#ifndef BIAS_HAVE_WXWIDGETS
#  error BIAS_HAVE_WXWIDGETS not defined but required by ImageCanvas. Please enable BIAS_HAVE_WXWIDGETS in CMake configure step.
#endif

#ifndef BIAS_HAVE_PTHREADS
#  error BIAS_HAVE_PTHREADS not defined. But BIAS::ImageCanvas needs pthreads (on Unix and Win32). Please enable USE_PTHREADS in CMake configure step.
#endif

#include <wx/wx.h>
#include <wx/spinctrl.h>
#include <wx/textctrl.h>
#include <wx/event.h>
#include <wx/dcbuffer.h>
#include <Image/PyramidImage.hh>
#include <Base/Debug/Debug.hh>
#include <pthread.h>
#include <Gui/StringConv.hh>
#include <iostream>

#define D_IC_IM (1<<0)


namespace BIAS{

  /** @brief used by ImageCanvas and ScaledImageCanvas, should not be used
      directly
      \todo Create a common base class for ImageCnvas and ImageCanvasGL and derive both classes from it (JW)
      @author woelk */

  /** @class Histogramm_float_Entry
      @brief base class for storing computed histogramm information
      in HistoCanvas */
  class Histo_Entry_Float {
  public:
    //Interval ends
    float value_begin;
    float value_end;
    //Number of values in this interval
    int   value_count;
  };

  //void ZoomImageCanvas::OnMouseLeave(wxMouseEvent &event)

  //-------------------------------------------------------
  //----------------------ZOOM CANVAS----------------------

  class BIASGui_EXPORT ZoomImageCanvas : public wxPanel
  {
  public:
    ZoomImageCanvas(wxWindow* parent, wxStatusBar *sbar=NULL, int sbarid=0,
                    wxWindowID id = -1, const wxPoint& pos = wxDefaultPosition,
                    const wxSize& size = wxSize(320, 240),
                    long style = wxTAB_TRAVERSAL,
                    const wxString& name = wxT("ZoomImageCanvas"));
    ~ZoomImageCanvas();

    void NewImage(BIAS::ImageBase& im, BIAS::Image<unsigned char>& ucim);
    void NewCoo(unsigned x, unsigned y);
    void NewZoom(int zoom);

    /**
     * Canvas now uses double buffering to avoid flickering, so it should be ok
     * to always clear the canvas.
     */
    inline void SetClearDrawingContext(bool clear)
    { ClearDrawingContext_=clear; }

  protected:
    void UpdateStatusBar(unsigned x, unsigned y);
    void GetZoomImage();

    void OnResize(wxSizeEvent & event);
    void OnPaint(wxPaintEvent & event);
    void OnMouseMove(wxMouseEvent &event);
    void OnMouseLeave(wxMouseEvent &event);
    void OnRightMouse(wxMouseEvent &event);
    void OnLeftMouse(wxMouseEvent &event);
    void OnKey(wxKeyEvent &event);

    wxBitmap *Bitmap_;
    wxImage *wximage_;
    BIAS::Image<unsigned char> ucim_, zim_;
    BIAS::ImageBase origim_;
    wxStatusBar *SBar_;
    int SBarID_;

    // Canvas now uses double buffering to avoid flickering, so it should be ok
    // to always clear the canvas.
    bool ClearDrawingContext_;

    unsigned ZoomFactor_, ZoomX_, ZoomY_; // zoom coo, set from NewCoo
    // upper left corner of ZoomX_, ZoomY_ in Window_
    unsigned ZoomXPos_, ZoomYPos_;
    int wwidth_, wheight_; // width & height of Window_
    int mousex_, mousey_;

    pthread_mutex_t *wxbitmap_mutex_, *origim_mutex_;

    DECLARE_EVENT_TABLE()
      }; // class ZoomImageCanvas

  //-------------------------------------------------------
  //-----------------------ZOOM FRAME----------------------

  /** @brief used by ImageCanvas and ScaledImageCanvas, should not be used
      directly
      @author woelk */
  class BIASGui_EXPORT ZoomFrame : public wxFrame
  {
    enum { MIA_ZoomQuitButton };
  public:
    ZoomFrame(wxWindow* parent, const wxString& title);
    ~ZoomFrame();

    void NewImage(BIAS::ImageBase& im, BIAS::Image<unsigned char>& ucim,
                  const wxString& name);
    void NewCoo(unsigned x, unsigned y);

    void SetZoom(unsigned z);

  protected:
    void Setup();

    void OnQuit(wxCommandEvent & event);
    void OnClose(wxCloseEvent & event);
    void OnZoomFactor(wxScrollEvent & event);
    //void OnKeyPress(wxKeyEvent &event);

    wxString ImageName_;
    wxString Title_;
    wxButton *QuitButton_;
    ZoomImageCanvas *Window_;
    wxSlider *SLScale_;

    DECLARE_EVENT_TABLE()
      }; // class ZoomFrame

  //-------------------------------------------------------
  //-----------------HISTOGRAMM CANVAS---------------------

  class BIASGui_EXPORT HistoImageCanvas : public wxPanel
  {
  public:
    HistoImageCanvas(wxWindow* parent, wxStatusBar *sbar=NULL, int sbarid=0,
                     wxWindowID id = -1, 
                     const wxPoint& pos = wxDefaultPosition,
                     const wxSize& size = wxSize(272, 216),
                     long style = wxTAB_TRAVERSAL,
                     const wxString& name = wxT("HistoImageCanvas"));
    ~HistoImageCanvas();

    void UpdateHistogramm(ImageBase image);
    void NewImage(BIAS::ImageBase& im);
    bool SameName(wxString name, bool settothenewone);
    void RefitHistogramm();

  protected:

    void OnMouseLeave(wxMouseEvent&);
    void OnMouseMove(wxMouseEvent &event);

    void UpdateStatusBar(unsigned mouseX, unsigned mouseY);

    void OnResize(wxSizeEvent & event);
    void OnPaint(wxPaintEvent& event);

    wxBitmap                   *Bitmap_;
    wxImage                    *wximage_;
    BIAS::Image<unsigned char>  ucim_, zim_;
    BIAS::ImageBase             origim_;
    wxStatusBar                *SBar_;

    //indicates how wide should be 1 step of histogramm, when painted
    //(in pixels), is set automatically when the histoframe is resized
    wxCoord                     histoZoom_;

    int SBarID_;

    unsigned int mousex_, mousey_;

    int wwidth_, wheight_; // width & height of Window_

    pthread_mutex_t            *wxbitmap_mutex_, *origim_mutex_;

    //----HISTOGRAMMS----
    //Histogramm type:
    // -1 - no histogramm computed
    //
    //  0 - unsigned char
    //  1 - floating point (float)
    int                         histo_type;

    //Histogramm for float-images for each channel
    //std::vector<std::vector<Histo_Entry_Float> > histo_float;
    std::vector<std::vector<Histo_Entry_Float> > histo_float;
    int                              histo_float_zero;
    int                              histo_float_steps;

    float                            histo_float_min;
    float                            histo_float_step;

    //Histogramm for unsigned char-images for each channel
    std::vector<std::vector<int> >   histo_UC;
   
    int                              histo_channels;
    
    std::vector<std::vector<float> > histo_percentage;
    std::vector<float>               histo_percentage_biggest;

    //count of pixels in an image with histogramm
    int                         histo_size;
    //-------------------
    wxString                    histo_name;

    DECLARE_EVENT_TABLE()
      }; // class HistoImageCanvas

  //-------------------------------------------------------
  //-----------------HISTOGRAMM FRAME----------------------
  class BIASGui_EXPORT HistoFrame : public wxFrame
  {
    enum { MIA_HistoQuitButton };
  public:
    HistoFrame(wxWindow* parent, const wxString& title, 
               wxPoint pos = wxPoint(10,10),
               wxSize size = wxSize(300, 250));
    ~HistoFrame();

    void NewImage(BIAS::ImageBase& im,
                  const wxString& name);

  protected:
    void Setup(wxSize size = wxSize(300, 250));
    void OnSize(wxSizeEvent& event);

    void OnQuit(wxCommandEvent & event);
    void OnClose(wxCloseEvent & event);

    wxString          ImageName_;
    wxString          Title_;
    wxButton         *QuitButton_;
    HistoImageCanvas *Window_;

    DECLARE_EVENT_TABLE()
      }; // class HistoFrame

  //-------------------------------------------------------
  //-------------------------------------------------------

  /** @class ImageCanvas
      @ingroup g_gui
      @brief display image in wx application, provides zoom and investigation
      functionality
      @author woelk */

  class BIASGui_EXPORT ImageCanvas : public wxScrolledWindow , public BIAS::Debug
  {
  public:
    ImageCanvas(wxWindow* parent, wxStatusBar* sbar=NULL,
                int sbarid=0, wxWindowID id = -1,
                const wxPoint& pos = wxDefaultPosition,
                const wxSize& size = wxSize(320, 240),
                long style = wxHSCROLL | wxVSCROLL & wxTRANSPARENT,
                const wxString& name = wxT("ScrolledWindow"));
    ~ImageCanvas();


    void Show(BIAS::Image<unsigned char>& im, std::string name);
    void Show(BIAS::ImageBase& origim, BIAS::Image<unsigned char>& scaledrgbim,
              std::string name);
    void Show(BIAS::PyramidImage<unsigned char>& pim, std::string name);

    //returns the position of the histogramm window
    wxPoint GetHistoPos();
    //returns the size of the histogramm window
    wxSize GetHistoSize();
    //The Method that shows Histogramm-Window if it's not visible yet
    void ShowHistogramm(wxPoint pos  = wxPoint(10,10),
                        wxSize  size = wxSize(100,100));
    void ShowHistogramm(ImageBase& image,
                        wxPoint pos  = wxPoint(10,10),
                        wxSize  size = wxSize(100,100));

    // to let user access classic hide/show function
    virtual bool Show(bool show=true) {return wxScrolledWindow::Show(show);};

    inline BIAS::ImageBase *GetOrigImage() { return &origim_; }

    bool GetMousePos(int &x, int &y);

    inline void SetDefaultZoom(unsigned z)
    { DefaultZoom_=z; }

    /**
     * Canvas now uses double buffering to avoid flickering, so it should be ok
     * to always clear the canvas.
     */
    inline void SetClearDrawingContext(bool clear)
    { ClearDrawingContext_=clear; }

    inline void SetShowScaled(bool on){ShowScaled_ = on;};
    inline bool GetShowScaled(){return ShowScaled_;};

    /** Sets the size of the scaled image. Only used if show scaled is on. */
    inline void SetScaledImageSize(int width, int height)
    {
      ScaledImageSizeX_ = width;
      ScaledImageSizeY_ = height;
    }

    bool HistogrammIsShown();

  protected:
    void InitHisto(wxPoint pos  = wxPoint(10,10),
                   wxSize  size = wxSize(100,100));
    void UpdateStatusBar(unsigned x, unsigned y);
    void Zoom(unsigned x, unsigned y);

    void OnPaint(wxPaintEvent& event);

    void OnErase(wxEraseEvent& event);

    void OnSize(wxSizeEvent& event);
    void OnMouseMove(wxMouseEvent &event);
    void OnMouseLeave(wxMouseEvent &event);
    void OnRightMouseButton(wxMouseEvent &event);
    //void OnKeyPress(wxKeyEvent &event);

    wxBitmap *Bitmap_;
    wxImage *wximage_;
    wxStatusBar *SBar_;
    int SBarID_;
    ZoomFrame *MyZoomWin_;
    HistoFrame *MyHistoWin_;
    unsigned width_, height_;
    int mousex_, mousey_;
    BIAS::ImageBase origim_;
    BIAS::Image<unsigned char> rgbim_;

    // for scaled-down display
    bool ShowScaled_;
    int ScaledImageSizeX_,ScaledImageSizeY_;

    pthread_mutex_t *wxbitmap_mutex_, *origim_mutex_;

    unsigned DefaultZoom_;
    
    // Canvas now uses double buffering to avoid flickering, so it should be ok
    // to always clear the canvas.
    bool ClearDrawingContext_;

    std::string ImageName_;
    wxString Title_;

    DECLARE_EVENT_TABLE()
      }; // ImageCanvas


  /** @class ScaledImageCanvas
      @ingroup g_gui
      @brief display float images in wx application, provides zoom, scale,
      shift  and investigation functionality
      @author woelk */
  class BIASGui_EXPORT ScaledImageCanvas : public wxPanel
  {
  public:
    ScaledImageCanvas(wxWindow* parent, wxStatusBar* sbar=NULL,
                      int sbarid=0, wxWindowID id = -1,
                      const wxPoint& pos = wxDefaultPosition,
                      const wxSize& size = wxDefaultSize,
                      long style = wxTAB_TRAVERSAL,
                      const wxString& name = wxT("ScaledScrolledWindow"));
    ~ScaledImageCanvas();

    void Show(BIAS::ImageBase& im, std::string name);

    void Show(BIAS::PyramidImage<float>& pim, std::string name);

    // to let user access classic hide/show function
    virtual bool Show(bool show=true) {return MiaScWi_->Show(show);};

    inline void SetScaleRange(double min, double max)
    { BIASASSERT(max>min); MinScale_=min; MaxScale_=max; }

    inline void SetOffsetRange(double min, double max)
    { BIASASSERT(max>min); MinOffSet_=min; MaxOffSet_=max; }

    int SetScale(double scale);
    int SetScale(double scale, bool update);

    int SetOffset(double offset);
    int SetOffset(double offset, bool update);

    inline void SetDefaultZoom(unsigned z)
    { MiaScWi_->SetDefaultZoom(z); }

    /** @return  (pointer) to class member ImageCanvas used for display JW */
    ImageCanvas* GetImageCanvas();

    /** @return datat image origim_ class member JW */
    BIAS::ImageBase & GetOrigimg();

    /** pointer to sbar to be used JW */
    void SetStatusBarPtr(wxStatusBar* p_sbar);

    /** display text to sbar if one is assigned */
    void SetStatusText(const std::string & s);

    /** recalculate scale and shift for each new frame or not ?*/
    void SetAutoadaptScale(const bool b){AutoadaptScale_ = b;}

    /** set ranges so that they would look beautiful **/
    void SetNiceRanges(double scale_factor, 
                       double offset_factor);

    /** set ranges of sliders **/
    void SetRanges(double value_sc_double_l, 
                   double value_sc_double_r,
                   double value_of_double_l,
                   double value_of_double_r);

    void UpdateHistogramm(ImageBase image);
      
  protected:
    void OnScaleOffset       (wxScrollEvent & event);
    void OnScaleOffsetSteps  (wxSpinEvent & event);
    void OnScaleOffsetBorders(wxCommandEvent & event);

    void OnAutoScaleOffset   (wxCommandEvent& event);
    void OnHisto             (wxCommandEvent& event);

    void UpdateScaleOffset();
    void UpdateScaleOffsetBorders();

    wxCheckBox*   Check_Auto_Sc_Of_;
    wxCheckBox*   Check_Histo_;

    //Controls to change Scale-Slider Parameters
    wxStaticText* Scale_Text_;

    wxTextCtrl*   Scale_L_;
    wxTextCtrl*   Scale_R_;
    wxSpinCtrl*   Scale_Steps_;

    double        scale_l_old;
    int             scale_old;
    double        scale_r_old;

    //Controls to change Offset-Slider Parameters
    wxStaticText* Offset_Text_;

    wxTextCtrl*   Offset_L_;
    wxTextCtrl*   Offset_R_;
    wxSpinCtrl*   Offset_Steps_;

    double        offset_l_old;
    int             offset_old;
    double        offset_r_old;

    //Sizes of text/spin/checkboxes
    wxSize box_size;
    wxSize box_size_steps;
    wxSize box_size_checks;
    
    //Count of steps in sliders
    int    slider_sc_digits;
    int    slider_of_digits;

    //Old Values of sliders, text/spin/checkboxes
    wxString scale_border_l_value;
    wxString scale_border_r_value;

    wxString offset_border_l_value;
    wxString offset_border_r_value;
    

    wxString string_base;

    // data member
    ImageCanvas *MiaScWi_; // true data image
    wxSlider *ScaleSL_, *OffsSL_;
    BIAS::ImageBase origim_;
    BIAS::Image<float> scaledf_;
    BIAS::Image<unsigned char> rgbim_, ucim_;
    std::string ImageName_;
    double MinScale_, MaxScale_, MinOffSet_, MaxOffSet_;

    // pointer to statusbar to use if any
    wxStatusBar * p_StatBar;
    bool AutoadaptScale_;
    DECLARE_EVENT_TABLE()
      }; // ScaledImageCanvas


} // namespace


#endif // __ImageCanvas_hh__
