#ifdef WIN32
#  pragma warning (disable: 4005) // for VS8 + WX
#endif

#include <Gui/ProjectionGUI.hh>
#include <Geometry/ProjectionParametersSpherical.hh>
#include <Gui/StringConv.hh>
#include <sstream>

using namespace std;
using namespace BIAS;

/*
===============================================================================
VectorPanel
===============================================================================
*/

VectorPanel::VectorPanel(wxWindow* parent, wxWindowID id, bool editable)
  : wxPanel(parent, id)
{
  editable_ = editable;
}

void VectorPanel::SetVector(const Vector<double>& vector)
{
  // need to re-create controls?
  if (vector.Size() != controls_.size())
  {
    DestroyChildren();
    wxSizer* sizer = new wxFlexGridSizer(vector.Size(), 1);
    SetSizer(sizer);

    for (int i = 0; i < (int)vector.Size(); i++)
    {
      wxTextCtrl* control = new wxTextCtrl(this, 
					   GetId()
                                           ,_T("-1"), 
                                           wxDefaultPosition, 
                                           wxDefaultSize, 
                                           wxTE_PROCESS_ENTER

);
      control->SetEditable(editable_);
      sizer->Add(control);
      controls_.push_back(control);
    }
  }

  // fill controls
  for (int i = 0; i < (int)vector.Size(); i++)
  {
    //instead of using this blocking technique when focused
    //use TEXT_ENTER event which occures when enter is pressed
    //within the TextCtr.
    //if (FindFocus() == controls_[i]) continue;
  
    stringstream str;
    str << vector[i];
    //cout<<" "<<str.str()<<endl;
    controls_[i]->SetValue(AsciiToWx(str.str()));
  }
}

Vector<double> VectorPanel::GetVector() const
{
  Vector<double> v(controls_.size());
  for (int i = 0; i < (int)controls_.size(); i++)
  {
    stringstream str(WxToAscii(controls_[i]->GetValue()));
    str >> v[i];
  }
  return v;
}

/*
===============================================================================
MatrixPanel
===============================================================================
*/

MatrixPanel::MatrixPanel(wxWindow* parent, wxWindowID id, bool editable)
  : wxPanel(parent, id)
{
  editable_ = editable;
}

void MatrixPanel::SetMatrix(const Matrix<double>& matrix)
{
  // need to re-create controls?
  if ((int)matrix.GetRows() != rows_ || (int)matrix.GetCols() != cols_)
  {
    DestroyChildren();
    wxSizer* sizer = new wxFlexGridSizer(matrix.GetRows(), matrix.GetCols());
    SetSizer(sizer);

    rows_ = matrix.GetRows();
    cols_ = matrix.GetCols();

    for (int i = 0; i < rows_; i++)
    {
      for (int j = 0; j < cols_; j++)
      {
        wxTextCtrl* control = new wxTextCtrl(this, GetId());
        control->SetEditable(editable_);
        sizer->Add(control);
        controls_.push_back(control);
      }
    }
  }

  // fill controls
  for (int i = 0; i < rows_; i++)
  {
    for (int j = 0; j < cols_; j++)
    { 
      
      /** There is definitely a bug here,
          sometimes it goes beyond controls_.size()
          ischiller 18/07/07
      */
      if (((i*cols_+j)< (int)controls_.size()) &&
          FindFocus() == controls_[i*cols_+j]){
        
        stringstream str;
        str << matrix.GetDataArray()[i][j];
       controls_[i*cols_+j]->SetValue(AsciiToWx(str.str()));
      }
    }
  }
}

Matrix<double> MatrixPanel::GetMatrix()
{
  Matrix<double> matrix(rows_, cols_);
  for (int i = 0; i < rows_; i++)
  {
    for (int j = 0; j < cols_; j++)
    {
      stringstream str(WxToAscii(controls_[i*cols_+j]->GetValue()));
      str >> matrix[i][j];
    }
  }
  return matrix;
}

/*
===============================================================================
ProjectionGUI
===============================================================================
*/

// event ids
enum
{
  ID_New,
  ID_Load,
  ID_Save,
  ID_NormalizeQuaternion
};

// panel ids
enum
{
  ID_Center,
  ID_Quaternion,
  ID_RMatrix,
  ID_ImageSize,
  ID_AspectRatio,
  ID_Principal,
  ID_KMatrix,
  ID_FocalLength,
  ID_Undistortion,
  ID_Radius
};

BEGIN_EVENT_TABLE(ProjectionGUI, wxPanel)
  EVT_TEXT_ENTER(-1, ProjectionGUI::OnModified)
  EVT_BUTTON(ID_NormalizeQuaternion, ProjectionGUI::OnNormalizeQuaternion)
  EVT_BUTTON(ID_New,  ProjectionGUI::OnNew)
  EVT_BUTTON(ID_Load, ProjectionGUI::OnLoad)
  EVT_BUTTON(ID_Save, ProjectionGUI::OnSave)
END_EVENT_TABLE()

ProjectionGUI::ProjectionGUI(wxWindow* parent, wxWindowID id,
                             const wxPoint& pos, const wxSize& size,
                             long style, const wxString& name, bool createButtons)
  : wxPanel(parent, id, pos, size, style, name)
{
  panelCenter_        = NULL;
  panelQuaternion_    = NULL;
  panelRMatrix_       = NULL;
  panelImageSize_     = NULL;
  panelAspectRatio_   = NULL;
  panelPrincipal_     = NULL;
  panelKMatrix_       = NULL;
  panelFocalLength_   = NULL;
  panelUndistortion_  = NULL;
  panelRadius_        = NULL;

  selfModifying_ = false;

  projectionPanel_ = new wxScrolledWindow(this);
  projectionPanel_->SetScrollRate(1, 1);

  wxSizer* buttonSizer = new wxBoxSizer(wxHORIZONTAL);

  if (createButtons)
  {
    buttonSizer->Add(new wxButton(this, ID_New,  wxT("New")));
    buttonSizer->Add(new wxButton(this, ID_Load, wxT("Load")));
    buttonSizer->Add(new wxButton(this, ID_Save, wxT("Save")));
  }

  wxSizer* mainSizer = new wxBoxSizer(wxVERTICAL);
  mainSizer->Add(buttonSizer, 0, wxALL, 5);
  mainSizer->Add(projectionPanel_, 1, wxEXPAND | wxALL, 5);

  SetSizer(mainSizer);
}

void ProjectionGUI::SetProjection(const Projection& projection)
{
  // clear old projection panel
  projectionPanel_->DestroyChildren();

  // create sizer
  wxFlexGridSizer* sizer = new wxFlexGridSizer(2, 5, 5);
  sizer->AddGrowableCol(1);
  projectionPanel_->SetSizer(sizer);

  // base parameters
  const ProjectionParametersBase* paramsBase = projection.GetParameters();

  panelCenter_      = AddVectorPanel(wxT("Center"),       ID_Center,      true);
  panelQuaternion_  = AddVectorPanel(wxT("Quaternion"),   ID_Quaternion,  true);

  sizer->AddSpacer(0);
  sizer->Add(new wxButton(projectionPanel_, ID_NormalizeQuaternion, wxT("Normalize Quaternion")));

  panelRMatrix_     = AddMatrixPanel(wxT("R Matrix"),     ID_RMatrix,     false);
  panelImageSize_   = AddVectorPanel(wxT("Image Size"),   ID_ImageSize,   true);
  panelAspectRatio_ = AddVectorPanel(wxT("Aspect Ratio"), ID_AspectRatio, true);
  panelPrincipal_   = AddVectorPanel(wxT("Principal"),    ID_Principal,   true);

  // perspective parameters
  const ProjectionParametersPerspective* paramsPersp =
    dynamic_cast<const ProjectionParametersPerspective*>(paramsBase);
  if (paramsPersp != NULL)
  {
    panelKMatrix_       = AddMatrixPanel(wxT("K Matrix"),     ID_KMatrix,       false);
    panelFocalLength_   = AddVectorPanel(wxT("Focal Length"), ID_FocalLength,   true);
    panelUndistortion_  = AddVectorPanel(wxT("Undistortion"), ID_Undistortion,  true);
  }
  else
  {
    panelKMatrix_       = NULL;
    panelFocalLength_   = NULL;
    panelUndistortion_  = NULL;
  }

  // spherical parameters
  const ProjectionParametersSpherical* paramsSpherical =
    dynamic_cast<const ProjectionParametersSpherical*>(paramsBase);
  if (paramsSpherical != NULL)
  {
    panelRadius_ = AddVectorPanel(wxT("Radius"), ID_Radius, true);
  }
  else
  {
    panelRadius_ = NULL;
  }

  projection_ = projection;

  // fill gui with values from projection
  UpdateGUI();

  sizer->Layout();
  Layout();
}

const Projection& ProjectionGUI::GetProjection() const
{
  return projection_;
}

VectorPanel* ProjectionGUI::AddVectorPanel(const wxString& name,
                                           wxWindowID id, bool editable)
{
  VectorPanel* panel = new VectorPanel(projectionPanel_, id, editable);

  wxSizer* sizer = projectionPanel_->GetSizer();
  sizer->Add(new wxStaticText(projectionPanel_, -1, name));
  sizer->Add(panel, 1, wxEXPAND);

  return panel;
}

MatrixPanel* ProjectionGUI::AddMatrixPanel(const wxString& name,
                                           wxWindowID id, bool editable)
{
  MatrixPanel* panel = new MatrixPanel(projectionPanel_, id, editable);

  wxSizer* sizer = projectionPanel_->GetSizer();
  sizer->Add(new wxStaticText(projectionPanel_, -1, name));
  sizer->Add(panel, 1, wxEXPAND);

  return panel;
}

void ProjectionGUI::UpdateGUI()
{
  selfModifying_ = true;

  // base parameters
  const ProjectionParametersBase* paramsBase = projection_.GetParameters();

  unsigned int imageWidth, imageHeight;
  paramsBase->GetImageSize(imageWidth, imageHeight);

  double principalX, principalY;
  paramsBase->GetPrincipal(principalX, principalY);
  if (paramsBase->CValid())
    panelCenter_->SetVector(paramsBase->GetC());
  if (paramsBase->QValid()) {
    panelQuaternion_->SetVector(paramsBase->GetQ());
    panelRMatrix_->SetMatrix(BIAS::Matrix<double>(paramsBase->GetR()));
  }
  panelImageSize_->SetVector(Vector2<double>(imageWidth, imageHeight));// TODO: this is no vec2!
  panelAspectRatio_->SetVector(Vector<double>(1, paramsBase->GetAspectratio()));
  panelPrincipal_->SetVector(Vector2<double>(principalX, principalY)); // TODO: this is no vec2!

  // perspective parameters
  const ProjectionParametersPerspective* paramsPersp =
    dynamic_cast<const ProjectionParametersPerspective*>(paramsBase);
  if (paramsPersp != NULL)
  {
    double focalLength;
    double kc1, kc2, kc3, kc4;
    paramsPersp->GetFocalLength(focalLength);
    paramsPersp->GetUndistortion(kc1, kc2, kc3, kc4);

    panelKMatrix_->SetMatrix((Matrix<double>)(paramsPersp->GetK()));
    panelFocalLength_->SetVector(Vector<double>(1, focalLength));
    panelUndistortion_->SetVector(Vector4<double>(kc1, kc2, kc3, kc4));
  }

  // spherical parameters
  const ProjectionParametersSpherical* paramsSpherical =
    dynamic_cast<const ProjectionParametersSpherical*>(paramsBase);
  if (paramsSpherical != NULL)
  {
    panelRadius_->SetVector(Vector<double>(1, paramsSpherical->GetRadius()));
  }

  selfModifying_ = false;
}

void ProjectionGUI::OnNormalizeQuaternion(wxCommandEvent& /*event*/)
{
  ProjectionParametersBase* paramsBase = projection_.GetParameters();
  Quaternion<double> Q = paramsBase->GetQ();
  Q.Normalize();
  paramsBase->SetQ(Q);
  UpdateGUI();
}

void ProjectionGUI::OnModified(wxCommandEvent& event)
{
  if (selfModifying_)
  {
    return;
  }

  ProjectionParametersBase* paramsBase = projection_.GetParameters();
  ProjectionParametersPerspective* paramsPersp =
    dynamic_cast<ProjectionParametersPerspective*>(paramsBase);
  ProjectionParametersSpherical* paramsSpherical =
    dynamic_cast<ProjectionParametersSpherical*>(paramsBase);

  switch (event.GetId())
  {
    case ID_Center:
    {
      Vector<double> C = panelCenter_->GetVector();
      paramsBase->SetC(Vector3<double>(C));
    } break;

    case ID_Quaternion:
    {
      Vector<double> Q = panelQuaternion_->GetVector();
      paramsBase->SetQ(Quaternion<double>(Q[0], Q[1], Q[2], Q[3]));
    } break;

    case ID_RMatrix:
    {
      Matrix<double> R = panelRMatrix_->GetMatrix();
      paramsBase->SetR(Matrix3x3<double>(R));
    } break;

    case ID_ImageSize:
    {
      Vector<double> imageSize = panelImageSize_->GetVector();
      paramsBase->SetImageSize((int)imageSize[0], (int)imageSize[1]);
    } break;

    case ID_AspectRatio:
    {
      Vector<double> aspectratio = panelAspectRatio_->GetVector();
      paramsBase->SetAspectratio(aspectratio[0]);
      BIASASSERT(false); // aspect ratio is read-only
    } break;

    case ID_Principal:
    {
      Vector<double> principal = panelPrincipal_->GetVector();
      paramsBase->SetPrincipal(principal[0], principal[1]);
    } break;

    case ID_KMatrix:
    {
      BIASASSERT(false); // K matrix is read-only
    } break;

    case ID_FocalLength:
    {
      Vector<double> focalLength = panelFocalLength_->GetVector();
      paramsPersp->SetFocalLengthAndAspect(focalLength[0],
                                           paramsPersp->GetAspectratio());
    } break;

    case ID_Undistortion:
    {
      Vector<double> undist = panelUndistortion_->GetVector();
      paramsPersp->SetUndistortion(undist[0], undist[1], undist[2], undist[3]);
    } break;

    case ID_Radius:
    {
      Vector<double> radius = panelRadius_->GetVector();
      paramsSpherical->SetRadius(radius[0]);
    } break;
  }

  UpdateGUI();
}

void ProjectionGUI::OnNew(wxCommandEvent& /*event*/)
{
  wxString choices[] = { wxT("Perspective"), wxT("Spherical") };
  int numChoices = 2;

  wxSingleChoiceDialog dlg(NULL, wxT("Choose projection type:"),
                           wxT("New Projection"), numChoices, choices);
  if (dlg.ShowModal() == wxID_OK)
  {
    switch (dlg.GetSelection())
    {
    case 0:
      SetProjection(Projection(ProjectionParametersPerspective()));
      break;
    case 1:
      SetProjection(Projection(ProjectionParametersSpherical()));
      break;
    }
  }
}

void ProjectionGUI::OnLoad(wxCommandEvent& /*event*/)
{
#ifdef BIAS_HAVE_XML2
  wxFileDialog dlg(NULL, wxT("Load Projection"), wxT(""), wxT(""), wxT("*.*"),
                   wxOPEN | wxFILE_MUST_EXIST);
  if (dlg.ShowModal() == wxID_OK)
  {
    Projection loadedProjection;
    if (loadedProjection.Load(WxToAscii(dlg.GetPath())) != 0)
    {
      wxMessageDialog dlg(this, wxT("Failed to load projection"), wxT("Error"));
      dlg.ShowModal();
    }
    else
    {
      SetProjection(loadedProjection);
    }
  }
#else
  BIASERR("XML2 missing");
#endif
}

void ProjectionGUI::OnSave(wxCommandEvent& /*event*/)
{
#ifdef BIAS_HAVE_XML2
  wxFileDialog dlg(NULL, wxT("Save Projection"), wxT(""), wxT(""), wxT("*.*"), wxSAVE);
  if (dlg.ShowModal() == wxID_OK)
  {
    if (GetProjection().XMLWrite(WxToAscii(dlg.GetPath())) != 0)
    {
      wxMessageDialog dlg(this, wxT("Failed to save projection"), wxT("Error"));
      dlg.ShowModal();
    }
  }
#else
  BIASERR("XML2 missing");
#endif
}

/*
===============================================================================
ProjectionDialog
===============================================================================
*/

ProjectionDialog::ProjectionDialog(const Projection& projection, wxWindow* parent,
                                   wxWindowID id, const wxString& title,
                                   bool createButtons)
  : wxDialog(parent, id, title, wxDefaultPosition, wxSize(500, 550),
             wxDEFAULT_DIALOG_STYLE | wxRESIZE_BORDER)
{
  projgui_ = new ProjectionGUI(this, -1,
                  wxDefaultPosition,
                  wxDefaultSize,
                  wxTAB_TRAVERSAL,
                  wxT("panel"),
                  createButtons);
  
  projgui_->SetProjection(projection);

  wxSizer* buttonSizer = new wxBoxSizer(wxHORIZONTAL);
  buttonSizer->Add(new wxButton(this, wxID_OK,     wxT("OK")));
  buttonSizer->Add(new wxButton(this, wxID_CANCEL, wxT("Cancel")));

  wxSizer* mainSizer = new wxBoxSizer(wxVERTICAL);
  mainSizer->Add(projgui_, 1, wxEXPAND | wxALL, 5);
  mainSizer->Add(buttonSizer, 0, wxBOTTOM | wxALIGN_CENTER, 5);

  SetSizer(mainSizer);
}

const Projection& ProjectionDialog::GetProjection() const
{
  return projgui_->GetProjection();
}
