#include "CalibratedPyramid.hh"

#include <Base/Math/Vector2.hh>
#include <Geometry/ProjectionParametersBase.hh>

using namespace BIAS;
using namespace std;

template <class StorageType>
CalibratedPyramid<StorageType>::
CalibratedPyramid()
  : PyramidImage_(), Data_()
{
  const int size = PyramidImage_.Size();
  Data_.resize(size);
  for (int i=0; i<size; i++){
    Data_[i].projection = SharedPtr<ProjectionParametersBase>(NULL);
    Data_[i].image = PyramidImage_[i];
  }
}


template <class StorageType>
CalibratedPyramid<StorageType>::
CalibratedPyramid(const CalibratedPyramid<StorageType>& cp)
  : PyramidImage_(), Data_(PyramidImage_.Size())
{
  operator=(cp);
}


template <class StorageType>
CalibratedPyramid<StorageType>::
CalibratedPyramid(const PyramidImage<StorageType>& cp)
  : PyramidImage_(cp), Data_(cp.size())
{
  const unsigned size = Size();
  for (unsigned i=0; i<size; i++){
    Data_[i].projection = SharedPtr<ProjectionParametersBase>(NULL);
    Data_[i].image = PyramidImage_[i];
  }
}


template <class StorageType>
CalibratedPyramid<StorageType>::
~CalibratedPyramid()
{
  Clear();
}


template <class StorageType>
void CalibratedPyramid<StorageType>::
Init(const CalibratedImage<StorageType>& ci,
     const unsigned pyramid_size, const double rescale_factor)
{
  BIASASSERT(pyramid_size>0);
  Clear();
  PyramidImage_.SetRescaleFactor(rescale_factor);
  PyramidImage_.Init(*(ci.image), pyramid_size);
  PyramidImage_.Downsample();
  Data_.resize(pyramid_size);
  if (ci.projection)
    Data_[0].projection = SharedPtr<ProjectionParametersBase>(ci.projection->Clone());
  else 
    Data_[0].projection = SharedPtr<ProjectionParametersBase>(NULL);
  Data_[0].image = PyramidImage_[0];
  for (unsigned i=1; i<pyramid_size; i++){
    if (Data_[i-1].projection){
      Data_[i].projection = SharedPtr<ProjectionParametersBase>(Data_[i-1].projection->Clone());
      Data_[i].projection->Rescale((float)rescale_factor);
    } else {
      Data_[i].projection = SharedPtr<ProjectionParametersBase>(NULL);
    }
    Data_[i].image = PyramidImage_[i];
  }
}


template <class StorageType>
void CalibratedPyramid<StorageType>::
Clear()
{
  Data_.clear();
  //PyramidImage_.Clear();
}

template <class StorageType>
CalibratedPyramid<StorageType>& CalibratedPyramid<StorageType>::
operator=(const CalibratedPyramid<StorageType>& cp)
{
  Clear();
  const int size = cp.Size();
  PyramidImage_ = cp.PyramidImage_;
  Data_.resize(size);
  for (int i=0; i<size; i++){
    if (cp[i].projection)
      Data_[i].projection = SharedPtr<ProjectionParametersBase>(cp[i].projection->Clone());
    if (cp[i].image)
      Data_[i].image = PyramidImage_[i];
  }
  return *this;
}


template <class StorageType>
void CalibratedPyramid<StorageType>::
CoordinateTransf(const Vector2<double>& coo0, const unsigned index,
                 Vector2<double>& coo_index) const
{
  if (index==0) { coo_index=coo0; return ; }
  coo_index[0] = (coo0[0]-PyramidImage_.GetPositionOffset()) / 
    (PyramidImage_.GetFactors()[index]);
  coo_index[1] = (coo0[1]-PyramidImage_.GetPositionOffset()) / 
    (PyramidImage_.GetFactors()[index]);
// //   coo_index[0] = (coo0[0]+PyramidImage_.GetPositionOffset()) / (PyramidImage_.GetFactors()[index]);
// //   coo_index[1] = (coo0[1]+PyramidImage_.GetPositionOffset()) / (PyramidImage_.GetFactors()[index]);
//   coo_index[0] = coo0[0] / PyramidImage_.GetFactors()[index];
//   coo_index[1] = coo0[1] / PyramidImage_.GetFactors()[index];
}


//////////////////////////////////////////////////////////////////////////
//                 instantiation
//////////////////////////////////////////////////////////////////////////

#define INST(type) template class BIASImage_EXPORT CalibratedPyramid<type>

INST(unsigned char);
INST(float);

#ifdef BUILD_IMAGE_INT
INST(int);
#endif
#ifdef BUILD_IMAGE_CHAR
INST(char);
#endif
#ifdef BUILD_IMAGE_SHORT
INST(short);
#endif
#ifdef BUILD_IMAGE_USHORT
INST(unsigned short);
#endif
#ifdef BUILD_IMAGE_UINT
INST(unsigned int);
#endif
#ifdef BUILD_IMAGE_DOUBLE
INST(double);
#endif
