/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#ifndef __COLOR_MULTI_HISTOGRAM_HH__
#define __COLOR_MULTI_HISTOGRAM_HH__

#include <Base/Math/Math.hh>
#include "ColorHistogram.hh"


// start from big bits, because small bits are used in Image
#define D_COLOR_MULTI_HIST_VEC 0x9000000

namespace BIAS{

  /**
     @class ColorMultiHistogram
     This Class handels MultiHistogram matters on Images.
     it is derived from Colorhistogram, to make it usable just like a 
     normal ColorHistogram 
     It represents a Multihistogram and offers the same Methods as 
     ColorHistogram
     @author Ingo Schiller March 2004
     MultiHistogram is represented by two ColorHistograms 
     @status untested
   */

template <class StorageType>
class BIASImage_EXPORT ColorMultiHistogram 
  {
  public:
    /// 
    ColorMultiHistogram();
    ~ColorMultiHistogram();

  
    /**
       Deletes ALL current Histogram information
    */
    int Clear();

   
    /**
       set the coords for the ROI in first Histogram
    */
    int SetROI1(int ux, int uy, int lx, int ly);

    /**
       set the coords for the ROI in second Histogram
    */
    int SetROI2(int ux, int uy, int lx, int ly);


    /**
       set the binsize (must be divideable by 2pown)
    */
    int SetBinSize(int binsize);

    /**
       set minsat_, which defines the minimum saturation
    */
    inline void SetMinSat(int minsat)
    {minSat_ = minsat;}

    /**
       sets Lambda, which is used to calculate the dist
       in CalcSimilarity
    */
    inline void SetLambda(int lambda)
    {lambda_=lambda;}
   
    /**
       GetBinSize
    */
    inline int GetBinSize() const {return binSize_;}   

    /**
       Get Hist1
       @return the first Histogram
    */
    ColorHistogram<StorageType> GetHist1() const;


    /**
       Get Hist2
       @return the second Histogram
    */
    ColorHistogram<StorageType> GetHist2() const;

    /**
       @returns the distance of the two centers of the ROIs in pixels
       first x coords than y coord
    */
    void GetDistROIs(std::vector<double>& dist) const;

    /**
      calculates the Histogram of the image
      ATTENTION: ROI must be set in upper function SetROI if this function 
      is used
      @param image = the image
      @returns -1 if image does not exist
    */
    int GenerateHist(BIAS::Image<StorageType>& image);

    /** Updates the histogramm with values from hist, 
        where the factor gives the weighting of the new values from hist.
        @return 0 in case of no error
        @author Daniel Grest, March 2004
     */
    int Update(const BIAS::ColorMultiHistogram<StorageType> &hist, 
               double factor=0.5);

    /**
       Calculates the Bhattacharyya similarity Coeiffient^2
       which is the distance or difference of 2 ColorMultiHistograms
       @param the hist to compare with
       @returns double value, which indicates the similarity. 1 if similar.
       @author schiller March 2004
     */
    double CalcSimilarity(const ColorMultiHistogram<StorageType> &chist) const;

     /**
       Calculates the Bhattacharyya similarity Coeiffient^2
       which is the distance or difference of 2 ColorMultiHistograms.
       Same As CalcSimilarity, but the two histogram values are weighted
       according to the amount of color bins which are not zero.
       @param the hist to compare with
       @returns double value, which indicates the similarity. 1 if similar.
       @author grest May 2004
     */
    double CalcSimilarityColorWeighted(const BIAS::ColorMultiHistogram<StorageType> &chist) const;

    /**
       dumps information on the MultiHistogram
       calls the Dump of ColorHistogram
       @param os = the outstream, default std::cout
       @author Schiller
    */
    void Dump(std::ostream& os=std::cout,std::ostream& os2=std::cout);

    /**
       reads a histogram from instream and fills the internal data 
       structures
       @param the filenames of the two histograms
       @author Schiller March 2004
       @returns 0 if ok, -1 if no file
     */
    int Read(char *filename1, char *filename2);

  protected:
     
    // two ColorHistograms representating the Multihistogram 
    BIAS::ColorHistogram<StorageType> hist1_, hist2_;
    //modifieable Parameters for GUI
    int binSize_,minSat_,lambda_;
  
  }; // class


} // namespace BIAS
#endif // __COLOR_MULTI_HISTOGRAM_HH__
