/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
Multimediale Systeme der Informationsverarbeitung
Institut fuer Informatik
Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <Base/Common/BIASpragmaStart.hh>

#ifndef __CylindricalRectification_hh__
#define __CylindricalRectification_hh__

#include <Image/RectificationViaProjectionMappingBase.hh>


namespace BIAS {

#define DEG_TO_RAD 0.0017453292

  /** Class implements rectification by image \b projection onto a cylinder.
   *  ProjectionParametersCylindric are calculated and used to represent
   *  the rectified state. The input projections are for now restricted to 
   *  *Perspective and *Spherical cameras.
   *
   * \ingroup g_filter
   *  \author bartczak 11/2006
   */
  template <class InputStorageType, class OutputStorageType> 
  class BIASImage_EXPORT CylindricalRectification :
    public RectificationViaProjectionMappingBase<InputStorageType, 
                                                 OutputStorageType>
  {
  public:
    virtual ~CylindricalRectification();

    /** 
     * Parameter determines the extend of the cylinder along the cylinder axis.
     * This parameter controls how close optical rays are allowed to point into
     * the directions of the epipoles.
     */
    CylindricalRectification(double maxAngleRAD = 70*DEG_TO_RAD);

    /** only perspective and spherical cameras are allowed */
    virtual bool 
    IsInputCameraValid(const BIAS::Image<InputStorageType>& img,
                       const BIAS::ProjectionParametersBase* proj);
    
  protected:
    virtual int DetermineRectificationParameters_();

    /** @{
     *   Determines the internal parameters of rectifying camera.
     *  Untill now the worst case parameters are determined trying to
     *  achieve no pixel loss.
     * 
     *  \param camera input (original) camera
     *  \param rectPose the base of the (new) rectifying camera
     *  \param minX/maxX length of the needed cylinder length
     *  \param minPhi/maxPhi extend of the cylinder around its axis
     *  \param relativeXRes pixels per length unit along cylinder axis 
     *  \param relativePhiRes pixels per angle unit around cylinder axis 
     */
      int DetermineCylindricCameraBoundries_(
                               const ProjectionParametersBase* camera,
                               const Pose& rectPose,
                               double& minX, double& maxX,
                               double& minPhi, double& maxPhi,
                               double& relativeXRes,
                               double& relativePhiRes) const;


      int DetermineCylindricCameraBoundriesPerspective_(
                                 const ProjectionParametersBase* camera,
                                 const Pose& rectPose,
                                 double& minX, double& maxX,
                                 double& minPhi, double& maxPhi,
                                 double& relativeXRes,
                                 double& relativePhiRes) const;

      int DetermineCylindricCameraBoundriesSpherical_(
                                 const ProjectionParametersBase* camera,
                                 const Pose& rectPose,
                                 double& minX, double& maxX,
                                 double& minPhi, double& maxPhi,
                                 double& relativeXRes,
                                 double& relativePhiRes) const;

      int DetermineCylindricCameraBoundriesCylindric_(
                                 const ProjectionParametersBase* camera,
                                 const Pose& rectPose,
                                 double& minX, double& maxX,
                                 double& minPhi, double& maxPhi,
                                 double& relativeXRes,
                                 double& relativePhiRes) const;

      /** @} */
    /** 
     * The maximally allowed half viewing angle of cylindrical rectification
     * camera. Indirectly determines the minimal angle between cylinder
     * axis and all optical rays.
     */
    double maxHalfAngleInX_;

	//HELPERS
	static int ProjectOntoUnitCylinder_(BIAS::HomgPoint3D& point, double& phi);
	static void SetMinMaxFrom_(BIAS::HomgPoint3D& rectPoint, 
				               double& minX, double& maxX,
								double& minPhi, double& maxPhi);
  };


}//namespace 

#endif
#include <Base/Common/BIASpragmaEnd.hh>

