/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
Multimediale Systeme der Informationsverarbeitung
Institut fuer Informatik
Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <Base/Common/BIASpragmaStart.hh>

#ifndef __HomographyMapping_hh__
#define __HomographyMapping_hh__

#include <Image/BackwardMapping.hh>
#include <Geometry/HMatrix.hh>


namespace BIAS {

  /** @class HomographyMapping
      @ingroup g_filter
      @brief  Maps image src to image sink with homography H (software
      implementation)
      @author koeser, edited by grest Nov. 2005
  */ 
  template <class InputStorageType, class OutputStorageType> 
  class BIASImage_EXPORT HomographyMapping 
    : public  BackwardMapping<InputStorageType, OutputStorageType>  {
  public:

    virtual ~HomographyMapping();

    HomographyMapping();

    /**  required because assignment is 
         @author Jan Woetzel */
    HomographyMapping(const HomographyMapping & src);

    /**  required because of const members 
         @author Jan Woetzel */
    HomographyMapping<InputStorageType, OutputStorageType> & 
    operator=(const HomographyMapping & src);

    /** @brief set your homography H (source = H * sink) before calling Map()*/
    inline void SetHomography(const HMatrix& H) { H_ = H; };

 
       
    /** calculates the bounding box in sink image, 
        where to do the backward mapping.
        The resulting coordinates must be valid in sink, and the resulting 
        br position should be outside the region.
        @author grest, Nov. 2005
     */
    virtual int GetBoundingBox(unsigned int srcwidth, unsigned int srcheight,
                               unsigned int sinkwidth, unsigned int sinkheight,
                               int &tlx, int &tly, 
                               int &brx, int &bry);

    /** if Map() is called with newDist=true, the real applied Homography 
        is changed
     */
    HMatrix GetChangedHomography() const;    

  protected:
    /** @brief reimplementation for homography, takes sink and computes 
        coords in source */
    virtual int GetSourceCoordinates_(const HomgPoint2D& sink,
                                      HomgPoint2D& source) const;
    
    /** @brief analytic jacobian */
    virtual int GetJacobian_(const HomgPoint2D& sink,
                             Matrix2x2<double>& Jacobian) const;

    /// the homography for backward mapping: run over sink and compute
    /// source = H_ * sink
    HMatrix H_;

    using BackwardMapping<InputStorageType, OutputStorageType>::offsetX_;
    using BackwardMapping<InputStorageType, OutputStorageType>::offsetY_;
  };


  //
  //#ifdef WIN32
  //template <>
  //int HomographyMapping<unsigned char, unsigned char>::
  //Map(Image<unsigned char> &source, Image<unsigned char> &dest, 
  //    const HMatrix& H);
  //#endif
  //

} // namespace 

#endif

#include <Base/Common/BIASpragmaEnd.hh>
