/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
Multimediale Systeme der Informationsverarbeitung
Institut fuer Informatik
Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/



#include "Image/MixtureOfGaussians.hh"
#include <Base/Image/ImageConvert.hh>
#include <Base/Image/ImageDraw.hh>
#include <Base/Image/ROI.hh>
#include <bias_config.h>

using namespace BIAS;
using namespace std;
using namespace BIAS::MOGHelper;

// public

template <class StorageType>
MixtureOfGaussians<StorageType>::
MixtureOfGaussians(const unsigned int numDistributions, 
                   const unsigned int sizeWindow, 
                   const float updateFactor) {
  dDistributions_ = numDistributions;
  dWindow_ = sizeWindow /2;
  bInited_ = false;
  fUpdateFactor_ = updateFactor;
  dMaxWeight_ = 50;
  dMinWeight_ = 20;
  fError_ = 2.9f * 2.9f;
  dFrameCounter_ = 0;
  bSmooth_ = false;
}

template <class StorageType>
void MixtureOfGaussians<StorageType>::
SetSmoothing(bool smooth) {
  bSmooth_ = smooth;
  smoothFilter_.SetSigma(1);
}

template <class StorageType>
int MixtureOfGaussians<StorageType>::
InitWeights(const int maxWeight,
            const int minWeight, 
            const float errorMargin) {
  dMaxWeight_ = maxWeight;
  dMinWeight_ = minWeight;
  fError_ = errorMargin * errorMargin;
  return 0;
}

template <class StorageType>
MixtureOfGaussians<StorageType>::~MixtureOfGaussians() {
  
  
}

template <class StorageType>
int MixtureOfGaussians<StorageType>::
Apply(const BIAS::Image<StorageType> in) {
  if (!bInited_) {
    channels_ = in.GetChannelCount();
    width_ = in.GetWidth();
    height_ = in.GetHeight();
  }
  in.GetROI()->GetCorners(roiULx_, roiULy_, roiLRx_, roiLRy_);
  CreateNormalizedImage_(in, rgbimage_);
  idaImage = rgbimage_.GetImageDataArray();
  if (!bInited_)
    Init_();
  else
    Update_();
  if (dFrameCounter_ > dMinWeight_) {
    Cluster_(difference_, corners_);
  }
  dFrameCounter_++;

  return 0;
}

template <class StorageType>
int MixtureOfGaussians<StorageType>::
GetNormalizedImage(BIAS::Image<float> &normalized) {
  normalized = rgbimage_;
  return 0;
}

template <class StorageType>
int MixtureOfGaussians<StorageType>::
GetWeightImage(BIAS::Image<float> &weight) {
  if (channels_ == 3) {
    weight.ReInit(width_,height_,3,ImageBase::ST_float,false, ImageBase::CM_RGB);
  } else {
    weight.ReInit(width_,height_,1,ImageBase::ST_float,false, ImageBase::CM_Grey);
  }
  weight.FillImageWithConstValue(0.0);
  float **idaRet = weight.GetImageDataArray();
  float val;
  for (unsigned int y=0;y<height_;y++) {
    for (unsigned int x=0;x<width_*channels_;x++) {
      for (unsigned int i=0;i<dCurDistributions_;i++) {
        val = idaWeight[y][x*dDistributions_+i];
        if (val > idaRet[y][x])
          idaRet[y][x] = val;
      }
    }
  }

  return 0;
}

template <class StorageType>
int MixtureOfGaussians<StorageType>::
GetDifferenceImage(BIAS::Image<unsigned char> &diff) {
  diff = difference_;
  if (dFrameCounter_ < dMinWeight_) 
    return -1;
  else
    return 0;
}

template <class StorageType>
int MixtureOfGaussians<StorageType>::
GetDifferenceImageWithVisuals(BIAS::Image<unsigned char> &guiImage) {
  if (dFrameCounter_ < dMinWeight_) 
    return -1;
  unsigned char **visu = differenceVisu_.GetImageDataArray();
  for (unsigned int y=0;y<height_;y++) {
    for (unsigned int x=0;x<width_;x++) {
      for (unsigned int c=0;c<3;c++) {
        visu[y][x*3+c] = idaDiff[y][x];
      }
    }
  }
  unsigned char green[3] = { 0,255,0 };
  unsigned char red[3] = { 255,0,0 };
  unsigned char yellow[3] = { 255,255,0 };
  Corners c;
  for (unsigned int i=0; i<corners_.size();i++) {
    c = corners_[i];
    ImageDraw<unsigned char>::RectangleCorners(differenceVisu_, c.UL.x,
                                               c.UL.y, c.LR.x, c.LR.y, green);
    ImageDraw<unsigned char>::RectangleCorners(differenceVisu_, c.centerofmass.x-1,
                                               c.centerofmass.y-1, c.centerofmass.x+1, 
                                               c.centerofmass.y+1, red);
  }
  if (GetLargestBoundingBox(c) == 0) {
    ImageDraw<unsigned char>::RectangleCorners(differenceVisu_, 
                                               c.UL.x, c.UL.y, 
                                               c.LR.x, c.LR.y, yellow);
  }
  guiImage = differenceVisu_;
  return 0;
}

template <class StorageType>
int MixtureOfGaussians<StorageType>::
GetBoundingBoxes(std::vector<MOGHelper::Corners> &corners) {
  if (corners_.size()>0) {
    corners = corners_;
    return 0;
  }
  else 
    return -1;
}

template <class StorageType>
int MixtureOfGaussians<StorageType>::
GetLargestBoundingBox(MOGHelper::Corners &corners) {
  if (corners_.size()==0)
    return -1;
  corners = maxCorners_;
  return 0;
}


//  Protected

template <class StorageType>
int MixtureOfGaussians<StorageType>::
CreateNormalizedImage_(const BIAS::Image<StorageType> in, BIAS::Image<float> &out) {
  const StorageType **idaUC = in.GetImageDataArray();
  if (channels_ == 3) {
    out.ReInit(width_,height_,3,ImageBase::ST_float,true, ImageBase::CM_RGB);
    float **idaF = out.GetImageDataArray();
    float val, norm;
    for (unsigned int y=0;y<height_;y++) {
      for (unsigned int x=0;x<width_*3;x+=3) {
        val = 0, norm = 0;
        norm = (float) (idaUC[y][x]*idaUC[y][x] + idaUC[y][x+1]*idaUC[y][x+1]
          + idaUC[y][x+2]*idaUC[y][x+2]);
        norm = (float) sqrt(norm);
        idaF[y][x] = idaUC[y][x] / norm;
        idaF[y][x+1] = idaUC[y][x+1] / norm;
        idaF[y][x+2] = idaUC[y][x+2] / norm;
      }
    }
  } else {
    // assume already normalized!
    out = in;
//    out.ReInit(width_,height_,1,ImageBase::ST_float,false, ImageBase::CM_Grey);
//    float **idaF = out.GetImageDataArray();
//    for (unsigned int y=0;y<height_;y++) {
//      for (unsigned int x=0;x<width_;x++) {
//        idaF[y][x] = idaUC[y][x] / 7500;
//      }
//    }
  }
  return 0;
}

template <class StorageType>
int MixtureOfGaussians<StorageType>::
Init_() {
  dCurDistributions_ = 1;
  guess_.ReInit(width_*channels_,height_,dDistributions_,
                ImageBase::ST_float,false, ImageBase::CM_Grey);
  cov_.ReInit(width_*channels_,height_,dDistributions_,
              ImageBase::ST_float,false, ImageBase::CM_Grey);
  weight_.ReInit(width_*channels_,height_,dDistributions_,
                 ImageBase::ST_float,false, ImageBase::CM_Grey);
  if (channels_ == 3) {
    maxWeight_.ReInit(width_,height_,3,ImageBase::ST_float,false, ImageBase::CM_RGB);
  } else {
    maxWeight_.ReInit(width_,height_,1,ImageBase::ST_float,false, ImageBase::CM_Grey);
  }
  difference_.ReInit(width_,height_,1,ImageBase::ST_unsignedchar,false, ImageBase::CM_Grey);
  differenceSmooth_.ReInit(width_,height_,1,ImageBase::ST_unsignedchar,false, ImageBase::CM_Grey);
  differenceVisu_.ReInit(width_,height_,3,ImageBase::ST_unsignedchar,
                         false, ImageBase::CM_RGB);
  idaImage = rgbimage_.GetImageDataArray();
  idaWeight = weight_.GetImageDataArray();
  idaGuess = guess_.GetImageDataArray();
  idaCov = cov_.GetImageDataArray();
  idaMaxWeight = maxWeight_.GetImageDataArray();
  idaDiff = difference_.GetImageDataArray();
  for (unsigned int y=0;y<height_;y++) {
    for (unsigned int x=0;x<width_*channels_;x++) {
       float val = 0;
       int cntPix = 0;
       for (unsigned int yw=y-dWindow_;yw<y+dWindow_;yw++) {
         for (unsigned int xw=x-dWindow_*channels_;xw<x+dWindow_*channels_;xw++) {
           if (yw<0 || xw<0 || yw>=height_ || xw >= width_*channels_)
             continue;
           val += idaImage[yw][xw];
           cntPix++;
         }
       }
       val /= (float)cntPix;
       idaGuess[y][x*dDistributions_] = val;
    }
  }
  int xc = 0;
  for (unsigned int y=0;y<height_;y++) {
    for (unsigned int x=0;x<width_*channels_;x++) {
       float val = 0;
       int cntPix = 0;
       for (unsigned int yw=y-dWindow_;yw<y+dWindow_;yw++) {
         for (unsigned int xw=x-dWindow_*channels_;xw<x+dWindow_*channels_;xw++) {
           if (yw<0 || xw<0 || yw>=height_ || xw >= width_*channels_)
             continue;
           xc = xw*dDistributions_;
           val += (idaImage[yw][xw] - idaGuess[yw][xc]) * 
             (idaImage[yw][xw] - idaGuess[yw][xc]);
           cntPix++;
         }
       }
       val /= (float)cntPix;
       idaCov[y][x*dDistributions_] = val;
    }
  }
  for (unsigned int y=0;y<height_;y++) {
    for (unsigned int x=0;x<width_*channels_;x++) {
      for (unsigned int i=1;i<dDistributions_;i++) {
        xc = x*dDistributions_+i;
        idaWeight[y][xc] = -1;
        idaGuess[y][xc] = -1;
        idaCov[y][xc] = -1;
      }
    }
  }
  bInited_ = true;
  return 0;
}

template <class StorageType>
int MixtureOfGaussians<StorageType>::
InitOnePixel_(unsigned int x,unsigned int y,unsigned int distrib) {
  float val = 0;
  int xc = 0;
  int cntPix = 0;
  for (unsigned int yw=y-dWindow_;yw<y+dWindow_;yw++) {
    for (unsigned int xw=x-dWindow_*channels_;xw<x+dWindow_*channels_;xw++) {
      if (yw<0 || xw<0 || yw>=height_ || xw >= width_*channels_)
        continue;
      val += idaImage[yw][xw];
      cntPix++;
    }
  }
  val /= (float)cntPix;
  idaGuess[y][x*dDistributions_+distrib] = val;
  
  val = 0;
  cntPix = 0;
  for (unsigned int yw=y-dWindow_;yw<y+dWindow_;yw++) {
    for (unsigned int xw=x-dWindow_*channels_;xw<x+dWindow_*channels_;xw++) {
      if (yw<0 || xw<0 || yw>=height_ || xw >= width_*channels_)
        continue;
      xc = xw*dDistributions_+distrib;
      val += (idaImage[yw][xw] - idaGuess[yw][xc]) * (idaImage[yw][xw] - idaGuess[yw][xc]);
      cntPix++;
    }
  }
  val /= (float)cntPix;
  xc = x*dDistributions_+distrib;
  idaCov[y][xc] = val;
  idaWeight[y][xc] = 0;
  return 0;
}

template <class StorageType>
int MixtureOfGaussians<StorageType>::
Update_() {
  int xc = 0;
  bool matched;
  // fill difference image white (matched pixel will be set to black,
  // so changes are white in the result)
  difference_.FillImageWithConstValue((unsigned char)255);
  // iterate through whole picture
  for (unsigned int y=roiULy_;y<roiLRy_;y++) {
    for (unsigned int x=roiULx_;x<roiLRx_*channels_;x++) {
      // initialized with large number
      float min_weight = 10000;
      // initialized with small number
      float max_weight = -10000;
      // worst distribution (i.e. matched minimal times)
      unsigned int min_distrib = 0;
      // best distribution (i.e. matched maximal times)
      unsigned int max_distrib = 0;

      // find min/max distribs
      for (unsigned int i=0;i<dDistributions_;i++) {
        xc = x*dDistributions_+i;
        // find distribution with maximal weight
        if (idaWeight[y][xc] > max_weight) {
          max_weight = idaWeight[y][xc];
          max_distrib = i;
        }
        // find distribution with minimal weight
        if (idaWeight[y][xc] < min_weight) {
          min_weight = idaWeight[y][xc];
          min_distrib = i;
        }
        if (idaWeight[y][xc] < 0)
          break;
      }
      
      matched = false;
      // value in orig image
      float val = idaImage[y][x];
      // check for match in all distributions
      for (unsigned int i=0;i<dDistributions_;i++) {
        xc = x*dDistributions_+i;
        // guess and covariance in this distribution
        float guess = idaGuess[y][xc];
        if (guess<0)
          break;
        float cov = idaCov[y][xc];
          if ((val - guess) * (val - guess) < fError_*cov) {
            // match! weight is increased
            if (idaWeight[y][xc] <= dMaxWeight_) 
              idaWeight[y][xc]++;
            // guess and covariance get updated
            idaGuess[y][xc] = (1-fUpdateFactor_)*guess + fUpdateFactor_*val;
            idaCov[y][xc] = (1-fUpdateFactor_)*cov + fUpdateFactor_*(val-guess)*(val-guess);
            
            // if weight is high enough to satisfy as "non moving", 
            //paint pixel black in difference image
            if (dMinWeight_ > 0) {
              if (idaWeight[y][xc] > dMinWeight_) {
                idaDiff[y][x/channels_] = 0;
              }
            } else {
              if (i == max_distrib) {
                idaDiff[y][x/channels_] = 0;
              }
            }
            matched = true;
          } else {
            // no match! decrease weight
            if (idaWeight[y][xc] > 0)
              idaWeight[y][xc]--;
          }
      }
      // if no match is found, overwrite least matched distribution 
      //with newly initialized values
      if (!matched) {
        InitOnePixel_(x,y,min_distrib);
        idaDiff[y][x/channels_] = 255;
      }
      // update max weight image (only for vis)
      idaMaxWeight[y][x] = max_weight;
    }
  }
  return 0;
}

template <class StorageType>
int MixtureOfGaussians<StorageType>::
Cluster_(const BIAS::Image<unsigned char> in, 
         std::vector<MOGHelper::Corners> &out) {
  out.clear();
  unsigned int w = in.GetWidth();
  unsigned int h = in.GetHeight();
  marker.ReInit(w,h,1);
  marker.FillImageWithConstValue((unsigned char)0);
  const unsigned char **idaIn = in.GetImageDataArray();
  
  if (bSmooth_) {
    smoothFilter_.Filter(in, differenceSmooth_);
    idaIn = (const  unsigned char **)differenceSmooth_.GetImageDataArray();
  }
  

  unsigned char **idaMark = marker.GetImageDataArray();
  MOGHelper::Corners corners;
  unsigned int maxcornerweight = 0;
  for (unsigned int y=roiULy_;y<roiLRy_;y++) {
    for (unsigned int x=roiULx_;x<roiLRx_;x++) {
      unsigned char val = idaIn[y][x];
      unsigned char mar = idaMark[y][x];
      if (val > 0 && mar == 0) {
        if (ClusterDescent_(x,y,w,h, idaIn, idaMark, corners) == 0) {
          out.push_back(corners);
          if (corners.weight > maxcornerweight) {
            maxCorners_ = corners;
            maxcornerweight = corners.weight; 
          }
        }
      }
    }
  }
  return 0;
}

template <class StorageType>
int MixtureOfGaussians<StorageType>::
ClusterDescent_(const unsigned int x, const unsigned int y,
                const unsigned int w, const unsigned int h,
                const unsigned char **idaIn, unsigned char **idaMark, 
                MOGHelper::Corners &corners) {
  MPoint start = { x, y };
  PointQueue queue;
  PointQueue cluster;
  queue.push_back(start);
  idaMark[y][x] = 255;
  while (!queue.empty()) {
    MPoint cur = queue.back();
    if (cluster.size() > 500000)
      break;
    queue.pop_back();
    cluster.push_back(cur);
    if (cur.x<roiLRx_-1) {
      if (idaMark[cur.y][cur.x+1] == 0) {
        idaMark[cur.y][cur.x+1] = 255;
        if (idaIn[cur.y][cur.x+1] > 128) {
          MPoint next = { cur.x+1, cur.y };
          cluster.push_back(next);
          queue.push_back(next);
        }
      }
      if (cur.y>roiULy_) {
        if (idaMark[cur.y-1][cur.x+1] == 0) {
          idaMark[cur.y-1][cur.x+1] = 255;
          if (idaIn[cur.y-1][cur.x+1] > 128) {
            MPoint next = { cur.x+1, cur.y-1 };
            cluster.push_back(next);
            queue.push_back(next);
          }
        }
        if (cur.x>roiULx_) {
          if (idaMark[cur.y-1][cur.x-1] == 0) {
            idaMark[cur.y-1][cur.x-1] = 255;
            if (idaIn[cur.y-1][cur.x-1] > 128) {
              MPoint next = { cur.x-1, cur.y-1 };
              cluster.push_back(next);
              queue.push_back(next);
            }
          }
        }
      }
    }
    if (cur.x>roiULx_) {
      if (idaMark[cur.y][cur.x-1] == 0) {
        idaMark[cur.y][cur.x-1] = 255;
        if (idaIn[cur.y][cur.x-1] > 128) {
          MPoint next = { cur.x-1, cur.y };
          cluster.push_back(next);
          queue.push_back(next);
        }
      }
    }
    if (cur.y<roiLRy_-1) {
      if (cur.x>roiULx_) {
        if (idaMark[cur.y+1][cur.x-1] == 0) {
          idaMark[cur.y+1][cur.x-1] = 255;
          if (idaIn[cur.y+1][cur.x-1] > 128) {
            MPoint next = { cur.x-1, cur.y+1};
            cluster.push_back(next);
            queue.push_back(next);
          }
        }
      }
      if (idaMark[cur.y+1][cur.x] == 0) {
        idaMark[cur.y+1][cur.x] = 255;
        if (idaIn[cur.y+1][cur.x] > 128) {
          MPoint next = { cur.x, cur.y+1 };
          cluster.push_back(next);
          queue.push_back(next);
        }
      }
      if (cur.x<roiLRx_-1) {
        if (idaMark[cur.y+1][cur.x+1] == 0) {
          idaMark[cur.y+1][cur.x+1] = 255;
          if (idaIn[cur.y+1][cur.x+1] > 128) {
            MPoint next = { cur.x+1, cur.y+1 };
            cluster.push_back(next);
            queue.push_back(next);
          }
        }
      }
    }
    
  }
  
  corners.weight = cluster.size();
  if (corners.weight > 50) {
    corners.UL.x = cluster[0].x;
    corners.UL.y = cluster[0].y;
    corners.LR.x = cluster[0].x;
    corners.LR.y = cluster[0].y;
    corners.centerofmass.x = cluster[0].x;
    corners.centerofmass.y = cluster[0].y;
    for (unsigned int i=1;i<corners.weight;i++) {
      MPoint cur = cluster[i];
      corners.centerofmass.x += cur.x;
      corners.centerofmass.y += cur.y;
      if (cur.x > corners.LR.x)
        corners.LR.x = cur.x;
      if (cur.x < corners.UL.x)
        corners.UL.x = cur.x;
      if (cur.y > corners.LR.y)
        corners.LR.y = cur.y;
      if (cur.y < corners.UL.y)
        corners.UL.y = cur.y;
    }
    corners.centerofmass.x /= corners.weight;
    corners.centerofmass.y /= corners.weight;
    return 0;
  } else {
    return -1;
  }
}

template class MixtureOfGaussians<unsigned char>;
template class MixtureOfGaussians<float>;
