/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef __TextureTransformAffine__hh__
#define __TextureTransformAffine__hh__

#include <Base/Common/BIASpragmaStart.hh>
#include <Image/TextureTransform.hh>


namespace BIAS {

  /** @class TextureTransformAffine
   *  @brief analytic properties of affine image warp
   *
   *  See BIAS::TextureTransform for explanation of methods.
   *  @author koeser 01/2008
   **/
  class BIASImage_EXPORT TextureTransformAffine: public TextureTransform {
  public:
    TextureTransformAffine():origin_(0.0,0.0) {
      P_.newsize(6);
      for (unsigned int i=0; i<6; i++) P_[i] = 0;
    }

    virtual ~TextureTransformAffine(){};

    int MapForward(const HomgPoint2D& src, HomgPoint2D& sink) const;

    int MapBackward(const HomgPoint2D& sink, HomgPoint2D& src) const;

    int TextureJacobianForward(const HomgPoint2D& src, 
                               Matrix2x2<double>& Jac) const {
      Jac = A_;
      return 0;
    }

    int TextureJacobianBackward(const HomgPoint2D& sink, 
                                Matrix2x2<double>& Jac) const {
      Jac = Ainv_;
      return 0;
    }

    /** @brief local warp is the same at any image position */
    virtual bool TextureJacobianIsConstant() const { return true;};

    int ParameterJacobianForward(Matrix<double>& Jac, 
                                 const HomgPoint2D& src) ;
    
    int ParameterJacobianBackward(Matrix<double>& Jac, 
                                  const HomgPoint2D& sink);

    /** @brief the jacobian depends on x */
    bool ParameterJacobianIsConstant() const { return false;};


    int ParameterInversionJacobian(Matrix<double>& Jac) const;


    /** @brief a11-1, a12, a21, a22-1, dx, dy */
    void SetParameters(const Vector<double>& p);

    void ComposeWithInverseDeltaP(const Vector<double>& deltaP);

    virtual Vector<double> GetInverseParameters() const {
      Matrix3x3<double> A(MatrixZero);
      A[0][0] = Ainv_[0][0];
      A[0][1] = Ainv_[0][1];
      A[0][2] = txinv_;
      A[1][0] = Ainv_[1][0];
      A[1][1] = Ainv_[1][1];
      A[1][2] = tyinv_;
      A[2][2] = 1;
      return ExtractParameters(A);
    }

    /** @brief origin relative to which rotation and scale is performed */
    void SetOrigin(const Vector2<double>& origin) {
      origin_ = origin;
      SetParameters(P_);
    }

    /** @brief origin relative to which rotation and scale is performed */
    const Vector2<double>& GetOrigin() const {
      return origin_;
    }

    virtual TextureTransformAffine* Clone() const {
      return new TextureTransformAffine(*this);
    }

    /** @brief set up A and t from paramters */
    inline void InterpretParameters(const Vector<double>& p,
                                    double& A00,double& A01,
                                    double& A10,double& A11,
                                    double& tx, double& ty) {
      BIASASSERT(p.Size()==6);
      A00 = p[0]+1.0;
      A01 = p[1];
      A10 = p[2];
      A11 = p[3]+1.0; 
      
      tx = p[4] - (p[0]*origin_[0]+p[1]*origin_[1]);
      ty = p[5] - (p[2]*origin_[0]+p[3]*origin_[1]);
    }

    Vector<double> ExtractParameters(const Matrix3x3<double>& A)
      const {
      Vector<double> p(6);
      p[0] = A[0][0] - 1.0;
      p[1] = A[0][1];
      p[2] = A[1][0];
      p[3] = A[1][1] - 1.0; 

      p[4] = A[0][2] + (p[0]*origin_[0]+p[1]*origin_[1]);
      p[5] = A[1][2] + (p[2]*origin_[0]+p[3]*origin_[1]);
      return p;
    }
 

  protected: 
    /// cached local warp and inverse
    Matrix2x2<double> A_, Ainv_;
    /// cached displacement and inverse
    double tx_, ty_, txinv_, tyinv_; 
    /// origin relative to which rotation and scale is performed
    Vector2<double> origin_;
  }; 

} // end namespace BIAS

#endif
