/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <Base/Common/BIASpragmaStart.hh>
#include <Image/TextureTransformHomography.hh>


int  BIAS::TextureTransformHomography::
MapForward(const HomgPoint2D& src, HomgPoint2D& sink) const {
  sink = H_ * src;
  if (fabs(sink[0])<1e-10) return -1;
  sink.Homogenize();
  return 0;
}

int  BIAS::TextureTransformHomography::
MapBackward(const HomgPoint2D& sink, HomgPoint2D& src) const {
  src = Hinv_ * sink;
  if (fabs(src[0])<1e-10) return -1;
  src.Homogenize();
  return 0;
}

int  BIAS::TextureTransformHomography::
ParameterJacobianForward(Matrix<double>& Jac, const HomgPoint2D& src) {
  Jac.newsize(2,8);

  // get direct access to H entries
  double* pH = H_.GetData();
  const double c11 = *pH++;
  const double c12 = *pH++;
  const double c13 = *pH++;
  const double c21 = *pH++;
  const double c22 = *pH++;
  const double c23 = *pH++;
  const double c31 = *pH++;
  const double c32 = *pH;
      

  // precompute often used values
  const double minusc1src    = -1.0*(c11*src[0]+c12*src[1]+c13);
  const double minusc2src    = -1.0*(c21*src[0]+c22*src[1]+c23);
  const double oneoverc3src  =  1.0/(c31*src[0]+c32*src[1]+1.0);
  const double oneoverc3src2 = oneoverc3src*oneoverc3src;

  // fill in jacobian
  double* pData = Jac.GetData();
  *pData++ = src[0] * oneoverc3src * AFFINE_SENSITIVITY;
  *pData++ = src[1] * oneoverc3src * AFFINE_SENSITIVITY;
  *pData++ = oneoverc3src;
  *pData++ = 0;
  *pData++ = 0;
  *pData++ = 0;
  *pData++ = minusc1src * oneoverc3src2 * src[0]* PROJECTIVE_SENSITIVITY;
  *pData++ = minusc1src * oneoverc3src2 * src[1]* PROJECTIVE_SENSITIVITY;

  *pData++ = 0;
  *pData++ = 0;
  *pData++ = 0;
  *pData++ = src[0] * oneoverc3src * AFFINE_SENSITIVITY;     
  *pData++ = src[1] * oneoverc3src * AFFINE_SENSITIVITY;  
  *pData++ = oneoverc3src;
  *pData++ = minusc2src * oneoverc3src2 * src[0]* PROJECTIVE_SENSITIVITY;
  *pData++ = minusc2src * oneoverc3src2 * src[1]* PROJECTIVE_SENSITIVITY;
  return 0;
}


int  BIAS::TextureTransformHomography::
ParameterJacobianBackward(Matrix<double>& Jac, const HomgPoint2D& sink) {
  Jac.newsize(2,8);

  // get direct access to H entries
  double* pH = H_.GetData();
  const double c11 = *pH++ -1.0;
  const double c12 = *pH++;
  const double c13 = *pH++;
  const double c21 = *pH++;
  const double c22 = *pH++ - 1.0;
  const double c23 = *pH++;
  const double c31 = *pH++;
  const double c32 = *pH;
      
  // precompute often used values
  const double x  = sink[0];
  const double xx = x*x;
  const double y  = sink[1];
  const double yy = y*y;
  const double longterm = (x*c21*c32-x*c31*c22-x*c31-y*c32*c11-y*c32+y*c12*c31+c11*c22+c11+c22+1-c21*c12);
  const double longtermlongterm =longterm*longterm;
  const double anothertermx = x*c22+x-x*c23*c32-y*c12+y*c13*c32+c12*c23-c13*c22-c13;
  const double anothertermy = x*c21-x*c23*c31-y*c11-y+y*c31*c13+c23*c11+c23-c13*c21;
// fill in jacobian
  double* pData = Jac.GetData();


  *pData++ = -(-y*c32+c22+1)*(anothertermx)/longtermlongterm* AFFINE_SENSITIVITY; ;
  *pData++ = (-y*c11*c22+c23*c11*c22+yy*c11*c32-c23*c11*y*c32-y*c11+c23*c11+x*c22*c21-x*c23*c31*c22-y*c22+y*c31*c13*c22+c23*c22-c13*c21*c22+y*c31*c13+x*c23*c31*y*c32+yy*c32-x*c23*c31+c13*c21*y*c32-yy*c31*c13*c32-x*c21*y*c32-c13*c21+c23-y-c23*y*c32+x*c21)/longtermlongterm* AFFINE_SENSITIVITY; ;
  *pData++ =  -(-y*c32+c22+1)/(longterm);
  *pData++ = (-x*c32+c12)*(anothertermx)/longtermlongterm* AFFINE_SENSITIVITY; ;
  *pData++ = -(c12*c23*c11-c23*c11*x*c32+x*y*c32*c11-y*c12*c11-y*c12+x*y*c32+c13*x*c21*c32+c13*y*c12*c31+c12*c23+x*c21*c12-x*c12*c23*c31-y*c31*c13*x*c32-xx*c21*c32+xx*c23*c31*c32-c13*c21*c12-x*c23*c32)/longtermlongterm* AFFINE_SENSITIVITY; ;
  *pData++ = (-x*c32+c12)/(longterm);

  *pData++ = -(y*c12-x*c22-x)*(anothertermx)/longtermlongterm * PROJECTIVE_SENSITIVITY;
  *pData++ = (y*c12*x*c21-xx*c22*c21-xx*c21-x*c23*y*c12*c31+xx*c23*c31*c22+xx*c23*c31-yy*c11*c12+x*c22*y*c11+x*y*c11-yy*c12+x*c22*y+x*y+yy*c13*c12*c31-y*c31*c13*x*c22-y*c31*c13*x+c23*c11*y*c12-c23*c11*x*c22-c23*c11*x+c23*y*c12-c23*x*c22-x*c23-y*c13*c21*c12+c13*c22*x*c21+c13*x*c21)/longtermlongterm * PROJECTIVE_SENSITIVITY;




  *pData++ =   (y*c31*c13*c22+x*c22*c21-c13*c21*c22-y*x*c31*c22-y*c21*c12+c23*c21*c12+y*c31*c13+c13*c21*y*c32-yy*c31*c13*c32-c13*c21+yy*c12*c31+x*c23*c31*y*c32-y*x*c31-c23*x*c21*c32-c23*y*c12*c31+x*c21)/longtermlongterm* AFFINE_SENSITIVITY; ;

  *pData++ =  -(c21-y*c31)*(anothertermy)/longtermlongterm* AFFINE_SENSITIVITY; 
  *pData++ =  (c21-y*c31)/longterm;


  *pData++ =   -(x*c11*c22+x*c11-c23*c11*x*c32-y*c12*c11+c13*c11*y*c32+c12*c23*c11-c13*c11*c22-c13*c11+x*c22-xx*c22*c31-c13*c22+c13*c22*x*c31+x*y*c12*c31-y*c12-x*c12*c23*c31-xx*c31+x+c12*c23-y*c31*c13*x*c32+c13*x*c31-x*c23*c32+y*c13*c32+xx*c23*c31*c32-c13)/longtermlongterm* AFFINE_SENSITIVITY; ;
  *pData++ =  (c11+1-x*c31)*(anothertermy)/longtermlongterm* AFFINE_SENSITIVITY; 

  *pData++ =  -(c11+1-x*c31)/longterm;
  *pData++ = (x*y*c11+c13*x*c21-xx*c21-yy*c12+c13*c22*x*c21-xx*c22*c21+x*c22*y-c12*c23*x*c21+x*c22*y*c11+y*c12*x*c21-y*c13*x*c21*c32-x*c23*y*c32*c11+yy*c13*c32*c11-y*c13*c11*c22+xx*c23*c21*c32-x*c23*y*c32+c23*c11*y*c12-y*c13+yy*c13*c32+c23*y*c12-yy*c11*c12-y*c13*c22-y*c13*c11+x*y)/longtermlongterm * PROJECTIVE_SENSITIVITY;
  *pData++ =   -(-x*c21+y*c11+y)*(anothertermy)/longtermlongterm * PROJECTIVE_SENSITIVITY;

  return 0;
}

void BIAS::TextureTransformHomography::
SetParameters(const Vector<double>& p) {
  P_ = p;
  H_ =  ConstructHFromParameters(p);
  H_.GetInverse(Hinv_); // should check return value !!
  Hinv_ /= Hinv_[2][2];
}
   
void BIAS::TextureTransformHomography::
ComposeWithInverseDeltaP(const Vector<double>& deltaP) {
  HMatrix H = ConstructHFromParameters(deltaP);
  HMatrix HInv;
  H.GetInverse(HInv);
  H = HInv* H_;
  Vector<double> newp = ExtractParameters(H);
  SetParameters(newp);
}


int BIAS::TextureTransformHomography::
ParameterInversionJacobian(Matrix<double>& Jac) const {
  Jac.newsize(8,8);
  const double c11 = P_[0] * AFFINE_SENSITIVITY;
  const double c12 = P_[1] * AFFINE_SENSITIVITY;
  const double c13 = P_[2];
  const double c21 = P_[3] * AFFINE_SENSITIVITY;
  const double c22 = P_[4] * AFFINE_SENSITIVITY;
  const double c23 = P_[5];
  const double c31 = P_[6] * PROJECTIVE_SENSITIVITY;
  const double c32 = P_[7] * PROJECTIVE_SENSITIVITY;


  const double det = c11*c22+c11+c22+1-c21*c12;
  const double detdet = det*det;

  double *pD = Jac.GetData();
  
  
 
  *pD++ =        -(c22+1-c23*c32)/detdet*(c22+1);
  *pD++ =            (c22+1-c23*c32)/detdet*c21;
  *pD++ =                                                           0;
  *pD++ =           (c22+1-c23*c32)/detdet*c12;
  *pD++ =   (-c21*c12+c11*c23*c32+c23*c32)/detdet;
  *pD++ =                                -c32/(det);
  *pD++ =                                                              0;
  *pD++ =                              -c23/(det);

  *pD++ =            (c12-c13*c32)/detdet*(c22+1);
  *pD++ =  -(c11*c22+c11+c22+1-c21*c13*c32)/detdet;
  *pD++ =                               c32/(det);
  *pD++ =              -(c12-c13*c32)/detdet*c12;
  *pD++ =            (c12-c13*c32)/detdet*(c11+1);
  *pD++ =                                                               0;
  *pD++ =                                                              0;
  *pD++ =                                c13/(det);

  *pD++ =   -(c12*c23-c13*c22-c13)/detdet*(c22+1);
  *pD++ =  (c23*c11*c22+c23*c11+c23*c22+c23-c13*c21*c22-c13*c21)/detdet;
  *pD++ =                       -(c22+1)/(det);
  *pD++ =    (c12*c23-c13*c22-c13)/detdet*c12;
  *pD++ =      -(c23*c11+c23-c13*c21)/detdet*c12;
  *pD++ =                            c12/(det);
  *pD++ =                                                         0;
  *pD++ =                                                        0;

  *pD++ =      (c21-c23*c31)/detdet*(c22+1);
  *pD++ =        -(c21-c23*c31)/detdet*c21;
  *pD++ =                                                       0;
  *pD++ =   -(c11*c22+c11+c22+1-c31*c12*c23)/detdet;
  *pD++ =          (c21-c23*c31)/detdet*(c11+1);
  *pD++ =                             c31/(det);
  *pD++ =                            c23/(det);
  *pD++ =                                                          0;

  *pD++ =    -(c21*c12-c31*c13*c22-c31*c13)/detdet;
  *pD++ =            (c11+1-c31*c13)/detdet*c21;
  *pD++ =                             -c31/(det);
  *pD++ =            (c11+1-c31*c13)/detdet*c12;
  *pD++ =       -(c11+1-c31*c13)/detdet*(c11+1);
  *pD++ =                                                            0;
  *pD++ =                             -c13/(det);
  *pD++ =                                                           0;

  *pD++ =      (c12*c23-c13*c22-c13)/detdet*c21;
  *pD++ =    -(c23*c11+c23-c13*c21)/detdet*c21;
  *pD++ =                             c21/(det);
  *pD++ = -(-c13*c11*c22-c13*c11-c13*c22-c13+c12*c23*c11+c12*c23)/detdet;
  *pD++ =                        (c23*c11+c23-c13*c21)/detdet*(c11+1);
  *pD++ =                                             -(c11+1)/(det);
  *pD++ =                                                                            0;
  *pD++ =                                                                         0;

  *pD++ =  -(c21*c32-c31*c22-c31)/detdet*(c22+1);
  *pD++ =       (c21*c32-c31*c22-c31)/detdet*c21;
  *pD++ =                                                            0;
  *pD++ =  (c32*c11*c22+c32*c11+c32*c22+c32-c12*c31*c22-c12*c31)/detdet;
  *pD++ =             -(c32*c11+c32-c12*c31)/detdet*c21;
  *pD++ =                                                        0;
  *pD++ =                    -(c22+1)/(det);
  *pD++ =                       c21/(det);

  *pD++ =                            (c21*c32-c31*c22-c31)/detdet*c12;
  *pD++ = -(-c31*c11*c22-c31*c11-c31*c22-c31+c21*c32*c11+c21*c32)/detdet;
  *pD++ =                                                                                  0;
  *pD++ =                 -(c32*c11+c32-c12*c31)/detdet*c12;
  *pD++ =              (c32*c11+c32-c12*c31)/detdet*(c11+1);
  *pD++ =                                                                       0;
  *pD++ =                                         c12/(det);
  *pD++ =                                    -(c11+1)/(det);
 



  for (unsigned int i=0; i<8; i++) {
    double factor = 1.0;
    switch (i) {
    case 0:
    case 1:
    case 3:
    case 4: factor = AFFINE_SENSITIVITY; break;
    case 6:
    case 7: factor = PROJECTIVE_SENSITIVITY; break;
    }
    for (unsigned int j=0; j<8; j++) {
      switch (j) {
      case 0:
      case 1:
      case 3:
      case 4: Jac[j][i] *= factor / AFFINE_SENSITIVITY; break;
      case 6:
      case 7: Jac[j][i] *= factor / PROJECTIVE_SENSITIVITY; break;
      }
    }
  }

  
  return 0;
}


BIAS::Vector<double> BIAS::TextureTransformHomography::
ExtractParameters(const BIAS::HMatrix& H) const{
  Vector<double> p(8);
  const double norminv = 1.0/H[2][2];
  for (unsigned int i=0; i<8; i++) p[i] = H[i/3][i%3]*norminv;

  // zero warp must be identity:
  p[0] -= 1.0;
  p[4] -= 1.0;

  p[0] /= AFFINE_SENSITIVITY;
  p[1] /= AFFINE_SENSITIVITY;
  p[3] /= AFFINE_SENSITIVITY;
  p[4] /= AFFINE_SENSITIVITY;

  p[6] /= PROJECTIVE_SENSITIVITY;
  p[7] /= PROJECTIVE_SENSITIVITY;

  return p;
}

    
BIAS::HMatrix BIAS::TextureTransformHomography::
ConstructHFromParameters(const Vector<double>&p) const {
  HMatrix H;
  H[2][2] = 1.0;

  H[0][0] = p[0] * AFFINE_SENSITIVITY + 1.0;
  H[0][1] = p[1] * AFFINE_SENSITIVITY;
  H[0][2] = p[2];
  H[1][0] = p[3] * AFFINE_SENSITIVITY;
  H[1][1] = p[4] * AFFINE_SENSITIVITY + 1.0;
  H[1][2] = p[5];
  H[2][0] = p[6] * PROJECTIVE_SENSITIVITY;
  H[2][1] = p[7] * PROJECTIVE_SENSITIVITY;

  return H;
}
