/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
Multimediale Systeme der Informationsverarbeitung
Institut fuer Informatik
Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <Base/Common/BIASpragmaStart.hh>

#ifndef __UndistortionMapping_hh__
#define __UndistortionMapping_hh__

#include <Image/BackwardMapping.hh>
#include <Image/CameraParam.hh>
#include <Image/MatlabTBUndistortion.hh>
#include <Image/SphericalUndistortion.hh>

namespace BIAS {

  /** @class UndistortionMapping
      @ingroup g_filter
      @brief  Maps image src to image sink with spherical or lens distortion
      @author streckel, Dez. 2005
  */

  template <class InputStorageType, class OutputStorageType> 
  class BIASImage_EXPORT UndistortionMapping 
    : public  BackwardMapping<InputStorageType, OutputStorageType>  {
  public:

    virtual ~UndistortionMapping();

    UndistortionMapping();

//     UndistortionMapping(const UndistortionMapping & src);

    /** @brief set your CameraParams for mapping */
    inline void SetCameraParam(const CameraParam& cp) { 
      cp_ = cp;
      if (cp_.IsSpherical()) {
        su_.Init(cp_);
      } else {
        pu_.Init(cp_);
      }
    };
        

    /** @brief Performs a projection of a spherical image given 
        a projection-type and a zoom factor.

        @param src source image 
        @param dst (output) uninitialized destination image
        @param SizeX, SizeY define the size of the output image
        @param FovX, FovY define the Field Of View angle (in deg) in x and y 
        direction that is used for the creation of the K-Matrix for the 
        projected image. If one of these is =0 then the resulting image will 
        be undistorted.
        @param CenterX, Centery is the pixel in the spherical uncalibrated 
        camera image, that lies in viewing direction. This point will be 
        mapped to the principal point in the projected image.
        @param zoomFactor expands the image by a given factor
        @param projectionType use the projection-type constants
        @author Birger Streckel */
    int ProjectSphereImage(const Image<InputStorageType>& src, 
                           Image<OutputStorageType>& dst,
                           const int& SizeX, const int& SizeY, 
                           const double& FovX, const double& FovY,
                           const double& CenterX, const double& CenterY, 
                           const int projectionType, const double& zoomFactor,
                           KMatrix& KMatDst);

    /** @brief Performs a projection of a perspective or other projection 
        type image to a sphere.

        @param src source (perspective)image 
        @param dst (output) uninitialized (spherical) destination image
        @param Ksrc is the KMatrix of the perspecive source image
        @param CenterX, Centery is the pixel in the spherical uncalibrated 
        camera image, that lies in viewing direction. The principal point 
        in the perspective image will be mapped to this point on the sphere
        @param projectionType use the projection-type constants
        @author Birger Streckel */
    int MapPerspToSphere(const Image<InputStorageType>& src, 
                         Image<OutputStorageType>& dst,
                         const KMatrix Ksrc,
                         const double& CenterX, const double& CenterY, 
                         const int projectionType);
      
   /** @brief Using the Matlab camera calibration toolbox parameters 
        for an undistortion of a distorted camera image.
        Reintroduces also aspect ratios !=1.
        If a KMatrix is given it is corrected for the aspect ratio 
        correction.
        @author streckel 06/2004 */
    int DistortPerspImage(const Image<InputStorageType>& src, 
                           Image<OutputStorageType>& dst);
    int DistortPerspImage(const Image<InputStorageType>& src, 
                           Image<OutputStorageType>& dst,
                           KMatrix& K);

   /** @brief Using the Matlab camera calibration toolbox parameters 
       for a distortion of an already undistorted image. 
       Corrects also aspect ratios !=1. Inverse to the 
       UndistortImage function.        
       If a KMatrix is given it is corrected for the aspect ratio 
       correction.
       @author streckel 06/2004 */
    int UndistortPerspImage(const Image<InputStorageType>& src, 
                             Image<OutputStorageType>& dst);
    int UndistortPerspImage(const Image<InputStorageType>& src, 
                             Image<OutputStorageType>& dst,
                             KMatrix& K);
    


  protected:


    /** @brief overload of virtual function in BackwardMapping. 
        Implements the coordinate mapping */
    virtual int GetSourceCoordinates_(const HomgPoint2D& sink,
                                      HomgPoint2D& source) const;

    /* Helper for GetSourceCoordinates */
    int CalculateSphericalCoord_(const HomgPoint2D& sink,
                                 HomgPoint2D& source) const;
    int CalculateToSphereMapping_(const HomgPoint2D& sink,
                                  HomgPoint2D& source) const;

    void CalculatePerspUndistorion_(const HomgPoint2D& sink,
                                    HomgPoint2D& source) const;
    void CalculatePerspDistorion_(const HomgPoint2D& sink,
                                  HomgPoint2D& source) const;

    typedef enum MapType 
      {UNDEF, PERSPECTIVE_DIST, PERSPECTIVE_UNDIST, SPHERICAL, TO_SPHERE};

    /// holds 1.0/log(2.0)
    const double log2inv_;

    /// The camera parameters for the undistortion
    CameraParam cp_;
    
    /// Spherical or perspective undistortion, depending on cp_
    SphericalUndistortion su_;
    MatlabTBUndistortion pu_;

    /// Parameters that influence the spherical mapping 
    /// i.e. direction, type, image part, ...
    double destCenterX_, destCenterY_;
    int destProjType_;
    BIAS::KMatrix destK_;

    ///Parameters that influence the perspectivev mapping

    
    /// spherical or perspective mapping
    MapType mapping_;
    
  };


} // namespace 

#endif

#include <Base/Common/BIASpragmaEnd.hh>
