/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#ifndef __ExampleRANSAC_double_hh__
#define __ExampleRANSAC_double_hh__

#include "MathAlgo/RANSAC.hh"

#include <string>

#define D_DOUBLE_RANSAC_SOLUTION 0x0100

namespace BIAS {

  class doubleRANSAC : public RANSAC<double>
  {
    /** this is a class evaluating the mean of the 
        double data assuming some outlier 
        using the RANSAC algorithm */
  public:
    inline doubleRANSAC();

    inline ~doubleRANSAC();

    inline void Init(std::vector<double>& data, unsigned int data_length, 
                     double inlier_distance_threshold, bool refine_solution, 
                     unsigned int sample_size, double expected_inlier_fraction,
                     double solution_error_threshold);

    inline int GetSampleSolutions(std::vector<unsigned int> &which_samples,
                                  std::vector<double> &solutions);

    inline bool EvaluateSolution(const double& solution,  
                                 bool existing_solution_flag,
                                 std::vector<bool>& inlier,    
                                 int& accept_count,
                                 double& evaluate_score,
                                 bool& ok_to_terminate_flag);

    inline bool RefineSolution(std::vector<unsigned int>& which_samples,
                               double& solution,
                               std::vector<bool>& new_inliers);

  protected:
    std::vector<double> _vData; 
    // data  closer to solution than _dInlierDistanceThreshold will be 
    // regarded as inlying
    double _dInlierDistanceThreshold;
    // if a sample yields more than _dExpectedInlierFraction inliers and 
    // the evaluate_score is lower than _dSolutionErrorThreshold, 
    // RANSACing will stop
    double _dExpectedInlierFraction;
    double _dSolutionErrorThreshold;
  };
  
  doubleRANSAC::doubleRANSAC() : RANSAC<double>()
  {}

  doubleRANSAC::~doubleRANSAC() 
  {}

  void doubleRANSAC::Init(std::vector<double>& data, unsigned int data_length, 
                          double inlier_distance_threshold, 
                          bool refine_solution, unsigned int sample_size,
                          double expected_inlier_fraction, 
                          double solution_error_threshold)
  {
    RANSAC<double>::Init(sample_size, 1, refine_solution);
    _vData=data;
    _uiDataSize=data_length;
    _dInlierDistanceThreshold=inlier_distance_threshold;
    _dExpectedInlierFraction=expected_inlier_fraction;
    _dSolutionErrorThreshold=solution_error_threshold;
  }
  
  int 
  doubleRANSAC::GetSampleSolutions(std::vector<unsigned int> &which_samples,
                                   std::vector<double> &solutions)
  {
    double mean=0.0;
    for (unsigned int i=0; i<which_samples.size(); i++)
      mean+=_vData[which_samples[i]];
    mean/=which_samples.size();
    solutions.clear();
    solutions.push_back(mean);
    return solutions.size();
  }

  bool doubleRANSAC::EvaluateSolution(const double& solution,  
                                      bool existing_solution_flag,
                                      std::vector<bool>& inlier,    
                                      int& accept_count,
                                      double& evaluate_score,
                                      bool& ok_to_terminate_flag)
  {
    bool good_flag=true;
    inlier.clear();
    inlier.resize(_uiDataSize);

    int my_accept_count = 0;
    evaluate_score = 0.0;

    for (unsigned int i=0; i<_uiDataSize; i++){
      inlier[i] = (fabs(solution-_vData[i])<_dInlierDistanceThreshold);
      if (inlier[i]) {
        my_accept_count++;
        evaluate_score += fabs(solution-_vData[i]);
        if (existing_solution_flag && my_accept_count>accept_count) break;
      }
    }
    if (my_accept_count!=0) {
      evaluate_score/=my_accept_count;
    } else {
      evaluate_score=DBL_MAX;
      good_flag=false;
    }
    accept_count=my_accept_count;
    
    ok_to_terminate_flag=
      ((((double)accept_count/(double)_uiDataSize) > _dExpectedInlierFraction)
      && (evaluate_score < _dSolutionErrorThreshold));
    std::string bo=(ok_to_terminate_flag)?("true"):("false");
    BIASDOUT(D_DOUBLE_RANSAC_SOLUTION, "\nsolution "<<solution
             <<"\tscore/expected_score: "
             <<evaluate_score<<"/"<<_dSolutionErrorThreshold
             <<"\tinliers: "<<accept_count<<"/"<<_uiDataSize<<" = "
             <<((double)accept_count/(double)_uiDataSize)<<"/"
             <<_dExpectedInlierFraction<<"\t"<<bo);
    
    return good_flag;
  }

  bool doubleRANSAC::RefineSolution(std::vector<unsigned int>& which_samples,
                                    double& solution,
                                    std::vector<bool>& new_inliers)
  {
    solution=0.0;
    for(unsigned int i=0; i<which_samples.size(); i++)
      solution+=_vData[which_samples[i]];
    solution/=which_samples.size();
    
    new_inliers.clear();
    new_inliers.resize(_uiDataSize);
    for(unsigned int i=0; i<_uiDataSize; i++)
      new_inliers[i]=(fabs(solution-_vData[i])<_dInlierDistanceThreshold);
    BIASDOUT(D_DOUBLE_RANSAC_SOLUTION, "refined solution "<<solution);
    return true;
  }

} // namespace 




#endif // __ExampleRANSAC_double_hh__
