/* This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <MathAlgo/SVD.hh>
#include <MathAlgo/SVD3x3.hh>
#include <Base/Math/Random.hh>
#include <Base/Math/Matrix3x3.hh>
#include <Base/Debug/TimeMeasure.hh>


using namespace BIAS;
using namespace std;

void Print(ostream& os, SVD& svd, SVD3x3&svd3x3)
{
  os << "U: SVD: "   <<svd.GetU()<<endl;
  os << "   SVD3x3: "<<svd3x3.GetU()<<endl;
  os << "   diff: "<<svd.GetU()-svd3x3.GetU()<<endl;
      
  os << "S: SVD: "   <<svd.GetS()<<endl;
  os << "   SVD3x3: "<<svd3x3.GetS()<<endl;
  os << "   diff: "<<svd.GetS()-svd3x3.GetS()<<endl;
      
  os << "V: SVD: "   <<svd.GetV()<<endl;
  os << "   SVD3x3: "<<svd3x3.GetV()<<endl;
  os << "   diff: "<<svd.GetV()-svd3x3.GetV()<<endl;
}

/** @brief tests the integrity of the Sqrt() operation
    @author woelk 01/2006 */
int MatrixSqrt()
{
  const int size = 4;
  const int count = 100;
  Matrix<double> M(size, size), sqrtM, diff;
  Random rand;
  SVD svd;
  for (int i=0; i<count; i++){
    // generate random symmetric positive definit matrix M
    for (int c=0; c<size; c++){
      for (int r=0; r<size; r++){
        M[c][r] = M[r][c] = rand.GetUniformDistributed(-10.0, 10.0);
      }
    }
    M = M * M;

    //cout << "M = "<<M<<endl;
    sqrtM = svd.Sqrt(M);
    //cout << "sqrtM = "<<sqrtM<<endl;
    //cout << "sqrtM*sqrtM = "<<sqrtM*sqrtM<<endl;
    diff = M - (sqrtM * sqrtM);
    if (!diff.IsZero(1e-8)){
      BIASERR("error in matrix square root");
      return -2;
    }
  }
  //cout <<"succefull\n";
  return 0;
}

/* @brief example for SVD usage
   @author woelk 07/2004 */
int main()
{
  int sqrt_res = MatrixSqrt(); 

  Random rand;
  Matrix3x3<double> M;
  SVD svd;
  SVD3x3 svd3x3;
  TimeMeasure tsvd, tsvd3;
  
  

  int num=90000;
  for (int k=0; k<num; k++){
//     if (k==0){
//       M.SetZero();
//       M[0][0]=16384;
//       M[1][1]=M[2][2]=-16384;
//     } else {
      for (int i=0; i<9; i++){
        M.GetData()[i]=rand.GetUniformDistributed(-10.0, 10.0);
      }
      //    }
    
    tsvd3.Start();
    svd3x3.Compute((Matrix<double>)M);
    tsvd3.Stop();
    tsvd.Start();
    svd.Compute((Matrix<double>)M);
    tsvd.Stop();
    
    if (k==0){
      Print(cerr, svd, svd3x3);
    } /*        this does not work, wince svd is not unique

      else {
      if (svd.GetU()!=svd3x3.GetU()){
        BIASERR("differnt U in "<<k<<"th try");
        Print(cerr, svd, svd3x3); 
        break;
      }
      if (svd.GetS()!=svd3x3.GetS()){
        BIASERR("differnt S in "<<k<<"th try");
        Print(cerr, svd, svd3x3);   
        break;
      }
      if (svd.GetV()!=svd3x3.GetV()){
        BIASERR("differnt V in "<<k<<"th try");
        Print(cerr, svd, svd3x3);   
        break;
      }
      } */
    
    Matrix<double> id, res, D, inv;
    Vector<double> d;
    id.newsize(3,3);
    res.newsize(3,3);
    D.newsize(3,3);
    inv.newsize(3,3);
    d.newsize(3);
    id.SetIdentity();
    inv=svd.Invert();
    res=M*inv-id;
    if (!res.IsZero(1e-8)){
      BIASERR("svd different decomposition in "<<k<<"th try");
      cerr << "res :"<<res<<endl;
      cerr << "M :"<<M<<endl;
      Print(cerr, svd, svd3x3);  
      cerr << "svd:   M : " << M << endl;
      d=svd.GetS();
      D.SetZero();
      for (int i=0; i<3; i++) D[i][i]=d[i];
      cerr << "     USV': "<<svd.GetU()*D*svd.GetVT()<<endl;
      break;
    }
    inv=svd3x3.Invert();
    res=M*inv-id;
    if (!res.IsZero(1e-8)){
      BIASERR("svd3x3 different decomposition in "<<k<<"th try");
      cerr << "res :"<<res<<endl;
      cerr << "M :"<<M<<endl;
      Print(cerr, svd, svd3x3);  
      cerr << "svd3x3:   M : " << M << endl;
      d=svd3x3.GetS();
      D.SetZero();
      for (int i=0; i<3; i++) D[i][i]=d[i];
      cerr << "        USV': "<<svd3x3.GetU()*D*svd3x3.GetVT()<<endl;
       break;
   }
  }

  cerr << "svd: ";
  tsvd.Print();
  cerr << "user: "<<(double)tsvd.GetUserTime()/(double)num << " ms / svd\n";
  cerr << "real: "<<(double)tsvd.GetRealTime()/(double)num << " us / svd\n\n";
  cerr << "svd3x3: ";
  tsvd3.Print();
  cerr << "user: "<<(double)tsvd3.GetUserTime()/(double)num << " ms / svd\n";
  cerr << "real: "<<(double)tsvd3.GetRealTime()/(double)num << " us / svd\n\n";

  return (sqrt_res==0)?0:-1;
}


