/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef _MonteCarloTransform_hh_
#define _MonteCarloTransform_hh_

#include <MathAlgo/UncertaintyTransformBase.hh>

#include <vector>

namespace BIAS {

  template <class T> class BIASMathBase_EXPORT Matrix;
  template <class T> class BIASMathBase_EXPORT Vector;

  /** @class MonteCarloTransform
      @brief monte carlo propagation of uncertainty

      @author woelk 11/2006 */
  class BIASMathAlgo_EXPORT MonteCarloTransform 
    : public UncertaintyTransformBase
  {
  public:
    MonteCarloTransform();

    virtual ~MonteCarloTransform();

    /** @brief computes the Monte Carlo approximation of the 
        transformations of the mean and the associated covariance for the 
        point transformation Transform_().
        @returns 0 on success 
                -1 when transformation of mean failed
                +1 when transfromation of sigma point failed */ 
    int Transform(const Vector<double>& src_mean, 
                  const Matrix<double>& src_cov,
                  Vector<double>& dst_mean, Matrix<double>& dst_cov) const;

    int Transform(const Vector<double>& src_mean, 
                  const Matrix<double>& src_cov,
                  std::vector<Vector<double> >& dst_samples) const;

    void Skew(const std::vector<Vector<double> >& dst_samples,
              double skewnness) const;

    /** Sets the number of samples per dimension */
    inline void SetNumSamplesPerDimension(const int num)
    { NumSamplesPerDimension_ = num; }

    inline void SetUseQuasiRandomNumbers(const bool uqrn)
    { UseQuasiRandomNumbers_ = uqrn; }

    /** switch gauss/uniform distribution */
    inline void SetUniformDistribution(const bool uqrn)
    { uniformDistribution_ = uqrn; }

  protected:
    int NumSamplesPerDimension_;
    bool UseQuasiRandomNumbers_;
    
    /// set to true if you want to sample uniform instead of gauss
    bool uniformDistribution_;

    /** @brief implements a point transformation
        @returns 0 on success */
    virtual int Transform_(const Vector<double>& src,
                           Vector<double>& dst) const = 0;
    
    /** computes src_mean.Size()* NumSamplesPerDimension_ samples drawn 
        from multidiemsional normal distribution with first two moments 
        src_mean and src_cov */ 
    int GetSamples_(const Vector<double>& src_mean, 
                     const Matrix<double>& src_cov,
                     std::vector<Vector<double> >& src_samples) const;

  }; // class

} // namespace

#endif // _MonteCarloTransform_hh_
