/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#ifndef __PreemptiveRANSAC_hh__
#define __PreemptiveRANSAC_hh__

#include <bias_config.h>
#include <Base/Common/BIASpragmaStart.hh>

#include <Base/Debug/Debug.hh>
#include <Base/Math/Random.hh>
#include <MathAlgo/Median1D.hh>

#include <list>
#include <vector>

namespace BIAS {

  template <class SolutionType>
  class PreemptiveRANSAC : virtual public Debug
  {
  public:
    virtual ~PreemptiveRANSAC() {};

    void Init(const unsigned data_size);

    void AddSolutionGuesses(const std::vector<SolutionType>& guesses);

    int SolveMaster(const double inlying_data_fraction, const double max_score,
                    SolutionType &solution, std::vector<bool> &inliers, 
                    unsigned &num_inliers, double& score);


    /** @brief Set max number of samples.
        
    Ransacing will stop after max_samples samples, regardless
    of solution quality.  Normally, more than a few thousand samples 
    signifies a problem. */
    inline void SetMaxSamples(const unsigned max_samples) 
    { _uiMaxSamples = max_samples; }

    /** @brief Set min number of samples.
        
    Ransacing will evaluate and compare solutions using sets of
    MinSamples solutions. */
    inline void SetMinSamples(const unsigned min_samples) 
    { _uiMinSamples = min_samples; }

    inline void SetGreedy(const bool greedy)
    { _bGreedy = greedy; }
    
    /** the number of measurements that is initially evaluated */ 
    inline void SetNumInitialChecked(const unsigned num_initial_checked)
    { _uiNumInitialChecked = num_initial_checked; }

    /** when less than MinNumInitialInliers are found, the solution is 
        rejected straight away */
    inline void SetMinNumInitialInliers(const unsigned min_num_initial_inl)
    { _uiMinNumInitialInliers = min_num_initial_inl; }

    inline void SetRefineSolutions(const unsigned refine_solutions)
    { _bRefineSolutions = refine_solutions; }

    /** after each EvaluationBlockSize-th measurement, the solutions are 
        compared and only the better half is kept */
    inline void SetEvaluationBlockSize(const unsigned eval_block_size)
    { _uiEvaluationBlockSize = eval_block_size; }
    
  protected:
    /// the number of samples needed by GetSampleSolutions
    unsigned _uiSampleSize;
    /// the minimu and maximum number of samples investigated
    unsigned _uiMinSamples, _uiMaxSamples;
    /// the number of data for evaluation
    unsigned _uiDataSize;
    /// should termination be aborted when first good solution is found?
    bool _bGreedy;
    /// the number of datat points that is initially checked
    unsigned _uiNumInitialChecked;
    /// the minimum number of inlier from _uiNumInitialChecked
    unsigned _uiMinNumInitialInliers;
    /// should the solutions be refined?
    bool _bRefineSolutions;
    /// after EvaluationBlockSize, bad solutions are removed
    unsigned _uiEvaluationBlockSize;

    struct SolStruct {
      SolutionType solution;
      /// evaluation of the solution has taken place up to eval_index
      unsigned eval_index;
      // the accumulated score of the so-far evaluated datas
      double score;
      // the number of inliers of the so-far evaluated datas
      unsigned num_inliers;
      /** inlier indicator for the data in the order given by _EvaluationOrder
          i.e. inliers[i] indicates if data[_EvaluationOrder[i]] is an inlier*/
      std::vector<bool> inliers;
      /// has a refined version of this solution lareday be added?
      bool refined;
    };
    
    /// all (remaining) solutions
    std::list<SolStruct> _Solutions;
    /// the order in which data is used for solutions evaluation
    std::vector<unsigned> _EvaluationOrder;

    PreemptiveRANSAC(const unsigned sample_size);

    virtual bool IsInlier_(const SolutionType& solution, 
                           const unsigned data_index,
                           double& score) = 0;
    
    /** @brief Compute solution(s) for the given set of samples
        @return the number of solutions found.

        The which_samples array will be of length sample_size, selecting the
        indices to be fit to from the subclass's data array. */
    virtual int GetSampleSolutions_(const std::vector<unsigned> &which_samples,
                                    std::vector<SolutionType> &solutions)= 0;

    virtual bool RefineSolution_(const std::vector<bool>& inliers,
                                 SolutionType& solution) = 0;

    virtual bool GenerateSamples_(const unsigned sample_index, 
                                  const unsigned sample_size,
                                  const unsigned data_size, 
                                  std::vector<unsigned>& which_samples)
    { return GenerateSamplesRandom_(sample_index, sample_size, data_size, 
                                    which_samples); }
    
    bool GenerateSamplesRandom_(const unsigned sample_index, 
                                const unsigned sample_size,
                                const unsigned data_size, 
                                std::vector<unsigned>& which_samples);

    void GenerateEvaluationOrder_(const unsigned data_size);

    /** returns the number of generated solution */
    int GenerateSolutions_(const unsigned block_num, 
                           const unsigned sample_size,
                           const unsigned data_size);

    /** returns the number of succesfully refined solutions */
    int GenerateRefinedSolutions_(const unsigned data_size);

    /** returns the number of rejected solutions */
    int InitialRejectTcd_(const unsigned min_num_inliers, 
                          const unsigned num_checked);

    void EvaluateSolutions_(const unsigned start_index, 
                            const unsigned data_size);

    /** returns the number of rejected solutions */
    int RejectSolutions_(const unsigned eval_index);

    inline bool IsInlier_(typename std::list<SolStruct>::iterator solution, 
                          const unsigned data_index, double& score);
 
    bool GetSolution_(const unsigned min_num_inliers, const double max_score, 
                      SolutionType& solution, std::vector<bool>& inliers, 
                      unsigned& num_inliers, double & score) const;

    const struct SolStruct *
    GetSolutionMaxInliers_(const unsigned min_num_inliers, 
                           const double max_score) const;

    const struct SolStruct *
    GetSolutionMinScore_(const unsigned min_num_inliers, 
                         const double max_score) const;

    inline void ConvertInliers_(const std::vector<bool>& inliers_in_eval_order,
                                std::vector<bool>& inliers) const; 
    
  };

  ///////////////////////////////////////////////////////////////
  
  template <class SolutionType>
  PreemptiveRANSAC<SolutionType>::PreemptiveRANSAC(const unsigned sample_size)
    : _uiSampleSize(sample_size), _uiMinSamples(50), _uiMaxSamples(1000),
      _uiDataSize(0), _bGreedy(true), _uiNumInitialChecked(8), 
      _uiMinNumInitialInliers(2), 
      _bRefineSolutions(true), _uiEvaluationBlockSize(50)
  {
    NewDebugLevel("D_PRANSAC_SolveMaster");
    NewDebugLevel("D_PRANSAC_EvalOrder");
    NewDebugLevel("D_PRANSAC_Evaluate");
    NewDebugLevel("D_PRANSAC_InitialReject");
    NewDebugLevel("D_PRANSAC_Inlier");
    NewDebugLevel("D_PRANSAC_GetSolution");
  }


  template <class SolutionType>
  void PreemptiveRANSAC<SolutionType>::
  Init(const unsigned data_size)
  {
    _uiDataSize = data_size;
    _Solutions.clear();
    _EvaluationOrder.clear();
    GenerateEvaluationOrder_(data_size);
  }

  template <class SolutionType>
  void PreemptiveRANSAC<SolutionType>::
  AddSolutionGuesses(const std::vector<SolutionType>& guesses)
  {
    BIASASSERT(_uiDataSize!=0);
    typename std::vector<SolutionType>::const_iterator it;
    SolStruct st;
    for (it=guesses.begin(); it!=guesses.end(); it++){
      st.solution = *it;
      st.eval_index = 0;
      st.score = 0.0;
      st.num_inliers = 0;
      st.inliers.resize(_uiDataSize);
      st.refined = false;
      _Solutions.push_back(st);
    }
  }


  template <class SolutionType>
  int PreemptiveRANSAC<SolutionType>::
  SolveMaster(const double inlying_data_fraction, const double max_score,
              SolutionType &solution, std::vector<bool> &inliers, 
              unsigned &num_inliers, double& score)
  {
    const int min_num_inliers = 
      (int)rint(inlying_data_fraction * (double)_uiDataSize);
    BIASASSERT(_uiDataSize!=0);
    BIASASSERT(_uiMaxSamples>=_uiMinSamples);
    BIASASSERT(_uiMinNumInitialInliers<=_uiNumInitialChecked);
    const unsigned num_blocks = _uiMaxSamples / _uiMinSamples;
    
    for (unsigned block=0; block<num_blocks; block++){
      BCDOUT(D_PRANSAC_SolveMaster,"\n########## block "<<block<<" #######\n");
      // generate and evaluate raw hypothesis
      GenerateSolutions_(block, _uiSampleSize, _uiDataSize);
      BCDOUT(D_PRANSAC_SolveMaster,"initially generated "<<_Solutions.size()
             <<" solutions\n");

      InitialRejectTcd_(_uiMinNumInitialInliers, _uiNumInitialChecked);
      BCDOUT(D_PRANSAC_SolveMaster, _Solutions.size()<<" solutions remained "
             "after initial rejection\n");
      if (_Solutions.empty()) continue;

      EvaluateSolutions_(_uiNumInitialChecked, _uiDataSize);
      BCDOUT(D_PRANSAC_SolveMaster, _Solutions.size()<<" solutions remained "
             "after evaluation of unrefined solutions\n");
      if (_Solutions.empty()) continue;
      
      // generate and evaluate the refined hypothesis
      if (_bRefineSolutions && !_Solutions.empty()){
        int num_refined = GenerateRefinedSolutions_(_uiDataSize);
        if (num_refined>0){
          BCDOUT(D_PRANSAC_SolveMaster, _Solutions.size()<<" solutions "
                 "after adding refinements\n");
          
          int num_rejected = 
            InitialRejectTcd_(_uiMinNumInitialInliers, _uiNumInitialChecked);
          BCDOUT(D_PRANSAC_SolveMaster, _Solutions.size()<<" solutions "
                 "after initial rejection of refinements\n");
          if (_Solutions.empty()) continue;
          
          if (num_rejected!=num_refined){
            EvaluateSolutions_(_uiNumInitialChecked, _uiDataSize);
            BCDOUT(D_PRANSAC_SolveMaster, _Solutions.size()<<" solutions "
                   "after evaluation of refinements\n");
            if (_Solutions.empty()) continue;
          } else {
            BCDOUT(D_PRANSAC_SolveMaster, "no refined solution passed the "
                   "initial test\n");
          }
        } else {
          BCDOUT(D_PRANSAC_SolveMaster, "no refined solutions\n");
        }
      }
      
      if (_bGreedy && GetSolution_(min_num_inliers, max_score, 
                                   solution, inliers, num_inliers, score)){
        BCDOUT(D_PRANSAC_SolveMaster, "solution found!\n");
        return 0;
      }
    } /// for (int block=0; block<num_blocks; block++){

    // check if any solution is left
    if (GetSolution_(min_num_inliers, max_score, 
                     solution, inliers, num_inliers, score)){
      BCDOUT(D_PRANSAC_SolveMaster, "solution found!\n");
      return 0;
    } 

    BIASERR("PreemptiveRANSAC found no solution, best score "<<score
            <<" / num inliers "<<num_inliers);
    return -1;
  }

  template <class SolutionType>
  void PreemptiveRANSAC<SolutionType>::
  GenerateEvaluationOrder_(const unsigned data_size)
  {
    static BIAS::Random ra;
    _EvaluationOrder.clear();
    const unsigned ds = data_size;
    std::list<unsigned> mlist;
    for (unsigned i=0; i<ds; i++)
      mlist.push_back(i);
    
    unsigned index;
    typename std::list<unsigned>::iterator it;
    BCDOUT(D_PRANSAC_EvalOrder, "evaluation order: ");
    for (unsigned i=0; i<ds; i++){
      index = ra.GetUniformDistributedInt(0u, ds-i-1);
      BIASASSERT(index>=0 && index<ds-i);
      it = mlist.begin();
      for (unsigned k=0; k<index; k++) it++;
      _EvaluationOrder.push_back(*it);
      BCDOUT(D_PRANSAC_EvalOrder, _EvaluationOrder.back()<<" ");
      mlist.erase(it);
    }
    BCDOUT(D_PRANSAC_EvalOrder, std::endl);
  }


  template <class SolutionType>
  int PreemptiveRANSAC<SolutionType>::
  GenerateSolutions_(const unsigned block_num, const unsigned sample_size,
                     const unsigned data_size)
  {
    const unsigned max_sample = (block_num+1)*_uiMinSamples;
    std::vector<unsigned> which_samples;
    std::vector<SolutionType> solutions;
    typename std::vector<SolutionType>::const_iterator it;
    SolStruct st;

    // Note: Counter sample_offset prevents that calls to GenerateSample_()
    //       generate the same random sample if sample solution creation fails!
    static int sample_offset;

    if (block_num == 0) sample_offset = 0;
    const int max_tries = 25;
    int num, tries;
    int num_solutions = 0;
    for (unsigned sample=block_num * _uiMinSamples; 
         sample<max_sample; sample++){
      tries = num = 0;
      do {
        if (GenerateSamples_(sample+sample_offset, sample_size,
                             data_size, which_samples)) {
          num = GetSampleSolutions_(which_samples, solutions);
          for (it=solutions.begin(); it!=solutions.end(); it++){
            st.solution = *it;
            st.eval_index = 0;
            st.score = 0.0;
            st.num_inliers = 0;
            st.inliers.resize(data_size);
            st.refined = false;
            _Solutions.push_back(st);
            num_solutions++;
          }
        } else {
          BIASERR("sample generation failed!");
          return -1;
        }
        if (num == 0) {
          tries++;
          sample_offset++;
        }
      } while (num==0 && tries<max_tries);
    }
    return num_solutions;
  }
  
  template <class SolutionType>
  int PreemptiveRANSAC<SolutionType>::
  GenerateRefinedSolutions_(const unsigned data_size)
  {
    typename std::list<SolStruct>::iterator it;
    SolStruct st;
    st.refined = true;
    std::vector<bool> inliers;
    int num_refined = 0;
    for (it=_Solutions.begin(); it!=_Solutions.end(); it++){
      if (!it->refined){
        BIASASSERT(it->eval_index == data_size);
        ConvertInliers_(it->inliers, inliers);
        st.solution = it->solution;
        if (RefineSolution_(inliers, st.solution)){
          st.eval_index = 0;
          st.score = 0.0;
          st.num_inliers = 0;
          st.inliers.resize(data_size);
          _Solutions.push_back(st);
          num_refined++;
        }
      }
    }
    return num_refined;
  }

  template <class SolutionType>
  int PreemptiveRANSAC<SolutionType>::
  InitialRejectTcd_(const unsigned min_num_inliers, 
                    const unsigned num_checked)
  {
    typename std::list<SolStruct>::iterator it;
    unsigned data = 0;
    double score = 0.0;
    int num_rejected = 0;
    BCDOUT(D_PRANSAC_InitialReject, "initial reject: ");
    for (it=_Solutions.begin(); it!=_Solutions.end();){
      if (it->eval_index>num_checked) { it++; continue; }
      BCDOUT(D_PRANSAC_InitialReject, " (");
      for (data=0; data<num_checked; data++){
        IsInlier_(it, data, score);
        BCDOUT(D_PRANSAC_InitialReject, ((it->inliers[data])?("+"):("-")));
      }
      if (it->num_inliers<min_num_inliers){
        it = _Solutions.erase(it);
        num_rejected++;
        BCDOUT(D_PRANSAC_InitialReject, ")*");
      } else {
        it++;
        BCDOUT(D_PRANSAC_InitialReject, ").");
      }
    }
    BCDOUT(D_PRANSAC_InitialReject, "\n");
    return num_rejected;
  }
      
  template <class SolutionType>
  bool PreemptiveRANSAC<SolutionType>::
  IsInlier_(typename std::list<SolStruct>::iterator solution, 
            const unsigned eval_index, double& score)
  {
    BCDOUT(D_PRANSAC_Inlier, "eval_index: "<<eval_index);
    BCDOUT(D_PRANSAC_Inlier, "solution->eval_index: "<<solution->eval_index);
    if (solution->eval_index<=eval_index){
      // cannot leave indices unevaluated
      //std::cerr<<solution->eval_index<<"  "<<eval_index<<std::endl;
      BIASASSERT(solution->eval_index==eval_index);
      BCDOUT(D_PRANSAC_Inlier, " (new score) "<<std::endl);
      BCDOUT(D_PRANSAC_Inlier, "_EvaluationOrder[eval_index] = "
             <<_EvaluationOrder[eval_index]<<std::endl);
      solution->inliers[eval_index] = 
        IsInlier_(solution->solution, _EvaluationOrder[eval_index], score);
      if (solution->inliers[eval_index]){
        solution->score += score;
        solution->num_inliers++;
      }
      solution->eval_index++;
    } else {
      BCDOUT(D_PRANSAC_Inlier, " (already scored) "<<std::endl);
      score = NAN;
    }
    BCDOUT(D_PRANSAC_Inlier, "inlier: "<<std::boolalpha
           <<solution->inliers[eval_index]<<std::endl);
    return solution->inliers[eval_index];
  }


  template <class SolutionType>
  void PreemptiveRANSAC<SolutionType>::
  EvaluateSolutions_(const unsigned start_index, const unsigned data_size)
  {
    BIASASSERT(start_index<data_size);
    typename std::list<SolStruct>::iterator it;
    double score = 0.0;
    for (unsigned data=start_index; data<data_size; data++){
      for (it=_Solutions.begin(); it!=_Solutions.end(); it++){
        if (it->eval_index>data) { continue; }
        IsInlier_(it, data, score);
      }
      if ( ( ((int)data+1-(int)start_index) % (int)_uiEvaluationBlockSize)
           == 0){
        BCDOUT(D_PRANSAC_Evaluate, "--- eval "<<data<<" ---\n");
        RejectSolutions_(data);
      } 
    }
  }

  template <class SolutionType>
  int PreemptiveRANSAC<SolutionType>::
  RejectSolutions_(const unsigned eval_index)
  {
    // always keep the better half
    std::vector<double> scores;
    typename std::list<SolStruct>::iterator it;
    for (it=_Solutions.begin(); it!=_Solutions.end(); it++){
      if (it->eval_index>eval_index+1) { continue; }
      scores.push_back(it->score);
    }
    if (scores.empty()) {
      BCDOUT(D_PRANSAC_Evaluate, "no solutions left to evaluate\n");
      return 0;
    }
    Median1D<double> med;
    med.Compute(scores);
    const double thresh = med.GetMedian();
    BCDOUT(D_PRANSAC_Evaluate, "score threshold = "<<thresh<<std::endl);
    int num_rejected=0;
    for (it=_Solutions.begin(); it!=_Solutions.end();){
      if (it->score > thresh){
        it = _Solutions.erase(it);
        BCDOUT(D_PRANSAC_Evaluate, "*");
        num_rejected++;
      } else {
        it++;
        BCDOUT(D_PRANSAC_Evaluate, ".");
      }
    }
    BCDOUT(D_PRANSAC_Evaluate, std::endl);
    BCDOUT(D_PRANSAC_Evaluate, _Solutions.size()<<" solutions remaining\n");

    return num_rejected;
  }


  template <class SolutionType>
  bool PreemptiveRANSAC<SolutionType>::
  GetSolution_(const unsigned min_num_inliers, const double max_score, 
               SolutionType& solution, std::vector<bool>& inliers, 
               unsigned& num_inliers, double & score) const
  {
    if (_Solutions.empty()) return false;
    
    const SolStruct *best_sol = 
      GetSolutionMinScore_(min_num_inliers, max_score);
    //GetSolutionMaxInliers_(min_num_inliers, max_score);

    if (best_sol){
      score = best_sol->score;
      num_inliers = best_sol->num_inliers;
      solution = best_sol->solution;
      BCDOUT(D_PRANSAC_GetSolution, "\nbest solution is: "
             <<solution<<"score: "<<score<<" / num inliers: "
             <<num_inliers<<std::endl);
      ConvertInliers_(best_sol->inliers, inliers);
      if (best_sol->num_inliers >= min_num_inliers && 
          best_sol->score <= max_score){
        BCDOUT(D_PRANSAC_GetSolution, "num_inliers: "<<num_inliers
               <<"\nscore: "<<score<<"\n");
        return true;
      } else {
        BCDOUT(D_PRANSAC_GetSolution, "minimum quality criteria not met "
               "(max_score: "<<max_score<<" / min inliers "<<min_num_inliers
               <<")\n");
        return false;
      }
    } else {
      BIASERR("no best solution returned");
      BIASABORT;
    }

    return false;
  }

  template <class SolutionType>
  const struct PreemptiveRANSAC<SolutionType>::SolStruct *
  PreemptiveRANSAC<SolutionType>::
  GetSolutionMaxInliers_(const unsigned min_num_inliers, 
                         const double max_score) const 
  {
    SolStruct const *best = NULL;
    unsigned max_inliers = 0;
    double min_score = DBL_MAX;
    typename std::list<SolStruct>::const_iterator it;
    for (it=_Solutions.begin(); it!=_Solutions.end();it++){
      BCDOUT(D_PRANSAC_GetSolution, "solution: "<<it->solution<<std::endl);
      if (it->num_inliers>max_inliers){
        max_inliers = it->num_inliers;
        min_score = it->score;
        BCDOUT(D_PRANSAC_GetSolution, "   is best (num inliers "<<max_inliers
               <<", score "<<min_score<<")"<<std::endl);
        best = &(*it);
      } else {
        BCDOUT(D_PRANSAC_GetSolution, "           (num inliers "<<max_inliers
               <<", score "<<min_score<<")"<<std::endl);
      }
      if (it->num_inliers==max_inliers && it->score < min_score){
        min_score = it->score;
        best = &(*it);
        BCDOUT(D_PRANSAC_GetSolution, "   new best score "<<min_score
               <<std::endl);
      }
    }
    
    return best;
  }

  template <class SolutionType>
  const struct PreemptiveRANSAC<SolutionType>::SolStruct *
  PreemptiveRANSAC<SolutionType>::
  GetSolutionMinScore_(const unsigned min_num_inliers, 
                       const double max_score) const 
  {
    SolStruct const *best = NULL;
    double min_score = DBL_MAX;
    typename std::list<SolStruct>::const_iterator it;
    for (it=_Solutions.begin(); it!=_Solutions.end(); it++){
      BCDOUT(D_PRANSAC_GetSolution, "solution: "<<it->solution
             <<std::endl);
      if (it->score < min_score){
        min_score = it->score;
        best = &(*it);
        BCDOUT(D_PRANSAC_GetSolution, "   is best (min_score "<<min_score
               <<" / num inliers: "<<it->num_inliers<<" )"<<std::endl);
      } else {
        BCDOUT(D_PRANSAC_GetSolution, "\nscore: "<<it->score
               <<" / num inliers: "<<it->num_inliers<<std::endl);
      }
    }
    return best;
  }

  template <class SolutionType>
  bool PreemptiveRANSAC<SolutionType>::
  GenerateSamplesRandom_(const unsigned sample_index,
                         const unsigned sample_size, const unsigned data_size, 
                         std::vector<unsigned>& which_samples)
  {
    BIASASSERT(sample_size<data_size);
    BIASASSERT(data_size>0);
    
    // initialzing random with 0 and 1 does not seem to work well
    //Random ran(sample_index+rand());
    Random ran(sample_index);
    unsigned count=0;
    unsigned index;
    which_samples.clear();
    which_samples.reserve(sample_size);
    const unsigned max_tries = 10000;
    do {
      index = ran.GetUniformDistributedInt(0u, data_size-1);
      if (find(which_samples.begin(), which_samples.end(), index)==
          which_samples.end()){
        which_samples.push_back(index);
        count++;
        //std::cout << index << "  ";
      }
    } while (count<sample_size && count<max_tries);
    //std::cout << std::endl;
    if (count>=max_tries){ 
      BIASERR("error generating samples (max_tries reached)");
      return false; 
    }
    
    return true;
  }

  template <class SolutionType>
  void PreemptiveRANSAC<SolutionType>::
  ConvertInliers_(const std::vector<bool>& inliers_in_eval_order,
                  std::vector<bool>& inliers) const
  {
    // convert the inlier vector
    inliers.resize(inliers_in_eval_order.size());
    const unsigned data_size = inliers.size();
    BIASASSERT(data_size == _EvaluationOrder.size());
    for (unsigned i=0; i<data_size; i++){
      BIASASSERT(_EvaluationOrder[i]<data_size);
      inliers[_EvaluationOrder[i]] = inliers_in_eval_order[i];
    }
  }

}
#endif // __PreemptiveRANSAC_hh__
