#ifndef __DPPanTiltControl_hh__
#define __DPPanTiltControl_hh__

#ifndef WIN32
#include <unistd.h>
#endif
#include <stdio.h>
#include <iostream>
#include "Base/Math/Vector.hh"
#include <vector>
#include "Base/Debug/Debug.hh"
#include <Geometry/RMatrix.hh>

#ifdef BIAS_HAVE_PTHREADS
#include <pthread.h>
#endif //BIAS_HAVE_PTHREADS

//for the dp-ptu
extern "C"
{
#include "include/ptu.h"
#ifndef WIN32
#include "include/linuxser.h"
#else
#include "include/W32SERIA.H"
#endif
}

//this is the resolution of the pantiltunit!!!!
#define RESOLUTIONPT 0.051428 
#define SPEEDMODE    0
#define ABSOLUTEMODE   1

#define MAXSPEED       2901

#define D_DP_RECURSIVE     (1<<0)
#define D_DP_MODE          (1<<1)
#define D_DP_PID           (1<<2)
#define D_DP_PIREC         (1<<3)

//define Object tracking control mode
#define TRACK_ABS_POS_MODE       100
#define TRACK_PID_SPEED_MODE     101
#define TRACK_REC_SPEED_MODE     102

namespace BIAS{

  /**
     @class DPPanTiltControl
     This class controls the Digital Perception Pan Tilt Unit 
     @author Ingo Schiller August - October 2004
  */ 
  class DPPanTiltControl : public BIAS::Debug{
  public:
    /**
       standard constructor / destructor 
    */
    DPPanTiltControl();  
    ~DPPanTiltControl();
    

    /**
       here define the mode in which the camera is to operate
       0 is SpeedControlMode
       1 is AbsolutePositioning mode 
       @returns 0 on no error
       @returns !=0 on error
    */
    int SetMode(int nMode);
    /**
       just resets the position of the ptu to 0
       @return bool indicating success of reset
    */
    bool ResetPanTiltUnit();
    
    /**
       just stops the currently moving PTU
    */
    void StopPanTiltUnit();

    /**
       opens the serial port and initialises pan tilt unit 
       @param int the number of the serial port
    */
    bool InitPanTiltUnit(int nCOMportNumber);

    /** 
       opens the given port and initialises pan tilt unit 
       @param char[] the name of the port
    */
    bool InitPanTiltUnit(const char *COMportName);
    
    /**
       closes the comport stream connected to ptu
    */
    void ClosePanTiltUnit();
    
    //do exactly what you think, just move your as nValue steps in direction
    void MoveUp(int nValue);
    void MoveDown(int nValue);
    void MoveLeft(int nValue);
    void MoveRight(int nValue);
    void StartCircularMovement();

    //the parameters of the Ziegler Nichols tuning are set here
    void SetKp( BIAS::Vector<int> &kp);
    void SetKi( BIAS::Vector<int> &ki);
    void SetKd( BIAS::Vector<int> &kd);

    //set basic stuff for movement
    char SetSpeed(int nPanSpeed,int nTiltSpeed);
    char SetAcceleration(int nPanAcc, int nTiltAcc);
   
    /**
       @returns the current speed
    */
    void GetSpeed(long &panSpeed, long &tiltSpeed);

    /**
       @returns the current Acceleration
    */
    void GetAcceleration(long &panAcc, long &tiltAcc);

    /**
       @return the current position of the PTU
    */
    void GetCurrentPosition(long &lPanPos,long &lTiltPos); 
    
    /**
       @return the current position of the PTU in degrees
    */
    void GetCurrentPositionDegrees(double &panDegree, double &tiltDegree);

    void SetPosition(int nPanPos, int nTiltPos, bool bWaitComplete=true);

    /**
       sets the initial orientation of the ptu
       to correct the angels of the sensor
    */
    void SetSensorInitOrientation(BIAS::RMatrix& initMatrix);
    void CompensateHeliMovement(BIAS::RMatrix& matrix);
    
    /**
       simple object tracking method, uses absolut positioning error
       and moves PTU/camera to that position(newMean)
       mode has to be set to: 
       dpc.SetMode(ABSOLUTEMODE)
    */
    void FollowMidPos(BIAS::Vector<double> &newMean);
    
  
    /**
       here do the PID Controlling of the PTU
       very simple algorithm with out deadtime and frequency
       mode has to be set to:
       dpc.SetMode(SPEEDMODE) 
    */
    void PIDSpeedAdjustFollow(BIAS::Vector<double> &newMean);
 
    
    /**
       this does the digital controlling of the ptu as described in 
       Herbert Schlitt pages 215ff.
       taces care of deadtime and frequency of the algo.
       makes recursive calculation and calcs the diff in velocity
       of ptu.
       mode has to be set to:
       dpc.SetMode(SPEEDMODE) 
    */
    void RecursivePIControl(BIAS::Vector<double> &newMean);

    void SetSamplingTime(float t);

    void SetResetTime(float tn);

    /**
	Send a command string, terminated by a space or <cr> character
    */
    void SendRawCommand(std::string cmd);

    /**
       	Read a line from the serial port, returns length of read string,
        zero on nothing to read
    */
    int ReadRawLine(std::string &cmd);
	
    /*	  
         For normal Orientation the bool = true (standard),
         for mounting upside down, the Orientation is false!! 
         e.g. under the heli
    */
    inline void SetOrientation(bool bOrientation){
      bDefaultOrientation_ = bOrientation;
    };
  
    inline void SetPictureSize(unsigned int xSize, unsigned int ySize){
      xPictureSize_ = xSize;
      yPictureSize_ = ySize;
    };

    inline void SetApertureAngle(double xAngle, double yAngle){
      xApertureAngle_ = xAngle;
      yApertureAngle_ = yAngle;
    };

#ifdef BIAS_HAVE_PTHREADS
    void GetCurrentPositionDegreesThreaded(double &panDegree, double& tiltDegree);
#endif //BIAS_HAVE_PTHREADS 

  protected:

#ifdef BIAS_HAVE_PTHREADS
    
    // thread to control camera movement
	  pthread_t control_thread; 
    static pthread_mutex_t condition_mutex;
    static long panThreaded_;
    static long tiltThreaded_;
    static void* threadFuncCircle_(void *ptr);

#endif //BIAS_HAVE_PTHREADS

    BIAS::RMatrix ptuMatrix_, deltaMatrix_, HandEyeMatrix_, sensorMatrixMinus1_;
    bool  bInitComp_;
    short int xPTSpeed_;
    short int yPTSpeed_; 
    short int maxSpeed_;
    short int panSpeedComp_;
    short int tiltSpeedComp_;

    double KpX_,KpY_,KiX_,KiY_,KdX_,KdY_;
    int ZiNiErrorX_; //The total error in the Ziegler-Nichols tuning method  x
    int ZiNiErrorY_; //The total error in the Ziegler-Nichols tuning method  y
    int ZiNiErrorTminus1X_;
    int ZiNiErrorTminus1Y_;
    portstream_fd COMstream_;
    int followObjectMaxDiff_;
    int followObjectMinDiff_;
    double xApertureAngle_;
    double yApertureAngle_;
    unsigned int xPictureSize_;
    unsigned int yPictureSize_;
    float ptStepDegrees_;
    //indicates if PanTitlUnit is upside down 
    bool bDefaultOrientation_; 
    
    //variables for RecursivePIControl
    float T_; //Time between discretisation of sampler 
    float Tn_; //Nachstellzeit (Verzoegerung wann der Wert am Ausgang des
               // Intergralen Elements anliegt)
    short int xPTDeltaSpeed_;  
    short int yPTDeltaSpeed_;
    short int xPTDeltaSpeedTMinus1_;  
    short int yPTDeltaSpeedTMinus1_;
    std::vector<int> ErrorTMinus1_;
    int RPICounter_; // the counter, shows which iteration 
                     // of RecursivePIControl it is
    
  };//end class

}// end namespace BIAS

#endif // __DPPanTiltControl_hh__
