#ifndef __ShowCamWxThreads_hh__
#define __ShowCamWxThreads_hh__

#include <bias_config.h>
#include <wx/wx.h>

#include <Gui/StringConv.hh>
#include <Image/Camera.hh>
#include <Base/Image/ImageIO.hh>

#include <VideoSource/VideoSource_Base.hh>

#include <Base/Common/TimeStamp.hh>
#ifdef BIAS_HAVE_DCAM
#include <VideoSource/VideoSource_DCAM_XB3.hh>
#endif

#ifdef BIAS_HAVE_SWISSRANGER
#include <VideoSource/VideoSource_SwissRanger.hh>
#endif
#ifdef BIAS_HAVE_PMD_PMDZess
#include  <VideoSource/VideoSource_usbPMD.hh>
#endif


namespace BIAS {
 
  /**
     @class  ImageSaveThread
     @brief  This class creates a thread which stores an image in a file
     if rescale is not zero then the stored image is downsampled
     by this factor.
     @author Christoffer Menk
  */
  class ImageSaveThread : public wxThread
  {
  public:
    // Constructor for single images
    ImageSaveThread(const wxString& filename,
                    const BIAS::Camera<unsigned char>& img,
                    const int& downsample,
                    wxCondition *condition,
                    wxMutex* mutex,
                    std::vector<Camera<unsigned char>*>& imageBuffer,
                    bool& writeToBuffer,
                    unsigned int& readIndex,
                    unsigned int& writeIndex);

    // constructor for triclops cam
    ImageSaveThread(const wxString& filename,
                    const BIAS::Camera<unsigned char>& img,
                    const BIAS::Camera<unsigned char>& imgCenter,
                    const BIAS::Camera<unsigned char>& imgRight,
                    const int& downsample,
                    wxCondition *condition,
                    wxMutex* mutex,
                    std::vector<Camera<unsigned char>*> &imageBuffer,
                    std::vector<Camera<unsigned char>*> &imageBufferCenter,
                    std::vector<Camera<unsigned char>*> &imageBufferRight,
                    bool& writeToBuffer,
                    unsigned int& readIndex,
                    unsigned int& writeIndex);

    // constructor for swissranger
    ImageSaveThread(const wxString& filename,
                    const BIAS::Camera<unsigned char>& img,
                    const BIAS::Camera<float>& imgDepth,
                    const BIAS::Camera<float>& imgAmp,
                    const int& downsample,
                    wxCondition *condition,
                    wxMutex* mutex,
                    std::vector<Camera<unsigned char>*> &imageBuffer,
                    std::vector<Camera<float>*> &imageBufferPmdDepth,
                    std::vector<Camera<float>*> &imageBufferPmdAmp,
                    bool& writeToBuffer,
                    unsigned int& readIndex,
                    unsigned int& writeIndex);

    // constructor for pmdzess
    ImageSaveThread(const wxString& filename,
                    const BIAS::Camera<unsigned char>& img2D,
                    const BIAS::Camera<unsigned char>& img,
                    const BIAS::Camera<float>& imgDepth,
                    const BIAS::Camera<float>& imgAmp,
                    const int& downsample,
                    wxCondition *condition,
                    wxMutex* mutex,
                    std::vector<Camera<unsigned char>*> &imageBuffer2d,
                    std::vector<Camera<unsigned char>*> &imageBuffer,
                    std::vector<Camera<float>*> &imageBufferPmdDepth,
                    std::vector<Camera<float>*> &imageBufferPmdAmp,
                    bool& writeToBuffer,
                    unsigned int& readIndex,
                    unsigned int& writeIndex);

    // constructor for single float images
    ImageSaveThread(const wxString& filename,
                    const BIAS::Camera<float>& img,
                    const int& downsample,
                    wxCondition *condition,
                    wxMutex* mutex,
                    std::vector<Camera<float>*> &imageBuffer,
                    bool& writeToBuffer,
                    unsigned int& readIndex,
                    unsigned int& writeIndex);


    ~ImageSaveThread() {}
 
    virtual ExitCode Entry();
    
    void SetSaveDirectory(std::string &path){
      saveDirectory_ = path;
    }
  
  private:
    void SaveImage_(Camera <unsigned char>& im);

    void SaveImage_(Camera <float>& imF);

    void SaveImage_(Camera <unsigned char>& im,
                    Camera <float>& imDepth,
                    Camera <float>& imAmp);
    
    void SaveImage_(Camera <unsigned char>& im,
                    Camera <unsigned char>& imC,
                    Camera <unsigned char>& imR);

    // set standard values for boolean flags
    inline void InitBoolean_(){
      isFloat_ = false;
      isTriclops_ = false;
      isAlive_ = true;
      isSwissRanger_ = false;
      isPMDZess_ = false;
    }
      
    std::string saveDirectory_;
    bool isFloat_;
    bool isTriclops_;
    bool isAlive_;
    bool isSwissRanger_;
    bool isPMDZess_;
    wxString Filename_;
    wxString FilenameOut_;
    BIAS::Camera<unsigned char> Img_; 
    BIAS::Camera<unsigned char> ImgCenter_; 
    BIAS::Camera<unsigned char> ImgRight_; 
    BIAS::Camera<unsigned char> Img2D_; 
    BIAS::Camera<float> ImgF_;
    BIAS::Camera<float> ImgDepth_;
    BIAS::Camera<float> ImgAmp_;
    int DownSample_;
    std::vector<Camera<unsigned char>*>* ImageBuffer_;
    std::vector<Camera<unsigned char>*>* ImageBufferCenter_;
    std::vector<Camera<unsigned char>*>* ImageBufferRight_;

    std::vector<Camera<float>*>* ImageBufferPMDDepth_;
    std::vector<Camera<float>*>* ImageBufferPMDAmp_;
    std::vector<Camera<float>*>* ImageBufferFloat_;
    
    wxMutex* BufferAccess_;
    bool* WriteToBuffer_;
    unsigned int* WriteIndex_;
    unsigned int* ReadIndex_;
    wxCondition* Condition_;
    unsigned int ImageCounter_;
  };






  /**
     @class  wxThreadReady
     @brief  Baseclass for Image Grabbing threads for various cameras
     @author Christoffer Menk
  */
  class wxThreadReady : public wxThread
  {
  public:
    wxThreadReady( BIAS::Camera<unsigned char>* img,
                   wxCondition *condition,
                   wxMutex* BufferAccessMutex,
                   std::vector<Camera<unsigned char>*> &imageBuffer,
                   bool& writeToBuffer,
                   unsigned int& readIndex,
                   unsigned int& writeIndex,
                   wxMutex *mutexCondition)
      :wxThread(wxTHREAD_JOINABLE)
    { 
      Img_ = img;  
      Condition_ = condition;
      BufferAccess_ =  BufferAccessMutex;
      ImageBuffer_ = &imageBuffer;
      WriteToBuffer_ = &writeToBuffer;
      ReadIndex_ = &readIndex;
      WriteIndex_ = &writeIndex;
      BufferSize_ = ImageBuffer_->size();
      mutexCondition_ = mutexCondition;

      //Create();
    }   

    void SetWriteToBuffer(bool buffer){
      *WriteToBuffer_ = buffer;
    }
  protected:
    bool *WriteToBuffer_;  
    wxMutex *mutexCondition_;
    wxCondition* Condition_;
    wxMutex* BufferAccess_;  
    unsigned int* WriteIndex_;
    unsigned int* ReadIndex_;
    unsigned int BufferSize_;
    BIAS::Camera<unsigned char>* Img_;
    std::vector<Camera<unsigned char>*>* ImageBuffer_;
  };

  /**
     @class  GrabSingleThread
     @brief  This class creates a thread which grabs a single image
     @author Christoffer Menk
  */  
  class GrabSingleThread : public wxThreadReady
  {
  public:
    GrabSingleThread(BIAS::VideoSource* cam,
                     BIAS::Camera<unsigned char>* img,
                     wxCondition *condition,
                     wxMutex* mutex,
                     std::vector<Camera<unsigned char>*> &imageBuffer,
                     bool& writeToBuffer,
                     unsigned int& readIndex,
                     unsigned int& writeIndex,
                     wxMutex *mutexCondition);
    ~GrabSingleThread() {}
 
    virtual ExitCode Entry();
   
    virtual void OnExit() {
    }
    
  private:
    BIAS::VideoSource* Cam_;
  };

#ifdef BIAS_HAVE_SWISSRANGER

  /**
     @class  GrabSingleThreadSR
     @brief  This class creates a thread which grabs a single image
     @author Ingo Schiller
  */  
  class GrabSingleThreadSR : public wxThreadReady
  {
  public:
    GrabSingleThreadSR(BIAS::VideoSource_SwissRanger* cam,
                       BIAS::Camera<unsigned char>* img,
                       BIAS::Camera<float>* imgDepth,
                       BIAS::Camera<float>* imgAmp,                   
                       wxCondition *condition,
                       wxMutex* mutex,
                       std::vector<Camera<unsigned char>*> &imageBuffer,
                       std::vector<Camera<float>*> &imageBufferPmdDepth,
                       std::vector<Camera<float>*> &imageBufferPmdAmp,
                       bool& writeToBuffer,
                       unsigned int& readIndex,
                       unsigned int& writeIndex,
                       wxMutex *mutexCondition);
    ~GrabSingleThreadSR() {}
 
    virtual ExitCode Entry();
  
    virtual void OnExit() {
    }
    
  private:  
    
    BIAS::VideoSource_SwissRanger* Cam_;
    BIAS::Camera<float>* ImgDepth_;
    BIAS::Camera<float>* ImgAmp_;
    std::vector<Camera<float>*>* ImageBufferPMDDepth_;
    std::vector<Camera<float>*>* ImageBufferPMDAmp_;
       
  };
#endif


#ifdef BIAS_HAVE_DCAM
  //#ifndef WIN32
  class GrabSingleThreadXB3 : public wxThreadReady
  {
  public:
    GrabSingleThreadXB3(BIAS::VideoSource_DCAM_XB3* cam,
                        BIAS::Camera<unsigned char>* imgLeft,
                        BIAS::Camera<unsigned char>* imgMiddle,
                        BIAS::Camera<unsigned char>* imgRight,
                        BIAS::Camera<unsigned char>* allImages,
                        wxCondition* condition,
                        wxMutex* mutex,
                        std::vector<Camera<unsigned char>*> &imageBuffer,
                        std::vector<Camera<unsigned char>*> &imageBufferCenter,
                        std::vector<Camera<unsigned char>*> &imageBufferRight,
                        bool& writeToBuffer,
                        unsigned int& readIndex,
                        unsigned int& writeIndex,
                        wxMutex *mutexCondition);
 
    ~GrabSingleThreadXB3() {}
    virtual ExitCode Entry();
    
    virtual void OnExit() {
    }
    
  
  private:
    BIAS::VideoSource_DCAM_XB3* Cam_;
    BIAS::Camera<unsigned char>* ImgLeft_;
    BIAS::Camera<unsigned char>* ImgMiddle_;
    BIAS::Camera<unsigned char>* ImgRight_;
    BIAS::Camera<unsigned char>* AllImages_;
    std::vector<Camera<unsigned char>*>* ImageBufferCenter_;
    std::vector<Camera<unsigned char>*>* ImageBufferRight_;
  };
  //#endif //BIAS_HAVE_DCAM  
#endif


#ifdef BIAS_HAVE_PMD_PMDZess
  /**
     @class  GrabSingleThreadPMDZess
     @brief  This class creates a thread which grabs 3 to 4 images from the PMD Zess camera
     @author Ingo Schiller
  */  
  class GrabSingleThreadPMDZess : public wxThreadReady
  {
  public:
    GrabSingleThreadPMDZess(BIAS::VideoSource_usbPMD* cam,
                            BIAS::Camera<unsigned char>* img2D,
                            BIAS::Camera<unsigned char>* img,
                            BIAS::Camera<float>* imgDepth,
                            BIAS::Camera<float>* imgAmp,
                            wxCondition *condition,
                            wxMutex* mutex,
                            std::vector<Camera<unsigned char>*> &imageBuffer2d,
                            std::vector<Camera<unsigned char>*> &imageBuffer,
                            std::vector<Camera<float>*> &imageBufferPmdDepth,
                            std::vector<Camera<float>*> &imageBufferPmdAmp,
                            bool& writeToBuffer,
                            unsigned int& readIndex,
                            unsigned int& writeIndex,
                            wxMutex *mutexCondition);
 
    ~GrabSingleThreadPMDZess() {}
 
    virtual ExitCode Entry();
    virtual void OnExit() {
    }
    
  private:
    BIAS::VideoSource_usbPMD* Cam_;
    BIAS::Camera<float>* ImgDepth_;
    BIAS::Camera<float>* ImgAmp_;  
    BIAS::Camera<unsigned char>* Img2D_;
    std::vector<Camera<unsigned char>*>* ImageBuffer2D_;
    std::vector<Camera<float>*>* ImageBufferPMDDepth_;
    std::vector<Camera<float>*>* ImageBufferPMDAmp_;
    
   
  };
#endif  

} // end of namespace BIAS
  
#endif // __ShowCamWxThreads_hh__
