/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef __GenSynthMatchesRig_hh__
#define __GenSynthMatchesRig_hh__

#include <bias_config.h>

#include <Base/Common/BIASpragmaStart.hh>

#include <iostream>
#include <string>
#include <vector>

#include <Base/Image/Image.hh>
#include <Base/Geometry/HomgPoint2D.hh>
#include <Base/Geometry/HomgPoint3D.hh>
#include <Base/Geometry/KMatrix.hh>
#include <Base/Geometry/FMatrixBase.hh>
#include <Base/Geometry/EuclideanTransf3D.hh>
#include <Base/Geometry/RMatrixBase.hh>
#include <Base/Geometry/PMatrixBase.hh>
#include <Utils/Param.hh>
#include <Base/Math/Random.hh>
#include "MatchDataBase.hh"
#include "GenSynthMatches.hh"
#include <Base/Math/Vector3.hh>
#include <Geometry/RMatrix.hh>

namespace BIAS{

#define GENSYNTHMATCHESRIG_MAX_NR_OF_CAMERAS 8

  /**
     needed for WIN32-DLL, templates of templates... would generate
     too long name symbols!
  **/
  struct VecOMapOStrDbl { 
    VecOMapOStrDbl() {}
    VecOMapOStrDbl(std::vector<std::map<std::string, double> > tmp) {
      vecOMaps = tmp;
    }
    std::vector<std::map<std::string, double> > vecOMaps;
  };

  /**
     @brief Generates synthetic matches for cameras in a rig
  */
  class BIASUtils_EXPORT GenSynthMatchesRig
  {
  public:
    ///@brief Constructor - register parameters
    GenSynthMatchesRig(Param &para);
    virtual ~GenSynthMatchesRig();

    // clears all internal vectors and resets K and Ki
    // also resets _rand
    void Reset();
    
    ///@brief Creates 3D points and generates 2D matches for them.
    int CreateMatches();

    ///@addtogroup ImageGeneration
    ///@{
    ///@brief Creates images showing the 2D matches for all cameras.
    void DrawTrue(std::vector<Image<unsigned char> >&im, int minindex=0, 
                  int maxindex=1) const;
    ///@brief Creates an image showing the 2D matches for a camera.
    void DrawTrue(Image<unsigned char> &img, int minindex, 
                  int maxindex, int camIndex) const;
    ///@brief Creates images showing the 2D matches for all cameras. 
    void DrawNoisy(std::vector<Image<unsigned char> >&im, int minindex=0, 
                   int maxindex=1) const;
    ///@brief Creates an image showing the 2D matches for a camera.
    void DrawNoisy(Image<unsigned char> &img, int minindex, 
                   int maxindex, int camIndex) const;
    ///@}

    //int Create3DPoints();

    ///@brief write in xml/ascii file format
    int Write(std::string file) const;
    ///@brief read in xml/ascii file format
    int Read(std::string file);

    ///@brief returns the vector of the used GenSynthMatches object
    inline std::vector<BIAS::GenSynthMatches> GetCams() const {
      return CamMatches_;
    }

    friend BIASUtils_EXPORT std::ostream& 
    operator<<(std::ostream& os, const GenSynthMatchesRig& m);

    friend BIASUtils_EXPORT std::istream& 
    operator>>(std::istream& os, GenSynthMatchesRig& m);

  protected:
    ///@brief Pointer to the parameter object
    Param *allParams_;

    ///@brief motion vectors read from the param file
    BIAS::Vector<double> MasterCenterX_;
    BIAS::Vector<double> MasterCenterY_;
    BIAS::Vector<double> MasterCenterZ_;
    BIAS::Vector<double> MasterAnglesX_;
    BIAS::Vector<double> MasterAnglesY_;
    BIAS::Vector<double> MasterAnglesZ_;

    //camID -> transform -> value (relative).
    std::vector<std::map<std::string, double> > SlaveCamsRelative_;
    //camID -> timestep -> transform -> value (global).
    std::vector<VecOMapOStrDbl > SlaveCamsGlobal_;
    ///@brief Vector containing the cameras
    std::vector<GenSynthMatches> CamMatches_;
 
    ///@brief dummy to initialize the param object
    GenSynthMatches GenMatches_;
    ///@brief The number of cameras used. Read from the param object.
    int *NumCameras_;

    ///@brief Update the param object for the given slave camera.
    void UpdateParams_(int camID);

    ///@brief Setup the global poses of the cameras and read the parameters.
    void Setup_();
    ///@brief read the parameters from the param object.
    void GetParameters_();
    ///@brief Add parameters to the param object.
    void AddParameters_(Param &para);
    ///@brief Print the poses of the cameras. For debugging only.
    void PrintCamPoses_();

  }; // class

  BIASUtils_EXPORT std::ostream& 
  operator<<(std::ostream& os, const GenSynthMatchesRig& m);

} // namespace


#include <Base/Common/BIASpragmaEnd.hh>

#endif 
