/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <wtypes.h>
#include <1394Camera.h>
#include <Base/Debug/TimeMeasure.hh>

#include "VideoSource_DCAM.hh"

using namespace BIAS;

using namespace std;

// tested with Scorpion, Dragonfly

VideoSource_DCAM::VideoSource_DCAM(int card): VideoSource(card)
{  
  ForceF7_ = false;
  DirectF7_ = false;
  UserTopLeft_ = false;
  UseBinning_ = false;
  ColorMode_ = ImageBase::CM_UYVY422;
  DcamFps_ = 3;
  Width_ = 0;
  Height_ = 0;
  Left_ = 0;
  Top_ = 0;
  BytesPerPacket_ = 0;
  ColorChannels_ = 0;
  FramesPerSecond_ = 0.0;

  IsControllable_ = true;
}

VideoSource_DCAM::~VideoSource_DCAM()
{
}

int VideoSource_DCAM::OpenDevice(int camid)
 {
  // check if a camera is connected.
	int res = camera.CheckLink();
	if (res == CAM_SUCCESS)
	{
		BIASDOUT(D_DCAM,"A Camera was found and selected.");
		SelectedCam_=camera.GetNode();
	} 
	else 
	{
		BIASERR("No Camera was found.");
	}

  BIASDOUT(D_DCAM,"VideoSource_DCAM::OpenDevice("<<camid<<")");

  int select = camera.SelectCamera(camid);
  
  // is there a camera with id: camid?
  if (select == CAM_SUCCESS)
  {
    SelectedCam_ = camid;
    BIASDOUT(D_DCAM,"VideoSource_DCAM::OpenDevice("<<camid<<") successful");
  } else {
    BIASERR("No Device found");   
    return -1; 
  }

  BIASDOUT(D_DCAM,"Setting up camera "<<SelectedCam_<<"");  

  // QueryCam(); is done by the camera

  return InitCam();
}

int VideoSource_DCAM::OpenDevice()
{
  return OpenDevice(0);
}

int VideoSource_DCAM::CloseDevice()
{
  return 0;
}

int VideoSource_DCAM::UsePort(int channel)
{
  BIASWARN("Not yet implemented for DCAM_CMU!!!");
  return -1;
}

int VideoSource_DCAM::ExternalTrigger(bool on)
{
  camera.GetCameraControl(::FEATURE_TRIGGER_MODE)->Inquire();
  if (camera.GetCameraControl(::FEATURE_TRIGGER_MODE)->HasPresence()) {
    camera.GetCameraControl(::FEATURE_TRIGGER_MODE)->SetOnOff(on);
    camera.GetCameraControl(::FEATURE_TRIGGER_MODE)->Status();
    if (!camera.GetCameraControl(::FEATURE_TRIGGER_MODE)->StatusOnOff()) {
      BIASERR("Trigger refuses to be switched on.");
    }
    return 0;
  } else {
    BIASERR("No trigger mode available for this camera");
    return -1;
  }
}


int VideoSource_DCAM::InitCam()
{	
	//Just for testing the code, tested with Scorpion, Dragonfly
	//Width_ = 400;
	//Height_ = 200;  
  //Top_ = 0;
  //Left_ = 0;
  //UserTopLeft_ = true;
	// Set Firewire B 
	

	
	//camera.Set1394b(true);
	if(camera.InitCamera()==CAM_SUCCESS)
	{
	
	cout<< "camera inited successfully" << endl;
    BIASDOUT(D_DCAM,"VideoSource_DCAM::InitCamera() successful");
	}
	else
	{
    BIASERR("VideoSource_DCAM::InitCamera() failed");
		return -1;
	}	
	if(camera.Has1394b()){
		int res = camera.Set1394b(true);
		cout << "Status Set1394b: " << res << endl;
		BIASDOUT(D_DCAM,"VideoSource_DCAM::InitCamera(): set cam successful to 1394b");
		cout << "Firewire Status: " << camera.Status1394b() << endl;
		cout << "CAM Max Speed: " << camera.GetMaxSpeed() << endl;

	}

	// Look up camera Model and Vendor
  LARGE_INTEGER camModel;
  camModel.QuadPart = 0;
  camera.GetCameraUniqueID(&camModel);
  //DBG? following call overrides user color settings, why?
  CamModel_ = GetModelForEuid(camModel);

  if(!DirectF7_){

    if (Width_ != 0 || Height_ != 0) // UseBinning_ not implemented.
	  {
		  // try to set user definitions
		  if ( UserSizeToMode() < 0) 
		  {
			  BIASERR("UserSizeToMode() failed");
			  return -1;
      }
	  }
    else
	  {
		  // Select best mode, format and framerate for camera.
		  SelectDefault();
	  }
  }

	// set camera to parameters.
	if (camera.SetVideoFormat(DcamFormat_) == CAM_SUCCESS && 
		  camera.SetVideoMode(DcamMode_)		 == CAM_SUCCESS)
	{
   
		if (DcamFormat_ != 7)
		{
      // higher value -> higher framerate
      // 2=7.5fps, 3=15fps, ...
      float fps = FramesPerSecond_;
      if (FramesPerSecond_ != 0.0)
      {
        int videoframerate = 7;

        // set video frame rate, for more details see CMU driver manual
        if (fps <= 1.875)
			    videoframerate = 0;
        else if (fps <= 3.75)
			    videoframerate = 1;
        else if (fps <= 7.5)
			    videoframerate = 2;
        else if (fps <= 15.0)
			    videoframerate = 3;
        else if (fps <= 30.0)
			    videoframerate = 4;
        else if (fps <= 60.0)
			    videoframerate = 5;
        else if (fps <= 120.0)
			    videoframerate = 6;
        else if (fps <= 240.0)
			    videoframerate = 7;

        // try to set user video frame rate, else set default
        if (camera.SetVideoFrameRate(videoframerate)!= CAM_SUCCESS)
        {
          BIASERR("Failed to set user frame rate, setting default.");
			    camera.SetVideoFrameRate(DcamFps_);
        }
      }
      else
			  camera.SetVideoFrameRate(DcamFps_);
		}
		else
		{
			camera.GetCameraControlSize()->SetColorCode(COLOR_CODE_RGB8);
			camera.GetCameraControlSize()->SetSize((unsigned short)Width_,
                                             (unsigned short)Height_);
	
			// Maximum bpp is selected.
			if (BytesPerPacket_ == 0) 
			{
				CalcBytesPerPacket();
			} 			
     
    
			camera.GetCameraControlSize()->
        SetBytesPerPacket((unsigned short)BytesPerPacket_);
			camera.GetCameraControlSize()->SetPos((unsigned short)Left_,
                                            (unsigned short)Top_);
		}

    // setTopLeft
    if(UserTopLeft_){
      int res = camera.GetCameraControlSize()->SetPos((unsigned short)Left_,
                                            (unsigned short)Top_);
      if(res != CAM_SUCCESS){
      }
    }

    CompleteInitialized_ = true;
		return 0;
	}

	// Initialization failed.
	BIASERR("Can not set DCAM Camera to requested parameters.");
	return -1;
}

int VideoSource_DCAM::GrabSingle(Camera <unsigned char> &image)
{
//cout << "In der normalen DCAM" << endl;
  BIASDOUT(D_DCAM,"VideoSource_DCAM::GrabSingle("<<SelectedCam_<<")");
  //TimeMeasure StopWatch_; 

  //StopWatch_.Reset();
  //StopWatch_.Start();

	// try to capture image.
	//int res = camera.CaptureImage();
	int res = camera.AcquireImage();
  //StopWatch_.Stop();
  //BIASERR("CaptureImage: " << StopWatch_.GetRealTime()/1000 << "ms");

  if (res == CAM_SUCCESS)
	{
    BIASDOUT(D_DCAM,"VideoSource_DCAM::CaptureImage successful");
    image.SetUID(BIAS::UUID::GenerateUUID());
	}
	else
	{
    BIASERR("VideoSource_DCAM::CaptureImage failed");		
	}

  //StopWatch_.Reset();
  //StopWatch_.Start();
	// initialized with Width_, Height_
  image.CopyIn_NoInit(camera.GetRawData(NULL));   

  //StopWatch_.Stop();
   // BIASERR("CopyIn_NoInit: " << StopWatch_.GetRealTime()/1000 << "ms");

  return 0;
}

int VideoSource_DCAM::QueryCam()
{  
  // is done automatically by the camera after initialisation
  return 0;
}


int  VideoSource_DCAM::SetFormat7(int mode, int bpp, int colorchannels)
{
  //BIASERR("PLEASE, (someone forgot to) implement me!!!");
  //BIASABORT;

  DcamFormat_  = 7;
  DcamMode_ = mode;
  BytesPerPacket_ = bpp;
  DirectF7_ = true;
  ColorChannels_ = colorchannels;

  return 0;
}

int VideoSource_DCAM::ShowCapabilities()
{  
  cout <<"Camera: "<<SelectedCam_<<endl;
  camera.RefreshControlRegisters();

  cout <<"Supported Formats: ";
  for (int i=0; i<8; i++)
  {
    if (camera.HasVideoFormat(i)) 
    {
      cout <<"Format"<<i<<" ";
    }
  }
  cout <<endl;

	// Get supported modes for available formats.
  cout <<"Suported Modes:"<<endl;
  for (int i=0; i<8; i++)
  {
    if (camera.HasVideoFormat(i))
    {
      cout <<"Format"<<i<<": ";
      for (int j=0; j<8; j++)
      {
        if (camera.HasVideoMode(i,j))
        {
          cout <<"Mode: "<<j<<"; ";
        }      
      cout <<endl;      
			}
    }
  }
  
  // Only for Formats 0..2 is a framerate meaningful
  cout <<"Supported Framerates:"<<endl;
  for (int format=0; format < 3; format++)
  {  
    cout <<" Format"<<format<<": ";
    for (int mode=0; mode<8; mode++)
    {
      if (camera.HasVideoMode(format,mode))
      {
        cout <<"Mode "<<mode<<":  ";
        for (int f=0; f<8; f++)
        {
          if (camera.HasVideoFrameRate(format,mode,f)) 
          {
            cout <<1.875 * double(0x1<<f)<<";  ";
          }
        }
      }
    }    
    cout <<endl;
  }

  return 0;
}

#define USE_NATIVE_XGA

int VideoSource_DCAM::UserSizeToMode()
{
  BIASDOUT(D_DCAM,"VideoSource_DCAM::UserSizeToMode("<<SelectedCam_<<"): "
           <<Width_<<"x"<<Height_<<" BytesPerPixel_: "<<BytesPerPixel_
           <<" ColorMode "<<ColorMode_);
  cout <<"BztesPerPixel" << BytesPerPixel_ << endl; 

	if (BytesPerPixel_ == 1.0)
	{
    if (ColorMode_ == ImageBase::CM_UYVY422) {
      ColorChannels_ = 2;
      if (!ForceF7_) {
				if (Width_ == 1600 && Height_ == 1200) 
				{	
					DcamFormat_ = 2;
					DcamMode_   = 3;
				}
				else if (Width_ == 1280 && Height_ == 960) 
				{
					DcamFormat_ = 2;
					DcamMode_   = 0;
				}
				else if (Width_ == 1024 && Height_ == 768) 
				{
					DcamFormat_ = 1;
					DcamMode_   = 3;
				}
				else if (Width_ == 800 && Height_ == 600) 
				{
					DcamFormat_ = 1;
					DcamMode_   = 0;
				}
				else if (Width_ == 640 && Height_ == 480) 
				{
					DcamFormat_ = 0;
					DcamMode_   = 3;
				}
				else if (Width_ == 320 && Height_ == 240) 
				{
					DcamFormat_ = 0;
					DcamMode_   = 1;
				}
				else  // Format 7 required
				{
					if (UserSizeToFormat7_() < 0)
					{
						BIASERR("VideoSource_DCAM::UserSizeToFormat7_(): failed");
						return -1;
					}
				}
      }
      else
			{
				if (UserSizeToFormat7_() < 0)
				{
					BIASERR("VideoSource_DCAM::UserSizeToFormat7_(): failed");
					return -1;
				}
			}
    }
    else if (ColorMode_ == ImageBase::CM_Grey ||ColorMode_ == ImageBase::CM_Bayer_RGGB
																						  || ColorMode_ == ImageBase::CM_Bayer_GBRG
																							|| ColorMode_ == ImageBase::CM_Bayer_GRBG
																							|| ColorMode_ == ImageBase::CM_Bayer_BGGR )
		{ 
      ColorChannels_ = 1;
      if (!ForceF7_ ) {
				if (Width_ == 1600 && Height_ == 1200)
				{
				  DcamFormat_ = 2;
					DcamMode_   = 5;	
				}
				else if (Width_ == 1280 && Height_ == 960)
				{
					DcamFormat_ = 2;
					DcamMode_   = 2;	
				}	
				else if (Width_ == 1024 && Height_ == 768)
				{
				  DcamFormat_ = 1;
					DcamMode_   = 5;	
				}
				else if (Width_ == 640 && Height_ == 480 && CamModel_!=CAMERA_PTGREY_DRAGONFLY_HIRES)
				{
				  DcamFormat_ = 0;
					DcamMode_   = 5;	
				}
				else
				{
					if (UserSizeToFormat7_() < 0)
					{
						BIASERR("VideoSource_DCAM::UserSizeToFormat7_(): failed");
						return -1;
					}
				}
      }
      else
			{
				if (UserSizeToFormat7_() < 0)
				{
					BIASERR("VideoSource_DCAM::UserSizeToFormat7_(): failed");
					return -1;
				}
			}
    }
    else 
		{ 
      BIASERR("VideoSource_DCAM::UserSizeToMode(): Unsupported ColorModel_"
              <<ColorMode_);
      return -1;
    }
	}
	else
	{
		BIASDOUT(D_DCAM,"VideoSource_DCAM::UserSizeToMode(): 16 Bit not implemented.");	
		return -1;
	}
	return 0;
}

 
int VideoSource_DCAM::UserSizeToFormat7_()
{

	// Set Camera to Format7 to inquire unitwidth and height.
	camera.SetVideoFormat(7);
	camera.SetVideoMode(0);

  unsigned short* tmpWidth = new unsigned short;
  unsigned short* tmpHeight = new unsigned short;
  camera.GetCameraControlSize()->GetSizeUnits(tmpWidth,tmpHeight);
	UnitWidth_  = *tmpWidth;
	UnitHeight_ = *tmpHeight;

  camera.GetCameraControlSize()->GetSizeLimits(tmpWidth,tmpHeight);
	MaxWidth_  = *tmpWidth;
	MaxHeight_ = *tmpHeight;

  delete tmpWidth;
  delete tmpHeight;

  if ((unsigned int)Width_ % UnitWidth_ >0)
	{
    BIASERR("VideoSource_DCAM::UserSizeToMode(): width of "<<Width_
	    <<" not multiple of: "<<UnitWidth_);
    return -1;
  }
  if ((unsigned int)Height_ % UnitHeight_ >0)
	{
    BIASERR("VideoSource_DCAM::UserSizeToMode(): height of "<<Height_
	    <<" not multiple of: "<<UnitHeight_);
    return -1;
  }

  if (!UseBinning_) 
	{
    DcamMode_   = 0;
    DcamFormat_ = 7;

    if (!UserTopLeft_) 
		{
      Left_ = (MaxWidth_ - Width_) / 2;
      Top_  = (MaxHeight_ - Height_) / 2;
    }
    BIASDOUT(D_DCAM_F7,"Using format7: "<<Width_<<"x"<<Height_<<"+"<<Left_
	     <<"+"<<Top_);
  }
  else 
	{ // Binning for PtGrey Dragonfly, not implemented yet.
    BIASDOUT(D_DCAM_F7,"Using format7 Mode 1 (Binning): "<<Width_<<"x"<<Height_
	     <<"+"<<Left_<<"+"<<Top_);

    DcamMode_      = 1; 
    DcamFormat_    = 7;
    ColorChannels_ = 1;
  }

  return 0;
}

int VideoSource_DCAM::SelectDefault()
{

	// Get best Mode for Camera
  for (int i=0; i<3; i++)
  {
    if (camera.HasVideoFormat(i))
    {
      for (int j=0; j<6; j++)					// ToDo: No 16-Bits Modes allowed.
      {
        if (camera.HasVideoMode(i,j))
        {
					DcamFormat_ = i;
					DcamMode_   = j;
        }         
			}
    }
  }

	// Select best Framerate for selected format and mode
	for (int k=0; k<8; k++)
	{
    if(camera.HasVideoFrameRate(DcamFormat_,DcamMode_,k)) {
			DcamFps_ = k;
    }
	}

  switch (DcamFormat_) {
  case 0:  // Format 0
    switch (DcamMode_) {
    case 0:  // Mode 0
      ColorChannels_ = 2;
      Width_ = 160; Height_ = 120; ColorMode_ = ImageBase::CM_YUV444;
      break;
    case 1:  // F0 Mode 1
      ColorChannels_ = 2;
      Width_ = 320; Height_ = 240; ColorMode_ = ImageBase::CM_UYVY422;
      break;
    case 2:  // F0 Mode 2
      ColorChannels_ = 2;
      Width_ = 640; Height_ = 480; ColorMode_ = ImageBase::CM_YUV411;
      break;
    case 3:  // F0 Mode 3
      ColorChannels_ = 2;
      Width_ = 640; Height_ = 480; ColorMode_ = ImageBase::CM_UYVY422;
      break;
    case 4:  // F0 Mode 4
      ColorChannels_ = 3;
      Width_ = 640; Height_ = 480; ColorMode_ = ImageBase::CM_RGB;
      break;
    case 5:  // F0 Mode 5
      ColorChannels_ = 1;
      Width_ = 640; Height_ = 480;
      break;
    case 6:     // F0 Mode 6   
      ColorChannels_ = 1;
      Width_ = 640; Height_ = 480;
      break;
    }
    
    break;
  case 1:   // Format 1
    switch (DcamMode_) {
    case 0: // F1 Mode 0
      ColorChannels_ = 2;
      Width_ = 800; Height_ = 600;  ColorMode_ = ImageBase::CM_UYVY422;
      break;
    case 1: // F1 Mode 1
      ColorChannels_ = 3;
      Width_ = 800; Height_ = 600;  ColorMode_ = ImageBase::CM_RGB;
      break;
    case 2: // F1 Mode 2
      ColorChannels_ = 1;
      Width_ = 800; Height_ = 600;  
      break;
    case 3:   // F1 Mode 3
      ColorChannels_ = 2;
      Width_ = 1024; Height_ = 768;  ColorMode_ = ImageBase::CM_UYVY422;
      break;
    case 4: // F1 Mode 4
      ColorChannels_ = 3;
      Width_ = 1024; Height_ = 768;  ColorMode_ = ImageBase::CM_RGB;
      break;
    case 5: // F1 Mode 5
      ColorChannels_ = 1;
      Width_ = 1024; Height_ = 768;  
      break;
    case 6: // F1 Mode 6
      ColorChannels_ = 1;
      Width_ = 800; Height_ = 600;  
      break;
    case 7: // F1 Mode 
      Width_ = 1024; Height_ = 768;  
      ColorChannels_ = 1;
      BytesPerPixel_ = 2.0;
      break;
    }
    
    break;
  case 2: // Format 2
    switch (DcamMode_) {
    case 0:
      ColorChannels_ = 2;
      Width_ = 1280; Height_ = 960;  ColorMode_ = ImageBase::CM_UYVY422;
      break;
    case 1:
      ColorChannels_ = 3;
      Width_ = 1280; Height_ = 960;  ColorMode_ = ImageBase::CM_RGB;
      break;
    case 2:
      ColorChannels_ = 1;
      Width_ = 1280; Height_ = 960; 
      break;
    case 3:
      ColorChannels_ = 2;
	  Width_ = 1600; Height_ = 1200;  ColorMode_ = ImageBase::CM_UYVY422;
      break;
    case 4:
      ColorChannels_ = 3;
      Width_ = 1600; Height_ = 1200;  ColorMode_ = ImageBase::CM_RGB;
      break;
    case 5:
      ColorChannels_ = 1;
	  Width_ = 1600; Height_ = 1200;  ColorMode_ = ImageBase::CM_Grey;
      break;
    case 6:
      ColorChannels_ = 1;
      Width_ = 1280; Height_ = 960;  
      break;
    case 7:
      ColorChannels_ = 1;
      Width_ = 1600; Height_ = 1200;  
      break;
    }
    break;
  }

	return 0;
}

#ifdef BUILD_IMAGE_USHORT
int VideoSource_DCAM::GrabSingle(BIAS::Camera <unsigned short> &image)
{
  return 0;
}
#endif  

CameraModel VideoSource_DCAM::GetModelForEuid(LARGE_INTEGER /*euid*/) 
{
	// Look up for Model and Vendor!!!
  char *vendorChr = new char[256];
  char *modelChr = new char[256];
  int vendorLen = 256;
  int modelLen = 256;

  camera.GetCameraVendor(vendorChr,vendorLen);
  camera.GetCameraName(modelChr,modelLen);

  string vendor(vendorChr);
  string model(modelChr);
  delete vendorChr;
  delete modelChr;
  ColorMode_ = ImageBase::CM_Grey;	// Default Colormodel
  if (vendor.find("Point Grey Research") != string::npos)
	{
    if (model.find("Dragonfly") != string::npos)
	  {
  	  unsigned long r1040;
	    camera.ReadQuadlet(0x1040, &r1040);  
      // little endian Red: 52 Green: 47 Blue: 42 Monochrome 59

      switch(r1040)
      { 	  
        case 0x52474742 :
          ColorMode_ = ImageBase::CM_Bayer_RGGB;	  
          break;
          
        case 0x47425247 :
          ColorMode_ = ImageBase::CM_Bayer_GBRG;	  
          break;
          
        case 0x47524247 :
          ColorMode_ = ImageBase::CM_Bayer_GRBG;	    
          break;
          
        case 0x42474752 :
          ColorMode_ = ImageBase::CM_Bayer_BGGR;	    
          break;

        case 0x59595959 :
					ColorMode_ = ImageBase::CM_Grey;
          return CAMERA_PTGREY_DRAGONFLY_HIRES;
          break;                                                       
          
        default :
					BIASERR("Error, unknown PtGrey color model!!!");
          return CAMERA_UNKNOWN;
          break;                                                       
      }
  	  return CAMERA_PTGREY_DRAGONFLY_COLOR;

    }else if(model.find("Scorpion") != string::npos){
			ColorMode_ = ImageBase::CM_Bayer_RGGB;
			return CAMERA_PTGREY_SCORPION_COLOR_1600;
	}else if(model.find("Flea2 FL2-08S2M") != string::npos){
	  ColorMode_ = ImageBase::CM_Grey;
	  return CAMERA_PTGREY_FLEA2_BW;
    }
	else if(model.find("Grasshopper GRAS-20S4C") != string::npos){
		//ColorMode_ = ImageBase::CM_YUYV422;
		return CAMERA_PTGREY_GRASSHOPPER_COLOR;
  }	else if(model.find("Bumblebee XB3 BBX3-13S2C") != string::npos){
        ColorMode_ = ImageBase::CM_Bayer_GBRG; 
        BIASDOUT(D_DCAM,"Model: PtGrey Bumblebee XB3");
        return CAMERA_PTGREY_BUMBLEBEE_XB3;        
  }
  }

  if ((vendor.find("SONY") != string::npos) &&
      (model.find("DFW-X700  v1.02B") != string::npos)) {
    ColorMode_ = ImageBase::CM_UYVY422;//why?
    return CAMERA_SONY_DFW_X700;
  }

  // other camera	
  return CAMERA_UNKNOWN;
}

void VideoSource_DCAM::ActivateScorpionTimestamp()
{
}

int VideoSource_DCAM::PreGrab()
{
 	if (Grabbing_) 
	{
		BIASERR("Camera is already grabbing.");
		return 0;
	}

  Grabbing_ = true;

	// Start Image Capturing
	//int res =	camera.StartImageCapture();
	int res =	camera.StartImageAcquisition();
   if (res == CAM_SUCCESS) 
	{
		BIASDOUT(D_DCAM,"VideoSource_DCAM::StartImageCapture successful");
	}
   else if (res == CAM_ERROR_INSUFFICIENT_RESOURCES) {
		BIASERR("VideoSource_DCAM::StartImageCapture failed (insufficient ressources), cam: " << SelectedCam_ << " result: " << res);
		Grabbing_ = false;
   } 
   else 
	{
		BIASERR("VideoSource_DCAM::StartImageCapture failed, cam: " << SelectedCam_ << " result: " << res);
		Grabbing_ = false;
	}
  return res;
}

int VideoSource_DCAM::PostGrab()
{
  // Stop Grabbing
	//camera.StopImageCapture();
	camera.StopImageAcquisition();
  Grabbing_=false;

  return 0;
}

unsigned int VideoSource_DCAM::CalcBytesPerPacket()
{
  unsigned short bpPacket;
  camera.GetCameraControlSize()->GetBytesPerPacket(&bpPacket);
	BytesPerPacket_ = bpPacket;
	return 0;
}

int  VideoSource_DCAM::ApplyBytesPerPacket()
{
	return 0;
}

int VideoSource_DCAM::StopIso() 
{
	return 0;
}

int VideoSource_DCAM::StartIso() 
{
	return 0;
}

float VideoSource_DCAM::GetShutter()
{
  camera.GetCameraControl(::FEATURE_SHUTTER)->Inquire();
  float shutter = -1;
  //if cam doesn't support absolute settings for shutter, it has to be
  //calculated from register data (long) -- TODO
  if (camera.GetCameraControl(::FEATURE_SHUTTER)->HasAbsControl())
    camera.GetCameraControl(::FEATURE_SHUTTER)->GetValueAbsolute(&shutter);
  else
    BIASERR("Cam doesn't support absolute values for shutter!!! FIXME");
	return shutter;
}

float VideoSource_DCAM::GetGain()
{  
  camera.GetCameraControl(::FEATURE_GAIN)->Inquire();
  float gain = -1;
  //if cam doesn't support absolute settings for gain, it has to be
  //calculated from register data (long) -- TODO
  if (camera.GetCameraControl(::FEATURE_GAIN)->HasAbsControl())
    camera.GetCameraControl(::FEATURE_GAIN)->GetValueAbsolute(&gain);
  else
    BIASERR("Cam doesn't support absolute values for shutter!!! FIXME");
	return gain;
}

void VideoSource_DCAM::PrintFormatAndMode_()
{
  BIASDOUT(D_DCAM,"Format: "<<DcamFormat_<<" Mode: "<<DcamMode_<<" ");
}

int VideoSource_DCAM::SetGain(float gain)
{
  if (camera.GetCameraControl(::FEATURE_GAIN)->Inquire() != CAM_SUCCESS) {
    BIASERR("Error getting settings for feature gain!!!");
    return -1;
  }

  //try to set gain via absolute value
  if (camera.GetCameraControl(::FEATURE_GAIN)->HasAbsControl()) {
    float min,max;
    camera.GetCameraControl(::FEATURE_GAIN)->GetRangeAbsolute(&min,&max);
    if ((gain < min)||(gain > max)) {
      BIASERR("Gain " << gain << " is out of range [" << min << ":" << max << "]");
      return -1;
    }
    if (!camera.GetCameraControl(::FEATURE_GAIN)->HasManualMode()) {
      BIASERR("Gain is not controlled manually!!!");
      return -1;
    }
    if (camera.GetCameraControl(::FEATURE_GAIN)->SetAutoMode(false)
        != CAM_SUCCESS) {
      BIASERR("Error setting gain to manual mode!!!");
      return -1;
    }
    if (camera.GetCameraControl(::FEATURE_GAIN)->SetAbsControl(true)
        != CAM_SUCCESS) {
      BIASERR("Error setting gain to manual mode!!!");
      return -1;
    }
    if (camera.GetCameraControl(::FEATURE_GAIN)->SetValueAbsolute(gain)
        != CAM_SUCCESS) {
      BIASERR("Error setting gain with SetValueAbsolute!!!");
      return -1;
    }
    return 0;
  }

  //if SetValueAbsolute is not supported calc register value
  switch(CamModel_)
  {
    case CAMERA_PTGREY_DRAGONFLY_COLOR:
      // Calculation from Technical Reference Manual for 640x480 Dragonfly
      unsigned short maxH,maxV;
      camera.GetCameraControlSize()->GetSizeLimits(&maxH,&maxV);
      if (maxV == 640)
      {
        if(gain < 512)
        {
          gain = -6,0404 + (20*log(658+gain)*(658-gain));
        }
        else
        {
          gain = -6,0404 + (0,0354*gain);
        }
      }
      // The camera is the 1024x768 Dragonfly
      else
      {
        if(gain < 512)
        {
          gain = -9,40 + (20*log(658+gain)*(658-gain));
        }
        else
        {
          gain = 9,40 + (0,0354*gain);
        }    
      }
      break;

    // see manual for more information, range is 0x800 to 0x8B4
    case CAMERA_SONY_DFW_X700:
      gain = gain * 8.0f + 0x800;
      if (gain > 2228.0f) gain = 2228.0f;
      if (gain < 2048.0f) gain = 2048.0f;
      break;

    default :
      BIASERR("VideoSource_DCAM::SetGain: Not implemented for this camera.");
      return -1;
      break;
  }

  // check if gain value can be set
  unsigned short min,max;
  camera.GetCameraControl(::FEATURE_GAIN)->GetRange(&min,&max);
  if (((int)gain < min) || ((int)gain > max))
  {
    BIASERR("Value for gain out of range.")
    return -1;
  }
  else if (!camera.GetCameraControl(::FEATURE_GAIN)->HasManualMode())
  {
    BIASERR("Feature is not set to be controlled manually.")
    return -1;
  }
  else
  {
    if (camera.GetCameraControl(::FEATURE_GAIN)->SetValue((unsigned short)gain) 
        != CAM_SUCCESS) {
      BIASERR("Error setting gain with CamControl->SetValue!!!");
      return -1;
    }
  }
	return 0;
}

int VideoSource_DCAM::SetWhiteBalance(float rvalue,float bvalue)
{
  // the range of the values is -0.5 to 0.5, so scale to 0-255
  rvalue = (rvalue + 0.5f) * 255.0f;
  bvalue = (bvalue + 0.5f) * 255.0f;

  // set min or max for values out of range
  if ( rvalue > 255.0f ) rvalue = 255.0;
  if ( bvalue > 255.0f ) bvalue = 255.0;
  if ( rvalue < 0.0f ) rvalue = 0.0;
  if ( bvalue < 0.0f ) bvalue = 0.0;

  camera.GetCameraControl(::FEATURE_WHITE_BALANCE)->Inquire();

  unsigned short min,max;
  camera.GetCameraControl(::FEATURE_WHITE_BALANCE)->GetRange(&min,&max);
  // check if whiteBalance values can be set.
  if ((rvalue < min) || (rvalue > max) || (bvalue < min) || (bvalue > max))
  {
    BIASERR("Values ("<<rvalue<<","<<bvalue<<") for white balance are"
            "out of range ["<<min<<":"<<max<<"]");
    return -1;
  }
  else if (!camera.GetCameraControl(::FEATURE_WHITE_BALANCE)->HasManualMode())
  {
    BIASERR("White balance is not set to be controlled manually.")
    return -1;
  }
  else
  {
    // conversion from float to int, maybe value must be computed.
    if (camera.GetCameraControl(::FEATURE_WHITE_BALANCE)
        ->SetValue((unsigned short)rvalue, (unsigned short)bvalue) != CAM_SUCCESS) {
      BIASERR("Error setting white balance with CamControl->SetValue!!!");
      return -1;
    }
  }
	return 0;
}

int VideoSource_DCAM::OnePushAuto( int feature)
{
  //!!!ATTENTION BIAS::FEATURE... may differ from ::FEATURE... (from CMU)!!!
  int res = 0;

  switch(feature)
  {//?????????? MAYBE AutoMode must be activated for OnePush-mode ??????????????
    case FEATURE_BRIGHTNESS:
      res = camera.GetCameraControl(::FEATURE_BRIGHTNESS)->SetOnePush(true);
      break;
    case FEATURE_EXPOSURE:
      res = camera.GetCameraControl(::FEATURE_AUTO_EXPOSURE)->SetOnePush(true);
      break;
    case FEATURE_SHARPNESS :
      res = camera.GetCameraControl(::FEATURE_SHARPNESS)->SetOnePush(true);
      break;
    case FEATURE_WHITE_BALANCE :
      res = camera.GetCameraControl(::FEATURE_WHITE_BALANCE)->SetOnePush(true);
      break;
    case FEATURE_HUE:
      res = camera.GetCameraControl(::FEATURE_HUE)->SetOnePush(true);
      break;
    case FEATURE_SATURATION:
      res = camera.GetCameraControl(::FEATURE_SATURATION)->SetOnePush(true);
      break;
    case FEATURE_GAMMA:
      res = camera.GetCameraControl(::FEATURE_GAMMA)->SetOnePush(true);
      break;
    case FEATURE_SHUTTER:
      res = camera.GetCameraControl(::FEATURE_SHUTTER)->SetOnePush(true);
      break;
    case FEATURE_GAIN:
      res = camera.GetCameraControl(::FEATURE_GAIN)->SetOnePush(true);
      break;
    case FEATURE_IRIS:
      res = camera.GetCameraControl(::FEATURE_IRIS)->SetOnePush(true);
      break;
    case FEATURE_FOCUS:
      res = camera.GetCameraControl(::FEATURE_FOCUS)->SetOnePush(true);
      break;
    case FEATURE_TRIGGER:
      break;
    case FEATURE_ZOOM:
      res = camera.GetCameraControl(::FEATURE_ZOOM)->SetOnePush(true);
      break;
    case FEATURE_PAN:
      break;
    case FEATURE_TILT:
      break;
    case FEATURE_OPTICAL_FILTER:
      break;
    default:
      BIASERR("VideoSource_DCAM::SetAutoMode(): Unkown Feature");
      return -1;
  }

  if (res != CAM_SUCCESS)
  {
    BIASERR("Failed to set OnePush!!!")
    return -1;
  }

	return 0;
}

int VideoSource_DCAM::SetShutter(float exptime)
{
  //if SetValueAbsolute is not supported calc register value
  switch(CamModel_)
  {
    // see manual for more information, range is 0x7E2 to 0xB20
    case CAMERA_SONY_DFW_X700:
      exptime = 2848.0f - exptime * 10000.0f;
      if (exptime > 2848.0f) exptime = 2848.0f;
      if (exptime < 2100.0f) exptime = 2100.0f;
      if (camera.GetCameraControl(::FEATURE_SHUTTER)->SetValue((unsigned short)exptime) == CAM_SUCCESS)
        return 0;
      break;
  }

  if (camera.GetCameraControl(::FEATURE_SHUTTER)->Inquire() != CAM_SUCCESS) {
      BIASERR("Error reading shutter settings!!!");
      return -1;
  }

  //try to set gain via absolute value
  if (camera.GetCameraControl(::FEATURE_SHUTTER)->HasAbsControl()) {
    float min,max;
    camera.GetCameraControl(::FEATURE_SHUTTER)->GetRangeAbsolute(&min,&max);
    cout<<"Min: "<<min<<" Max: "<<max<<endl;
    if ((exptime < min)||(exptime > max)) {
      BIASERR("Shutter " << exptime << " is out of range [" << min << ":" << max << "]");
      return -1;
    }
    if (!camera.GetCameraControl(::FEATURE_SHUTTER)->HasManualMode()) {
      BIASERR("Shutter is may not be controlled manually!!!");
      return -1;
    }
    if (camera.GetCameraControl(::FEATURE_SHUTTER)->SetAutoMode(false)
        != CAM_SUCCESS) {
      BIASERR("Error setting shutter to manual mode!!!");
      return -1;
    }
    if (camera.GetCameraControl(::FEATURE_SHUTTER)->SetAbsControl(true)
        != CAM_SUCCESS) {
      BIASERR("Error setting shutter to AbsControl-mode!!!");
      return -1;
    }
    int res = camera.GetCameraControl(::FEATURE_SHUTTER)->SetValueAbsolute(exptime);
    if (res != CAM_SUCCESS) {
      BIASERR("Error setting shutter with SetValueAbsolute, returns "<<res<<" !!!");
      return -1;
    }
    return 0;
  }

  BIASERR("Error: setting shutter with SetValueAbsolute not supported!!!");
  return -1;
}

int VideoSource_DCAM::SetLeftTop(unsigned int left, unsigned int top)
{
  int res;

  if (left<=(MaxWidth_ - Width_) && top <=(MaxHeight_ - Height_)) {
    Top_ = top;
    Left_ = left;
    UserTopLeft_ = true;
  }
  else {
    BIASERR("VideoSource_DCAM::SetLeftTop() values too large:"<<left<<"x"<<top)
    return -1;
  }		    
  
  // try to set defined position, if cam is initialized.
  if(CompleteInitialized_){
    res = camera.GetCameraControlSize()->SetPos((unsigned short)left,
                                              (unsigned short)top);

    if (res == CAM_SUCCESS)
    {
      BIASDOUT(D_DCAM,"VideoSource_DCAM::SetLeftTop: successful");
      return 0;
    } 
     else
    {
      BIASERR("VideoSource_DCAM::SetLeftTop: failed");
      return -1;
    }
  }
return 0;
}

unsigned int VideoSource_DCAM::GetFeature(int feature)
{
  unsigned short value = 0;

  switch(feature)
  {
    case FEATURE_BRIGHTNESS:
      camera.GetCameraControl(::FEATURE_BRIGHTNESS)->Inquire();
      camera.GetCameraControl(::FEATURE_BRIGHTNESS)->GetValue(&value);
      break;
    case FEATURE_EXPOSURE:
      BIASERR("Can't get exposure, get FEATURE_SHUTTER instead!!!");
      break;
    case FEATURE_SHARPNESS :
      camera.GetCameraControl(::FEATURE_SHARPNESS)->Inquire();
      camera.GetCameraControl(::FEATURE_SHARPNESS)->GetValue(&value);
      break;
    case FEATURE_WHITE_BALANCE :
      camera.GetCameraControl(::FEATURE_WHITE_BALANCE)->Inquire();
      camera.GetCameraControl(::FEATURE_WHITE_BALANCE)->GetValue(&value);
      break;
    case FEATURE_HUE:
      camera.GetCameraControl(::FEATURE_HUE)->Inquire();
      camera.GetCameraControl(::FEATURE_HUE)->GetValue(&value);
      break;
    case FEATURE_SATURATION:
      camera.GetCameraControl(::FEATURE_SATURATION)->Inquire();
      camera.GetCameraControl(::FEATURE_SATURATION)->GetValue(&value);
      break;
    case FEATURE_GAMMA:
      camera.GetCameraControl(::FEATURE_GAMMA)->Inquire();
      camera.GetCameraControl(::FEATURE_GAMMA)->GetValue(&value);
      break;
    case FEATURE_SHUTTER:
      camera.GetCameraControl(::FEATURE_SHUTTER)->Inquire();
      camera.GetCameraControl(::FEATURE_SHUTTER)->GetValue(&value);
      break;
    case FEATURE_GAIN:
      camera.GetCameraControl(::FEATURE_GAIN)->Inquire();
      camera.GetCameraControl(::FEATURE_GAIN)->GetValue(&value);
      break;
    case FEATURE_IRIS:
      camera.GetCameraControl(::FEATURE_IRIS)->Inquire();
      camera.GetCameraControl(::FEATURE_IRIS)->GetValue(&value);
      break;
    case FEATURE_FOCUS:
      camera.GetCameraControl(::FEATURE_FOCUS)->Inquire();
      camera.GetCameraControl(::FEATURE_FOCUS)->GetValue(&value);
      break;
    case FEATURE_TRIGGER:
      break;
    case FEATURE_ZOOM:
      camera.GetCameraControl(::FEATURE_ZOOM)->Inquire();
      camera.GetCameraControl(::FEATURE_ZOOM)->GetValue(&value);
      break;
    case FEATURE_PAN:
      break;
    case FEATURE_TILT:
      break;
    case FEATURE_OPTICAL_FILTER:
      break;
    default:
      BIASERR("VideoSource_DCAM::GetFeature(): Unkown Feature");
      break;
  }
  return value;
}


int VideoSource_DCAM::SetFeature(int feature, unsigned int value)
{
  int res = -1;
  unsigned short sValue = (unsigned short) value;

  switch(feature)
  {
    case FEATURE_BRIGHTNESS:
      res = camera.GetCameraControl(::FEATURE_BRIGHTNESS)->SetValue(sValue);
      break;
    case FEATURE_EXPOSURE:
      BIASERR("Can't set exposure, use FEATURE_SHUTTER instead!!!");
      return -1;
      break;
    case FEATURE_SHARPNESS :
      res = camera.GetCameraControl(::FEATURE_SHARPNESS)->SetValue(sValue);
      break;
    case FEATURE_WHITE_BALANCE :
      res = camera.GetCameraControl(::FEATURE_WHITE_BALANCE)->SetValue(sValue);
      break;
    case FEATURE_HUE:
      res = camera.GetCameraControl(::FEATURE_HUE)->SetValue(sValue);
      break;
    case FEATURE_SATURATION:
      res = camera.GetCameraControl(::FEATURE_SATURATION)->SetValue(sValue);
      break;
    case FEATURE_GAMMA:
      res = camera.GetCameraControl(::FEATURE_GAMMA)->SetValue(sValue);
      break;
    case FEATURE_SHUTTER:
      res = camera.GetCameraControl(::FEATURE_SHUTTER)->SetValue(sValue);
      break;
    case FEATURE_GAIN:
      res = camera.GetCameraControl(::FEATURE_GAIN)->SetValue(sValue);
      break;
    case FEATURE_IRIS:
      res = camera.GetCameraControl(::FEATURE_IRIS)->SetValue(sValue);
      break;
    case FEATURE_FOCUS:
      res = camera.GetCameraControl(::FEATURE_FOCUS)->SetValue(sValue);
      break;
    case FEATURE_TRIGGER:
      break;
    case FEATURE_ZOOM:
      res = camera.GetCameraControl(::FEATURE_ZOOM)->SetValue(sValue);
      break;
    case FEATURE_PAN:
      break;
    case FEATURE_TILT:
      break;
    case FEATURE_OPTICAL_FILTER:
      break;
    default:
      BIASERR("VideoSource_DCAM::SetFeature(): Unkown Feature");
      return -1;
  }

  if (res != CAM_SUCCESS) {
    BIASERR("Error setting feature "<<feature<<", SetValue returned "<<res);
    return -1;
  }

	return 0;
}

int VideoSource_DCAM::AutoMode( int feature, bool on)
{
  int res = 0;

  switch(feature)
    {
    case FEATURE_BRIGHTNESS:
      res = camera.GetCameraControl(::FEATURE_BRIGHTNESS)->SetAutoMode(on);
      break;
    case FEATURE_EXPOSURE:
      res = camera.GetCameraControl(::FEATURE_AUTO_EXPOSURE)->SetAutoMode(on);
      break;
    case FEATURE_SHARPNESS :
      res = camera.GetCameraControl(::FEATURE_SHARPNESS)->SetAutoMode(on);
      break;
    case FEATURE_WHITE_BALANCE :
      res = camera.GetCameraControl(::FEATURE_WHITE_BALANCE)->SetAutoMode(on);
      break;
    case FEATURE_HUE:
      res = camera.GetCameraControl(::FEATURE_HUE)->SetAutoMode(on);
      break;
    case FEATURE_SATURATION:
      res = camera.GetCameraControl(::FEATURE_SATURATION)->SetAutoMode(on);
      break;
    case FEATURE_GAMMA:
      res = camera.GetCameraControl(::FEATURE_GAMMA)->SetAutoMode(on);
      break;
    case FEATURE_SHUTTER:
      res = camera.GetCameraControl(::FEATURE_SHUTTER)->SetAutoMode(on);
      break;
    case FEATURE_GAIN:
      res = camera.GetCameraControl(::FEATURE_GAIN)->SetAutoMode(on);
      break;
    case FEATURE_IRIS:
      res = camera.GetCameraControl(::FEATURE_IRIS)->SetAutoMode(on);
      break;
    case FEATURE_FOCUS:
      res = camera.GetCameraControl(::FEATURE_FOCUS)->SetAutoMode(on);
      break;
    case FEATURE_TRIGGER:
      break;
    case FEATURE_ZOOM:
      res = camera.GetCameraControl(::FEATURE_ZOOM)->SetAutoMode(on);
      break;
    case FEATURE_PAN:
      break;
    case FEATURE_TILT:
      break;
    case FEATURE_OPTICAL_FILTER:
      break;
    default:
      BIASERR("VideoSource_DCAM::SetAutoMode(): Unkown Feature");
      return -1;
    }

  if (res != CAM_SUCCESS) {
    BIASERR("Error setting AutoMode for feature "<<feature
             <<", SetAutoMode("<<on<<") returned "<<res<<" !!!");
    return -1;
  }

  return 0;
}



void VideoSource_DCAM::SetAutoShutter(bool b)
{
  if (camera.GetCameraControl(::FEATURE_SHUTTER)->SetAutoMode(b)
      != CAM_SUCCESS) {
    BIASERR("Error setting shutter to auto mode!!!");
  }
}

bool VideoSource_DCAM::GetAutoShutter()
{
  return camera.GetCameraControl(::FEATURE_SHUTTER)->HasAutoMode();
}

void VideoSource_DCAM::SetAutoGain(bool b)
{
  if (camera.GetCameraControl(::FEATURE_GAIN)->SetAutoMode(b)
      != CAM_SUCCESS) {
    BIASERR("Error setting gain to auto mode!!!");
  }
}

bool VideoSource_DCAM::GetAutoGain()
{
  return camera.GetCameraControl(::FEATURE_GAIN)->HasAutoMode();
}
