/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
Multimediale Systeme der Informationsverarbeitung
Institut fuer Informatik
Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef __VIDEOSOURCE_SWISSRANGER_HH__
#define __VIDEOSOURCE_SWISSRANGER_HH__

#include <VideoSource/VideoSource_Base.hh>
#include <Image/Camera.hh>
#include <bias_config.h>
#include <vector>
#include <libusbSR.h>

namespace BIAS {

  /** @class VideoSource_SwissRanger
  @brief Support for SwissRanger usb cam
  @author fkellner 10/2007 */
  class BIASVideoSource_EXPORT VideoSource_SwissRanger : public VideoSource {

  public:
    VideoSource_SwissRanger();

    ~VideoSource_SwissRanger();

    //try to open first unused PMD-camera
    virtual int OpenDevice();
    //try to open a specific camera-device
    virtual int OpenDevice(int camNr);
    //close an active camera
    virtual int CloseDevice();

    //no needed for now
    virtual int PreGrab();
    virtual int PostGrab();


    /** 
    Returns current framerate
    */
    double GetFPS();

    virtual int InitImage(ImageBase &img);
    virtual int InitDepthImage(ImageBase &img);

    /** Returns the 2D image, and reads depth and modulation coefficients */
    virtual int GrabSingle(Camera<unsigned char> &image);

    /// Set IntegrationTime
    virtual int SetIntegrationTime(const unsigned char inttime) {
      integrationTime_ = inttime;
      return SR_SetIntegrationTime(srCam_, inttime);
    }

    virtual int GetIntegrationTime(unsigned char &val) {
      val = integrationTime_;
      return 0;
    }

    /* Set the Modulation frequency */
    int SetModulationFrequency(const unsigned char dModulationFrequency) {
      modulationFrequency_ = (ModulationFrq)dModulationFrequency;
      SetMaxDist_();
      return SR_SetModulationFrequency(srCam_, (ModulationFrq)modulationFrequency_);
    }
    /* Get the Modulation frequency*/
    int GetModulationFrequency(unsigned char &val) {
      val = (unsigned char)modulationFrequency_;
      return 0;
    }

    /* Set Amplitude Threshold */
    int SetAmplitudeThreshold(const unsigned short val) {
      amplitudeThreshold_ = val;
      return SR_SetAmplitudeThreshold(srCam_, amplitudeThreshold_);
    }

    /* Get Amplitude Threshold */
    inline int GetAmplitudeThreshold(unsigned short &val) {
      val = amplitudeThreshold_;
      return 0;
    }

    /* Set Saturation Threshold */
    int SetSaturationThreshold(const unsigned short val) {
      saturationThreshold_ = val;
      return SR_SetAmplitudeThreshold(srCam_, saturationThreshold_);
    }

    /* Get Saturation Threshold */
    inline int GetSaturationThreshold(unsigned short &val) {
      val = saturationThreshold_;
      return 0;
    }

    /* Set Distance Offset */
    int SetDistanceOffset(const unsigned char val) {
      distanceOffset_ = val;
      return SR_SetDistanceOffset(srCam_, distanceOffset_);
    }

    /* Get Distance Offset */
    int GetDistanceOffset(unsigned char &val) {
      val = distanceOffset_;
      return 0;
    }

    /** Returns the depth image. Must be called after GrabSingle() */
    inline int GrabSingleDepth(Camera<float> &image)
    { 
      if (!DepthImgF_.IsEmpty()) {
        image = DepthImgF_; 
        return 0;
      } 
      return -1;
    }

    /** Returns the amplitude image. Must be called after GrabSingle() */
    inline int GrabSingleModCoeff(Camera<float> &image)
    {       
      if (!ModCoeffImg_.IsEmpty()) {
        image = ModCoeffImg_; 
        return 0;
      } 
      return -1;
    }

    void SetGenerateNewUIDs(bool s) {  
      BIASERR("Always generating new UUIDS for PMD images");
      BIASABORT;
    }
    
    inline int SetAutoIllumination(const bool on) {
      if (on)
        return SetAutoIllumination(5, 255, 10, 45);
      else
        return SetAutoIllumination(0xff, 0, 0, 0);
    }
    
    inline int SetAutoIllumination(const unsigned char minIntTime, 
                                   const unsigned char maxIntTime,
                                   const unsigned char percentOverPos, 
                                   const unsigned char desiredPos) {
      return SR_SetAutoIllumination(srCam_, minIntTime, maxIntTime,
                                    percentOverPos, desiredPos);
    }
    
    inline int SetScatteringFilter(const unsigned char blkSzX, 
                                   const unsigned char blkSzY) {
      return SR_SetScatteringFilter(srCam_, blkSzX, blkSzY);
    }
    
    void EnsureLatestFrame(const bool ensure) {
      bEnsureLatestFrame_ = ensure;
    }
    
  protected:
    void ConvertImageFloatToCharNormalized_(const Image<float> src, Image<unsigned char> &dst);
    void ConvertRawImageToFloat_(const unsigned short* data, Image<float> &dst);
    void ConvertRawImage_(const unsigned short* data, Image<float> &dst);

    void SetMaxDist_() {
      float lightspeed = 299792.485f;
      switch (modulationFrequency_) {
        case MF_30MHz:
          max_dist_ = (lightspeed / 30.0f) / 2.0f;
          break;
        case MF_21MHz:
          max_dist_ = (lightspeed / 21.0f) / 2.0f;
          break;
        case MF_20MHz:
          max_dist_ = (lightspeed / 20.0f) / 2.0f;
          break;
        case MF_19MHz:
          max_dist_ = (lightspeed / 19.0f) / 2.0f;
          break;
        default:
          BIASERR("Can not set modulation frequency for SwissRanger!");
          break;
      }
    }
    bool bEnsureLatestFrame_;
    unsigned short amplitudeThreshold_;
    unsigned short saturationThreshold_;
    unsigned char integrationTime_;
    unsigned char distanceOffset_;
    int modulationFrequency_;

    bool bIsOpen_;

    //MIP images, holding grabbed data
    BIAS::Camera<unsigned char> DepthImg_;
    BIAS::Camera<float> DepthImgF_;
    BIAS::Camera<float> ModCoeffImg_;

    SwissrangerCam* srCam_;

    unsigned short* depth_;
    float* modcoeff_;

    float max_dist_;
    size_t bufSz_;
  };

}

#endif //__VIDEOSOURCE_SWISSRANGER_HH__

